/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "CGui.hpp"
#include <signal.h>
#include <iostream>
#include "libviz/CBitmap.hpp"
using namespace std;


bool CGui::takeScreenshot(string &filename)
{
	CBitmap24 bitmap(screen_width, screen_height);

	glReadPixels(0, 0, screen_width, screen_height, GL_BGR, GL_UNSIGNED_BYTE, bitmap.data);

	if (!bitmap.save(filename))
	{
		cout << bitmap.errorLog << endl;
		return false;
	}

	return true;
}

void CGui::init_rendering()
{
}

void CGui::finish_rendering()
{
	SDL_GL_SwapBuffers();
}


void CGui::set_video_mode(bool fullscreen)
{
	if (fullscreen && fullscreen_active)
		return;

	Uint32 flags = SDL_OPENGL;

	if (fullscreen)
	{
		screen_height = fullscreen_height;
		screen_width = fullscreen_width;
		flags |= SDL_FULLSCREEN;
		fullscreen_active = true;
	}
	else
	{
		screen_height = window_height;
		screen_width = window_width;
		flags |= SDL_RESIZABLE;
		fullscreen_active = false;
	}

//	const SDL_VideoInfo* video = SDL_GetVideoInfo();
	if( SDL_SetVideoMode(screen_width, screen_height, 32, flags) == 0 )
//	if( SDL_SetVideoMode(screen_width, screen_height, video->vfmt->BitsPerPixel, flags) == 0 )
	{
		fprintf(stderr, "Couldn't set video mode with 32bpp: %s\n", SDL_GetError());
		exit(1);
	}
	const SDL_VideoInfo* video = SDL_GetVideoInfo();
	if (video->vfmt->BitsPerPixel < 32)
	{
		cerr << "the video device is set to " << (int)video->vfmt->BitsPerPixel << " bits per pixel, but 32 are necessary for framebuffers (at least on nvidia hardware)... ABORTING" << endl;
		exit(1);
	}

	screen_aspect = (float)screen_width/(float)screen_height;

	setup_viewport();

	if (callback_viewport_changed)
		callback_viewport_changed(callback_class, screen_width, screen_height);
}


CGui::CGui() :	// public
	screen_height(480),	screen_width(640),
	fullscreen_active(false),
	callback_key_down(0), callback_key_up(0),
	callback_quit(0),
	callback_mouse_motion(0),
	callback_mouse_button_down(0), callback_mouse_button_up(0),
	callback_viewport_changed(0),

	// private
	window_height(480),	window_width(640),
	fullscreen_height(-1),	fullscreen_width(-1)
{
	const SDL_VideoInfo* video;

	if ( SDL_Init(SDL_INIT_VIDEO) < 0 )
	{
		fprintf(stderr, "Couldn't initialize SDL: %s\n", SDL_GetError());
		exit(1);
	}

	video = SDL_GetVideoInfo();

	// activate ctrl-c abort handling
#ifndef __WIN32__
	signal(SIGINT, SIG_DFL);
	signal(SIGQUIT, SIG_DFL);
#endif

	if( !video )
	{
		fprintf(stderr, "Couldn't get video information: %s\n", SDL_GetError());
		exit(1);
	}

	/* Set the minimum requirements for the OpenGL window */
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5 );
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16 );
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1 );

	// get available sdl modes
	SDL_Rect **modes;

	/* get available fullscreen/hardware modes */
	modes = SDL_ListModes(NULL, SDL_FULLSCREEN | SDL_HWSURFACE);

	/* check if there are any modes available */
	if(modes != (SDL_Rect **)0)
	{
		/* Print valid modes */
		for(int i=0; modes[i]; i++)
		{
			printf("%i x %i    ", modes[i]->w, modes[i]->h);
			if (modes[i]->w > fullscreen_width || modes[i]->h > fullscreen_height)
			{
				fullscreen_width = modes[i]->w;
				fullscreen_height = modes[i]->h;
			}
		}

		printf("\nUsing (%i x %i) as fullscreen mode\n", fullscreen_width, fullscreen_height);
	}

	// CEGui Stuff
	SDL_EnableUNICODE(1);
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
//	SDL_ShowCursor(SDL_DISABLE);

	set_video_mode(false);

}

CGui::~CGui()
{
	SDL_Quit();
}

void CGui::setup_viewport()
{
	glViewport(0, 0, screen_width, screen_height);
}

signed int CGui::get_ticks()
{
	return SDL_GetTicks();
}

void CGui::event_loop() 
{
	SDL_Event event;
	// process pending events
	while(SDL_PollEvent(&event)) 
	{
		switch(event.type)
		{
			case SDL_KEYDOWN:
				if (callback_key_down)
					callback_key_down(callback_class, event.key.keysym.sym, event.key.keysym.mod, event.key.keysym.scancode, event.key.keysym.unicode);
				break;

			case SDL_KEYUP:
				if (callback_key_up)
					callback_key_up(callback_class, event.key.keysym.sym, event.key.keysym.mod, event.key.keysym.scancode, event.key.keysym.unicode);
				break;

			case SDL_QUIT:
				if (callback_quit)
					callback_quit(callback_class);
				break;

			case SDL_MOUSEMOTION:
				if (callback_mouse_motion)
					callback_mouse_motion(callback_class, event.motion.x, event.motion.y);
				break;

			case SDL_MOUSEBUTTONDOWN:
				if (callback_mouse_button_down)
					callback_mouse_button_down(callback_class, event.button.button);
				break;


			case SDL_MOUSEBUTTONUP:
				if (callback_mouse_button_up)
					callback_mouse_button_up(callback_class, event.button.button);
				break;

			case SDL_VIDEORESIZE:
				window_width = event.resize.w;
				window_height = event.resize.h;
				set_video_mode(false);
				break;
		}
	}
}
