/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "CGuiSimulation.hpp"
#include <math.h>
#include "libviz/dspObjfile.hpp"
#include "libviz/CPrimitives.hpp"
#include "libviz/CMarchingCubes.hh"
#include "libviz/CTrackBall.hh"
#include "libmath/CMatrix3.hh"


#if HAVE_LIBCEGUIBASE
	#include <CEGUIDefaultResourceProvider.h>
#endif

#define SCALE_TEXTURE	1.0

#define DISPLACE_MARCHING_CUBES	1

#define RENDER_DATA_TEXTURE_MAX	14

#define PARTICLES_EMITTED_PER_ROUND	20
#define PARTICLES_MAX			20000

#define PARTICLE_TEXTURE_WIDTH	PARTICLES_EMITTED_PER_ROUND
#define PARTICLE_TEXTURE_HEIGHT	(PARTICLES_MAX/PARTICLES_EMITTED_PER_ROUND)


#define setCEGUIWindowStyle()													\
	myControlWindow->setPosition( CEGUI::UVector2( CEGUI::UDim( 0, 10 ), CEGUI::UDim( 0, gui->screen_height-200-10 ) ) );	\
	myControlWindow->setSize( CEGUI::UVector2( CEGUI::UDim( 0, 620 ), CEGUI::UDim( 0, 200) ) );				\


/**
 * on quit event, just set the quit_program to 'true'
 * this value is returned in the render loop
 */
void CGuiSimulation::gui_callback_quit()
{
	quit_program = true;
}


/**
 * if the viewport size is changed, this function is called
 */
void CGuiSimulation::gui_callback_viewport_changed(int width, int height)
{
#if HAVE_LIBCEGUIBASE
	myGuiRenderer->setDisplaySize(CEGUI::Size(width, height));
	setCEGUIWindowStyle();
#endif
}


/**
 * keypress event from the gui class
 */
void CGuiSimulation::gui_callback_key_down(int key, int mod, int scancode, int unicode)
{
#if HAVE_LIBCEGUIBASE
	if (guiVisible)
	{
//		CEGUI::System::getSingleton().renderGUI();

		CEGUI::System::getSingleton().injectKeyDown(scancode);
		if (unicode != 0)
		{
			CEGUI::System::getSingleton().injectChar(unicode);
		}
	}
#endif

	if (key >= '0' && key <= '9')
	{
		int nbr = key - '0';
		steps_to_do = nbr*nbr*nbr;
		return;
	}

	switch(key)
	{
		case 'q':	case 'Q':	// quit
			quit_program = true;
			break;

		case 'f':	case 'F':	// switch to fullscreen
			gui->set_video_mode(!gui->fullscreen_active);
			break;

		case 'r':	case 'R':	// reset simulation
			reset();
			break;

		case 't':	case 'T':	// toggle through rendering data textures or not to render
			render_data_texture = (render_data_texture+((mod & KMOD_SHIFT) ? -1 : 1)+RENDER_DATA_TEXTURE_MAX+1) % (RENDER_DATA_TEXTURE_MAX+1);
			break;

		case 's':	case 'S':	// output some statistics
			print_statistics = !print_statistics;
			break;

		case 'p':	case 'P':	// render particles
			render_particles = !render_particles;
			break;

		case 'w':	case 'W':	// render wireframe
			render_wireframe = !render_wireframe;
			break;

		case 'm':	case 'M':	// render marching_cubes 
			render_marching_cubes = !render_marching_cubes;
			break;

		case 'j':	case 'J':	// render next single marching cube
			single_marching_cube_id = (single_marching_cube_id+1)%256;
			cout << "cube id: " << single_marching_cube_id << endl;
			break;

		case 'k':	case 'K':	// render prev single marching cube
			single_marching_cube_id = (single_marching_cube_id+255)%256;
			cout << "cube id: " << single_marching_cube_id << endl;
			break;

		case 'n':	case 'N':	// render single marching cube
			render_single_marching_cube = !render_single_marching_cube;
			break;

		case 'v':	case 'V':	// render volume
			render_volume = !render_volume;
			break;

		case 'x':	case 'X':	// take screenshots
			take_screenshots = !take_screenshots;
			break;

		case 'u':	case 'U':	// visible trackball
			visible_trackball = !visible_trackball;
			break;


		case 'h':	case 'H':	// print help
			cout << "	q	quit program" << endl;
			cout << "	f	toggle fullscreen" << endl;
			cout << "	r	reset simulation" << endl;
			cout << "	t	cycle through textures" << endl;
			cout << "	v	render fluid volume" << endl;
			cout << "	m	render fluid marching cubes" << endl;
			cout << "	p	render particles" << endl;
			cout << "	x	toggle creating screenshots" << endl;
			cout << endl;
			cout << "	[enter]	run / stop simulation" << endl;
			cout << "	[space]	toggle gui display" << endl;
			cout << "	s	toggle statistics display" << endl;
			cout << "	v	toggle visibility of trackball" << endl;
			break;

		case SDLK_RETURN:
			run_simulation = !run_simulation;
#if HAVE_LIBCEGUIBASE
			myRunSimulation->setSelected(run_simulation);
#endif
			break;

		case ' ':	// reload gui
			guiVisible = !guiVisible;
//			reloadCEGui();
			break;
	}
}


#if HAVE_LIBCEGUIBASE
void CGuiSimulation::initCEGui()
{
	using namespace CEGUI;

	try
	{
		myGuiRenderer = new CEGUI::OpenGLRenderer(0, gui->screen_width, gui->screen_height);
		new CEGUI::System(myGuiRenderer);

		// initialise the required dirs for the DefaultResourceProvider
		CEGUI::DefaultResourceProvider* rp = static_cast<CEGUI::DefaultResourceProvider*>(CEGUI::System::getSingleton().getResourceProvider());

		rp->setResourceGroupDirectory("schemes", "./cedatafiles/schemes/");
		rp->setResourceGroupDirectory("imagesets", "./cedatafiles/imagesets/");
		rp->setResourceGroupDirectory("fonts", "./cedatafiles/fonts/");
		rp->setResourceGroupDirectory("layouts", "./datafiles/layouts/");
		rp->setResourceGroupDirectory("looknfeels", "./cedatafiles/looknfeel/");
		rp->setResourceGroupDirectory("lua_scripts", "./cedatafiles/lua_scripts/");

		Imageset::setDefaultResourceGroup("imagesets");
		Font::setDefaultResourceGroup("fonts");
		Scheme::setDefaultResourceGroup("schemes");
		WidgetLookManager::setDefaultResourceGroup("looknfeels");
		WindowManager::setDefaultResourceGroup("layouts");
		ScriptModule::setDefaultResourceGroup("lua_scripts");

		SchemeManager::getSingleton().loadScheme("WindowsLook.scheme");
		System::getSingleton().setDefaultMouseCursor("WindowsLook", "MouseArrow");
		Font *font = FontManager::getSingleton().createFont("DejaVuSans-10.font");
		font->setProperty("AutoScaled", "false");

		reloadCEGui();

		myRunSimulation->setSelected(run_simulation);
	}
	catch (CEGUI::Exception& exc)
	{
		cerr << exc.getMessage().c_str() << endl;
	}
	catch (std::exception& exc)
	{
		cerr << exc.what() << endl;;
	}
	catch(...)
	{
		cout << "unknown exception catched" << endl;
	}
}
#endif

/**
 * get value in flat texture  at position x,y,z
 * NO checks for out of boundary!
 * \param 	ff - the float data - 4 floats for every point
 * \param	xyz - 3d position
 * \param	res_x, res_y - simulation resolution
 * \param	ft_z_mod - slices in x direction
 */
inline float *getValuePosition(float *ff, int x, int y, int z, int res_x, int res_y, int ft_z_mod)
{
	int block_row = z / ft_z_mod;
	int block_col = z % ft_z_mod;

	int whole_block_row_bytes = block_row * res_x*res_y*ft_z_mod;
	int whole_block_col_bytes = y*res_x*ft_z_mod + block_col*res_x;

	return ff + 4*(whole_block_row_bytes + whole_block_col_bytes + x);
}



void CGuiSimulation::renderMarchingCube(int mc_index)
{
	glMatrixMode(GL_MODELVIEW);
	glEnable(GL_NORMALIZE);
	glPushMatrix();
//	glLoadIdentity();

	glTranslatef(-0.5,-0.5,-0.5);
//	glScalef(2,2,2);

	glColor3f(1.0, 0.0, 0);	// water should be blue :)
	for (int i = 0; i < 8; i++)
	{
		if (mc_index & (1 << i))
		{
			int x = (i>>0)&1;
			int y = (i>>1)&1;
			int z = (i>>2)&1;

			CPrimitives::sphere(0.1, (float)x, (float)y, (float)z, 100);
		}
	}

	glEnable(GL_BLEND);

	for (int s = 0; s < 4; s++)	// surfaces
	{
		if (marching_cubes[mc_index][0][0][0] == -111.0)
			break;

		for (int p = 0; p < MC_MAX_POLYGON_COUNT; p++)	// polygons
		{
			if (marching_cubes[mc_index][p][0][0] == -111.0)
				break;

			for (int i = 1; i < 4; i++)
			{
				glColor4f(0.0, 1.0, 0.0, 1.0);
				glBegin(GL_LINES);
				glVertex3fv(marching_cubes[mc_index][p][i]);
				glVertex3f(	marching_cubes[mc_index][p][i][0]+marching_cubes[mc_index][p][0][0]*0.2,
						marching_cubes[mc_index][p][i][1]+marching_cubes[mc_index][p][0][1]*0.2,
						marching_cubes[mc_index][p][i][2]+marching_cubes[mc_index][p][0][2]*0.2
						);
				glEnd();
			}

			glColor4f(0.2, 0.2, 1, 0.5);
			glEnable(GL_CULL_FACE);
			glBegin(GL_TRIANGLES);
			glNormal3fv(marching_cubes[mc_index][p][0]);
			glVertex3fv(marching_cubes[mc_index][p][1]);
			glVertex3fv(marching_cubes[mc_index][p][2]);
			glVertex3fv(marching_cubes[mc_index][p][3]);
			glEnd();
			glDisable(GL_CULL_FACE);
		}
	}
	glDisable(GL_BLEND);
	glPopMatrix();
}



void CGuiSimulation::renderMarchingCubes()
{
	CTexture *t = simulation->fluid_fraction_texture;
	float *ff = new float[t->width*t->height*4];
	t->getData(ff);

	glMatrixMode(GL_MODELVIEW);
	glEnable(GL_NORMALIZE);
	glPushMatrix();

	/**
	 * here the storage of the texture really sucks
	 * but later this should be driven directly on GPU
	 *
	 * and of course this whole function can be improved a lot but it's just for visualization - not for speed
	 */

	if (render_wireframe)
	{
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glDisable(GL_CULL_FACE);
	}
	else
	{
		glEnable(GL_CULL_FACE);
//		glDisable(GL_CULL_FACE);
	}

	int x, y, z;
	float px, py, pz;
	float *p000, *p001, *p010, *p011, *p100, *p101, *p110, *p111;
	float inv_x = 1.0f/(float)simulation->fres[0];
	float inv_y = 1.0f/(float)simulation->fres[1];
	float inv_z = 1.0f/(float)simulation->fres[2];

	glColor3f(0.2, 0.2, 1);	// water should be blue :)

	// handle offset within data right here!
	float *ff_handler = ff;
	ff += 0;

	for (z = 0; z < simulation->res_z-1; z++)
	{
		int block_row = z / simulation->ft_z_mod;
		int block_col = z % simulation->ft_z_mod;

		int whole_block_row_bytes = block_row * simulation->res_x*simulation->res_y*simulation->ft_z_mod;

		pz = (float)z / (float)simulation->res_z - 0.5 + 0.5/(float)simulation->res_z;
		for (y = 0; y < simulation->res_y-1; y++)
		{
			py = (float)y / (float)simulation->res_y - 0.5 + 0.5/(float)simulation->res_z;

			int whole_block_col_bytes = y*simulation->res_x*simulation->ft_z_mod + block_col*simulation->res_x;

			p000 = ff + 4*(whole_block_row_bytes + whole_block_col_bytes);

			for (x = 0; x < simulation->res_x-1; x++)
			{
//				p000 = getValuePosition(ff, x, y, z, simulation->res_x, simulation->res_y, simulation->ft_z_mod);
				p100 = p000+4;
				p010 = getValuePosition(ff, x, y+1, z, simulation->res_x, simulation->res_y, simulation->ft_z_mod);
				p110 = p010+4;
				p001 = getValuePosition(ff, x, y, z+1, simulation->res_x, simulation->res_y, simulation->ft_z_mod);
				p101 = p001+4;
				p011 = getValuePosition(ff, x, y+1, z+1, simulation->res_x, simulation->res_y, simulation->ft_z_mod);
				p111 = p011+4;


				float mc_gate = 0.5;

				int mc_index =	(int)(((*(p000)) > mc_gate)) << 0	|
						(int)(((*(p100)) > mc_gate)) << 1	|
						(int)(((*(p010)) > mc_gate)) << 2	|
						(int)(((*(p110)) > mc_gate)) << 3	|
						(int)(((*(p001)) > mc_gate)) << 4	|
						(int)(((*(p101)) > mc_gate)) << 5	|
						(int)(((*(p011)) > mc_gate)) << 6	|
						(int)(((*(p111)) > mc_gate)) << 7	;

				/**
				 * here we create a scale lookup table for the edges of marching cubes according to the fillgrade (given by p***)
				 * because the edges are positioned at 0.5, the scaling has to be within [0;2]
				 *
				 * the scaling of the vertices is made in this way:
				 * the 0.5 component has a coordinate index 'half_index' - e.g. (0.0, 0.5, 1.0) has the half_index 1
				 * by knowing that, the lookup is [half_index][first non 0.5 component][second non 0.5 component] for the new position of the 0.5 component
				 */

				if (mc_index != 0 && mc_index != 0xff)
				{
					px = (float)x / (float)simulation->res_x - 0.5 + 0.5/(float)simulation->res_z;

					glPushMatrix();
					glTranslatef(px, py, pz);
					glScalef(inv_x, inv_y, inv_z);
	//				glCallList(displayListBox);
					glBegin(GL_TRIANGLES);

					for (int p = 0; p < MC_MAX_POLYGON_COUNT; p++)	// polygons
					{
						if (marching_cubes[mc_index][p][0][0] == -111.0)
							break;

#if DISPLACE_MARCHING_CUBES
						vec3f vertices[3];
						for (int i = 0; i < 3; i++)
						{
							float *orig_vertex = marching_cubes[mc_index][p][i+1];

							int half_index;
							if (orig_vertex[0] == 0.5)		half_index = 0;
							else if (orig_vertex[1] == 0.5)	half_index = 1;
							else if (orig_vertex[2] == 0.5)	half_index = 2;
							else
							{
								cerr << "NO HALF INDEX FOUND" << endl;
								break;
							}

							switch(half_index)
							{
								case 0:	// x displacement
									if (orig_vertex[1] == 0.0)
									{
										if (orig_vertex[2] == 0.0)	vertices[i][0] = -(mc_gate-*p000)/(*p000 - *p100);
										else				vertices[i][0] = -(mc_gate-*p001)/(*p001 - *p101);
									}
									else
									{
										if (orig_vertex[2] == 0.0)	vertices[i][0] = -(mc_gate-*p010)/(*p010 - *p110);
										else				vertices[i][0] = -(mc_gate-*p011)/(*p011 - *p111);
									}

									vertices[i][1] = orig_vertex[1];
									vertices[i][2] = orig_vertex[2];
									break;

								case 1:	// y displacement
									if (orig_vertex[0] == 0.0)
									{
										if (orig_vertex[2] == 0.0)	vertices[i][1] = -(mc_gate-*p000)/(*p000 - *p010);
										else				vertices[i][1] = -(mc_gate-*p001)/(*p001 - *p011);
									}
									else
									{
										if (orig_vertex[2] == 0.0)	vertices[i][1] = -(mc_gate-*p100)/(*p100 - *p110);
										else				vertices[i][1] = -(mc_gate-*p101)/(*p101 - *p111);
									}

									vertices[i][0] = orig_vertex[0];
									vertices[i][2] = orig_vertex[2];
									break;

								case 2:	// z displacement
									if (orig_vertex[0] == 0.0)
									{
										if (orig_vertex[1] == 0.0)	vertices[i][2] = -(mc_gate-*p000)/(*p000 - *p001);
										else				vertices[i][2] = -(mc_gate-*p010)/(*p010 - *p011);
									}
									else
									{
										if (orig_vertex[1] == 0.0)	vertices[i][2] = -(mc_gate-*p100)/(*p100 - *p101);
										else				vertices[i][2] = -(mc_gate-*p110)/(*p110 - *p111);
									}

									vertices[i][0] = orig_vertex[0];
									vertices[i][1] = orig_vertex[1];
									break;
									
							}
						}

						vec3f vec0 = vertices[2]-vertices[0];
						vec3f vec1 = vertices[1]-vertices[0];

						vec3f normal = (vec0).getCrossProd(vec1).getNormalized();
						glNormal3fv(normal.data);
						glVertex3fv(vertices[0].data);
						glVertex3fv(vertices[1].data);
						glVertex3fv(vertices[2].data);
#else
						glNormal3fv(marching_cubes[mc_index][p][0]);
						glVertex3fv(marching_cubes[mc_index][p][1]);
						glVertex3fv(marching_cubes[mc_index][p][2]);
						glVertex3fv(marching_cubes[mc_index][p][3]);
#endif
					}
					glEnd();

					glPopMatrix();
				}
				p000+=4;
			}
		}
	}
	glDisable(GL_NORMALIZE);
	glPopMatrix();

//exit(0);

	if (render_wireframe)
	{
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_CULL_FACE);
	}

	delete ff_handler;
}


void CGuiSimulation::renderVolume()
{
	CTexture *t = simulation->fluid_fraction_texture;
	float *ff = new float[t->width*t->height*4];
	t->getData(ff);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glColor3f(0.2, 0.2, 1);	// water should be blue :)
	glEnable(GL_NORMALIZE);

	int x, y, z;
	float px, py, pz;
	float *p;

	float inv_x = 0.5f/(float)simulation->fres[0];
	float inv_y = 0.5f/(float)simulation->fres[1];
	float inv_z = 0.5f/(float)simulation->fres[2];

	for (z = 1; z < simulation->res_z-1; z++)
	{
		int block_row = z / simulation->ft_z_mod;
		int block_col = z % simulation->ft_z_mod;

		int whole_block_row_bytes = block_row * simulation->res_x*simulation->res_y*simulation->ft_z_mod;

		pz = (float)z / (float)simulation->res_z - 0.5 + 0.5/(float)simulation->res_z;
//		for (y = 0; y < simulation->res_y; y++)
		for (y = 1; y < simulation->res_y-1; y++)
		{
			py = (float)y / (float)simulation->res_y - 0.5 + 0.5/(float)simulation->res_z;

			int whole_block_col_bytes = y*simulation->res_x*simulation->ft_z_mod + block_col*simulation->res_x;

			p = ff + 4*(whole_block_row_bytes + whole_block_col_bytes) + 4;

//			for (x = 0; x < simulation->res_x; x++)
			for (x = 1; x < simulation->res_x-1; x++)
			{
//				if (abs(*p) < 10.0f)
//				if (!	((*p) >= 100.0f || (*p)<= -100))
//				if (p[0] >= 0.01 && p[0] <= 0.99)
				if (p[1] == 0)
				{
					px = (float)x / (float)simulation->res_x - 0.5 + 0.5/(float)simulation->res_z;

					glPushMatrix();
					glTranslatef(px, py, pz);
#if 0
					glScalef(	inv_x-0.001,
							inv_y-0.001,
							inv_z-0.001
						);
#else
					glScalef(inv_x, inv_y, inv_z);
#endif
					glCallList(displayListBox);
					glPopMatrix();
				}

				p+=4;
			}
		}
	}

	glDisable(GL_NORMALIZE);
	glPopMatrix();

	delete ff;
}



CGuiSimulation::CGuiSimulation(	CGui *p_gui,
								CSimulation *p_simulation,
								bool p_run_simulation,
								bool p_take_screenshots)
			:
	visible_trackball(false),
	enable_light(true),	render_wireframe(false),
	print_statistics(true),
	quit_program(false),
	render_volume(false),
	render_marching_cubes(true),
	render_single_marching_cube(false),
	single_marching_cube_id(1),
	render_particles(false),
#if HAVE_LIBCEGUIBASE
	mouseInGui(false),
#endif
	guiVisible(false),
	take_screenshots(p_take_screenshots),
	run_simulation(p_run_simulation)
{
	float max_res;

	error = false;

	gui = p_gui;
	simulation = p_simulation;

	font = new CFont();
	font->fontSize(16);

	last_fps_ticks = gui->get_ticks();	// ticks on last fps change
	last_fps = 0;			// last fps number
	fps_frame_count = 0;		// counting frames for next fps display

	gui->callback_class = (void*)this;

	gui->callback_key_down = (void(*)(void*,int, int,int, int))&_gui_callback_key_down;
	gui->callback_key_up = (void(*)(void*,int, int,int, int))&_gui_callback_key_up;
	gui->callback_quit = (void(*)(void*))&_gui_callback_quit;

	gui->callback_mouse_motion = (void(*)(void*,int,int))&_gui_callback_mouse_motion;
	gui->callback_mouse_button_up = (void(*)(void*,int))&_gui_callback_mouse_button_up;
	gui->callback_mouse_button_down = (void(*)(void*,int))&_gui_callback_mouse_button_down;

	gui->callback_viewport_changed = (void(*)(void*, int width, int height))&_gui_callback_viewport_changed;

	for (int i = 0; i < 10; i++)	mouse_button[i] = false;

	view_zoom = -1.1;	// no zoom at the beginning
	view_frustum_size = powf(M_E, view_zoom);

	old_mouse_x = -1;
	old_mouse_y = -1;

	render_data_texture = 0;

	simulation_steps = 0;
	steps_to_do = 0;

	particle_texture = new CTexture(PARTICLE_TEXTURE_WIDTH, PARTICLE_TEXTURE_HEIGHT, GL_RGBA32F_ARB, GL_RGBA, GL_FLOAT, GL_TEXTURE_RECTANGLE_ARB);
	particle_buf_texture = new CTexture(PARTICLE_TEXTURE_WIDTH, PARTICLE_TEXTURE_HEIGHT, GL_RGBA32F_ARB, GL_RGBA, GL_FLOAT, GL_TEXTURE_RECTANGLE_ARB);
	if (!particle_texture->valid || !particle_buf_texture->valid)
	{
		cerr << "failed to create texture for particle emission" << endl;
		goto error_particle_textures;
	}
	particle_fbo = new CFbo(particle_texture);
	particle_fbo->bind();
	particle_fbo->bindFramebufferTexture(particle_texture);
	particle_fbo->checkStatus();
	particle_fbo->unbind();
checkGlError();

	particle_buf_fbo = new CFbo();
	particle_buf_fbo->bind();
	particle_buf_fbo->bindFramebufferTexture(particle_buf_texture);
	particle_buf_fbo->unbind();
checkGlError();

	// load particle emission program
	particle_emission_program = new CProgram("shader/particle_emission_program.frag", NULL);
	if (particle_emission_program->error)
	{
		cerr << "ERROR after creating particle_emission_program" << endl;
		cerr << particle_emission_program->errorLog << endl;
		goto error_particle_emission_program;
	}

	particle_emission_program->enable();
	particle_emission_program->setUniform1i("TextureParticles", 0);
	particle_emission_program->setUniform1i("TextureVelocities", 1);
	particle_emission_program->setUniform2f("p_texture_size", particle_texture->width, particle_texture->height);
	particle_emission_program->setUniform3f("res", &(simulation->fres[0]));
	particle_emission_program->setUniform3f("res_sub_1", simulation->fres[0]-1.0f, simulation->fres[1]-1.0f, simulation->fres[2]-1.0f);
	particle_emission_program->setUniform1f("ft_z_mod", simulation->ft_z_mod);
	particle_emission_program->setUniform1f("emission_position", 0.5f);
	particle_emission_program->disable();

	// multiply program
	multiply_program = new CProgram("shader/multiply_texture.frag", NULL);
	if (multiply_program->error)
	{
		cerr << "ERROR after creating multiply_program" << endl;
		cerr << multiply_program->errorLog << endl;
		goto error_multiply_program;
	}

	multiply_program->enable();
	multiply_program->setUniform1i("Texture0", 0);
	multiply_program->setUniform1f("factor", 1.0f);
	multiply_program->setUniform1f("add", 0.0f);
	multiply_program->setUniform1i("alpha_color", 0);
	multiply_program->disable();

	// create buffer within gpu to store particles
	glGenBuffers(1, &particle_gpu_buffer);
	glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, particle_gpu_buffer);
	glBufferData(GL_PIXEL_PACK_BUFFER_ARB, particle_texture->height*particle_texture->width*sizeof(GLfloat)*4, 0, GL_DYNAMIC_DRAW);
	glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, 0);

	sphereDisplayList = glGenLists(1);
	glNewList(sphereDisplayList, GL_COMPILE);
	glColor3f(0.8, 0.8, 0.8);
	CPrimitives::sphere(0.4, 0, 0, 0, CSIMULATION_SPHERE_LOD);
	glEndList();
checkGlError();

	objFileDisplayList = glGenLists(1);
	glNewList(objFileDisplayList, GL_COMPILE);
	renderObjfile(*(simulation->objFile), false, 0.8, 0.8, 0.8);
	glEndList();
checkGlError();

	max_res = (simulation->fres[0] < simulation->fres[1] ? simulation->fres[1] : simulation->fres[0]);
	max_res = (max_res < simulation->fres[2] ? simulation->fres[2] : max_res);
	scale = simulation->fres / max_res;

	displayListBox = glGenLists(1);
	glNewList(displayListBox, GL_COMPILE);

	glBegin(GL_QUADS);
		glNormal3f(-1, 0, 0);	// left side
			glVertex3f(-1.0, -1.0, -1.0);
			glVertex3f(-1.0, 1.0, -1.0);
			glVertex3f(-1.0, 1.0, 1.0);
			glVertex3f(-1.0, -1.0, 1.0);
		glNormal3f(1, 0, 0);	// right side
			glVertex3f(1.0, -1.0, 1.0);
			glVertex3f(1.0, 1.0, 1.0);
			glVertex3f(1.0, 1.0, -1.0);
			glVertex3f(1.0, -1.0, -1.0);

		glNormal3f(0, -1, 0);	// bottom side
			glVertex3f(-1.0, -1.0, 1.0);
			glVertex3f(1.0, -1.0, 1.0);
			glVertex3f(1.0, -1.0, -1.0);
			glVertex3f(-1.0, -1.0, -1.0);
		glNormal3f(0, 1, 0);	// top side
			glVertex3f(-1.0, 1.0, -1.0);
			glVertex3f(1.0, 1.0, -1.0);
			glVertex3f(1.0,	1.0, 1.0);
			glVertex3f(-1.0, 1.0, 1.0);

		glNormal3f(0, 0, -1);	// front side
			glVertex3f(-1.0, -1.0, -1.0);
			glVertex3f(1.0, -1.0, -1.0);
			glVertex3f(1.0, 1.0, -1.0);
			glVertex3f(-1.0, 1.0, -1.0);
		glNormal3f(0, 0, 1);	// bottom side
			glVertex3f(-1.0, 1.0, 1.0);
			glVertex3f(1.0,	1.0, 1.0);
			glVertex3f(1.0, -1.0, 1.0);
			glVertex3f(-1.0, -1.0, 1.0);
	glEnd();

	glEndList();
checkGlError();

	reset();

#if HAVE_LIBCEGUIBASE
	initCEGui();
#endif

	return;

error_multiply_program:
	delete multiply_program;

error_particle_emission_program:
	delete particle_emission_program;
	delete particle_fbo;
	delete particle_buf_fbo;

error_particle_textures:
	delete particle_texture;
	delete particle_buf_texture;

	delete font;

	gui->callback_class = 0;

	gui->callback_key_down = 0;
	gui->callback_key_up = 0;
	gui->callback_quit = 0;

	gui->callback_mouse_motion = 0;
	gui->callback_mouse_button_up = 0;
	gui->callback_mouse_button_down = 0;

	error = true;
}


/**
 * deconstructor
 */
CGuiSimulation::~CGuiSimulation()
{
	if (error)		return;

	glDeleteLists(sphereDisplayList, 1);
	glDeleteLists(objFileDisplayList, 1);

	delete multiply_program;

	delete particle_emission_program;
	delete particle_fbo;
	delete particle_buf_fbo;

	delete particle_texture;
	delete particle_buf_texture;

	delete font;

	gui->callback_class = 0;

	gui->callback_key_down = 0;
	gui->callback_key_up = 0;
	gui->callback_quit = 0;

	gui->callback_mouse_motion = 0;
	gui->callback_mouse_button_up = 0;
	gui->callback_mouse_button_down = 0;
}



/**
 * reset the whole simulation (within the gui)
 */
void CGuiSimulation::reset()
{
	model_rotate_axis_y = 0;
	model_rotate_axis_x = 0;

	model_translate_x = model_translate_y = model_translate_z = 0;

	// setup modelview matrix for simulation
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glGetFloatv(GL_MODELVIEW_MATRIX, modelview_matrix);
	simulation->modelview_matrix = modelview_matrix;
	glPopMatrix();

	simulation->reset();	// reset simulation
	simulation->modelview_matrix = trackBall.rotationMatrixArray;
	trackBall.reset();
	trackBallGravity.reset();

	trackBall.setParameters(true, 5.0);
	trackBallGravity.setParameters(true, 5.0);

	gravitation_dir = simulation->getGravitation();
	gravitation_length = gravitation_dir.length();
	gravitation_dir.normalize();

// opengl stuff
	fbo_write_save_state();

	simulation->fbo_write_start(particle_texture->width, particle_texture->height);
checkGlError();
	simulation->clear_program->enable();
	simulation->clear_program->setUniform4f("color", 0, 0, 1000, 0);		// set alpha to 0 -> there's no particle
		particle_fbo->bind();
		glCallList(simulation->write_to_texture_call_list_z);	// we can use this calllist because we just paint one color on the framebuffer
		particle_fbo->unbind();
	simulation->clear_program->disable();
	simulation->fbo_write_end();

	fbo_write_restore_state();

checkGlError();

	particle_fbo->bind();
		glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, particle_gpu_buffer);
			glReadPixels(	0, 0,
					particle_texture->width, particle_texture->height,
					GL_RGBA, GL_FLOAT, 0);
		glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, 0);
	particle_fbo->unbind();
checkGlError();

	particle_emission_row = 0;
}


/**
 * particle rendering
 *
 * particles are stored in a non-quadratic texture
 * the width of the texture is the particles which are emitted in every new loop step
 * the height of the texture is the overall particle number divided the width of the texture
 *
 * for every rendering loop, the next row (emitting row) in the texture is used to emit new particles
 * the other particles are given their new location according to the velocity texture
 *
 * non-existing particles are markes with alpha == 0
 * to avoid rendering them, their position is set outside the clipping planes
 */
void CGuiSimulation::renderParticles()
{
	fbo_write_save_state();

	// use buffer of simulation to create velocity field (reverse the application of computation v/rho in the shader)

	if (steps_to_do)
	{
checkGlError();

		// emit particles & interpolate particle positions
		particle_buf_fbo->bind();
		simulation->fbo_write_start(particle_texture->width, particle_texture->height);

			particle_texture->bind();
			glActiveTexture(GL_TEXTURE1);
			simulation->velocity_density_texture->bind();

				particle_emission_program->enable();
				particle_emission_program->setUniform1f("emission_row", (float)particle_emission_row);
					glBegin(GL_QUADS);
						glTexCoord2f(0, 0);					glVertex2f(0, 0);
						glTexCoord2f(particle_texture->width, 0);		glVertex2f(1, 0);
						glTexCoord2f(particle_texture->width, particle_texture->height);	glVertex2f(1, 1);
						glTexCoord2f(0, particle_texture->height);		glVertex2f(0, 1);
					glEnd();
				particle_emission_program->disable();

			simulation->velocity_density_texture->unbind();
			glActiveTexture(GL_TEXTURE0);
			particle_texture->unbind();

		simulation->fbo_write_end();
		particle_buf_fbo->unbind();

		particle_emission_row = (particle_emission_row+1) % particle_texture->height;
	checkGlError();

		// finally swap textures
		simulation->swap(&particle_fbo, &particle_buf_fbo);
		simulation->swap(&particle_texture, &particle_buf_texture);

	checkGlError();

		// copy vertex data to vertex array
		particle_fbo->bind();
			glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, particle_gpu_buffer);
				glReadPixels(	0, 0,
						particle_texture->width, particle_texture->height,
						GL_RGBA, GL_FLOAT, 0);
			glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, 0);
		particle_fbo->unbind();
	}
	glDisable(GL_TEXTURE_2D);

	checkGlError();

	glEnable(GL_DEPTH_TEST);
// draw the vertices
	int a = (gui->screen_width < gui->screen_height ? gui->screen_width : gui->screen_height);
	glPointSize((float)(a) / (400.0f*view_frustum_size));
	glColor3f(1, 0.2, 0.2);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(-0.5, -0.5, -0.5);

	glDisable(GL_BLEND);
	glBindBuffer(GL_ARRAY_BUFFER, particle_gpu_buffer);
		glEnableClientState(GL_VERTEX_ARRAY);
			glVertexPointer(4, GL_FLOAT, sizeof(GLfloat)*4, 0);
			glDrawArrays(GL_POINTS, 0, particle_texture->width*particle_texture->height);
		glDisableClientState(GL_VERTEX_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	glPopMatrix();

	checkGlError();
	fbo_write_restore_state();
}



/**
 * this is the main render loop for the gui
 * the gui simulation includes also the visualisation of the simulation
 *
 * this contains also the particle emitting which is done by the function renderParticles()
 */
bool CGuiSimulation::render()
{
#if GET_SIMULATION_TICKS
	static float percent_voxelisation = 0, ticks_one_voxelisation = 0;
	static float percent_collide = 0, ticks_one_collide = 0;
	static float percent_propagate = 0, ticks_one_propagate = 0;
	static float percent_velocity_density = 0, ticks_one_velocity_density = 0;
	static float percent_update_state = 0, ticks_one_update_state = 0;
	static float percent_render = 0, ticks_one_render = 0;
	static int ticks_render = 0;
	static float round_count = 0;

glFinish();

	round_count++;
#endif

	int act_ticks = gui->get_ticks();
//	static int ticks_reset_count = 0;
	fps_frame_count++;

	if (act_ticks - last_fps_ticks > 1000)	// update if more than 1 second since the last update is over
	{
//		last_fps = (float)(fps_frame_count*1000) / (float)(act_ticks - last_fps_ticks);
		last_fps = fps_frame_count;
		fps_frame_count = 0;

		last_fps_ticks = act_ticks;

#if GET_SIMULATION_TICKS
		float ticks_all		= (float)(	simulation->ticks_voxelisation +
							simulation->ticks_collide + 
							simulation->ticks_propagate + 
							simulation->ticks_velocity_density + 
							simulation->ticks_update_state +
							ticks_render
						);

		percent_voxelisation	= (float)simulation->ticks_voxelisation		/ticks_all;
		percent_collide		= (float)simulation->ticks_collide		/ticks_all;
		percent_propagate	= (float)simulation->ticks_propagate		/ticks_all;
		percent_velocity_density = (float)simulation->ticks_velocity_density	/ticks_all;
		percent_update_state	= (float)simulation->ticks_update_state		/ticks_all;
		percent_render		= (float)ticks_render				/ticks_all;

		ticks_one_voxelisation	= (float)simulation->ticks_voxelisation		/round_count;
		ticks_one_collide	= (float)simulation->ticks_collide		/round_count;
		ticks_one_propagate	= (float)simulation->ticks_propagate		/round_count;
		ticks_one_velocity_density = (float)simulation->ticks_velocity_density	/round_count;
		ticks_one_update_state	= (float)simulation->ticks_update_state		/round_count;
		ticks_one_render	= (float)ticks_render				/round_count;

//		cout << "ticks all: " << ticks_all << endl;
//		cout << percent_render << endl;
//		if (ticks_reset_count > 10)
		{
			simulation->ticks_reset();
			ticks_render = 0;
//			ticks_reset_count = 0;
		}
//		ticks_reset_count++;
		round_count = 0;
#endif
	}

#if GET_SIMULATION_TICKS
	int render_ticks_start = gui->get_ticks();
#endif
	/*
	 * START RENDERING
	 */

	glClearColor(1,1,1,1);
	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-view_frustum_size, 1*view_frustum_size, -view_frustum_size, view_frustum_size, 1, 20);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	/*
	 * go back 2 units
	 * distance of eye to the center
	 */
	glTranslatef(0, 0, -3);


	/*****************************
	 * SETUP LIGHT
	 *****************************/
	if (enable_light)
	{
		const float lightPos[4] = {1, 0.5, 1, 1};
		const float lightAmbient[4] = {0.2, 0.2, 0.2, 1};
		const float lightDiffuse[4] = {0.8, 0.8, 0.8, 1};
		const float lightSpecular[4] = {0.5, 0.5, 0.5, 1};

		glDisable(GL_LIGHTING);

		// mark the light position
		glPushMatrix();
		glTranslatef(lightPos[0], lightPos[1], lightPos[2]);
		glScalef(0.1, 0.1, 0.1);
			glDisable(GL_LIGHTING);
			glDisable(GL_CULL_FACE);
			glColor3f(0.0, 0.7, 1.0);

			glBegin(GL_LINES);
				glVertex3f(-1.0, 0, 0);
				glVertex3f(1.0, 0, 0);
				glVertex3f(0, -1.0, 0);
				glVertex3f(0, 1.0, 0);
				glVertex3f(0, 0, -1.0);
				glVertex3f(0, 0, 1.0);
			glEnd();

			glEnable(GL_LIGHTING);
			glEnable(GL_CULL_FACE);
		glPopMatrix();

		glEnable(GL_COLOR_MATERIAL);
	        glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

		glEnable(GL_LIGHT0);
		glEnable(GL_LIGHTING);


		glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
		glLightfv(GL_LIGHT0, GL_AMBIENT, lightAmbient);
		glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);
		glLightfv(GL_LIGHT0, GL_SPECULAR, lightSpecular);

		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, lightAmbient);
		glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lightDiffuse);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, lightSpecular);

		glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, 100);

	}
	else
	{
		glDisable(GL_LIGHTING);
		glDisable(GL_LIGHT0);
	}

	/*
	 * standard view rotates the eye around the object
	 */
	glMultMatrixf(trackBall.rotationMatrixArray);

	/**
	 * mark the CUBE ORIGIN position
	 */
	glPushMatrix();
	glTranslatef(-0.55, -0.55, -0.55);
	glScalef(0.2, 0.2, 0.2);
		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);

		CPrimitives::originCross();
		
		glEnable(GL_LIGHTING);
		glEnable(GL_CULL_FACE);
	glPopMatrix();


	/**
	 * mark the gravitation direction
	 */
	glPushMatrix();
	glTranslatef(0, -0.55, 0);
	glScalef(0.2, 0.2, 0.2);

		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);

		glColor3f(0, 0, 0);
		glBegin(GL_LINES);
			glVertex3f(0, 0, 0);
			glVertex3f(gravitation_dir[0], gravitation_dir[1], gravitation_dir[2]);

			glVertex3f(-0.1, 0, 0);
			glVertex3f(+0.1, 0, 0);
			glVertex3f(0, -0.1, 0);
			glVertex3f(0, +0.1, 0);
			glVertex3f(0, 0, -0.1);
			glVertex3f(0, 0, +0.1);
		glEnd();


		glScalef(-1,-1,-1);
/*
		// rotate back (see below)
		glRotatef(-90, 0, 1, 0);
*/
		/*
		glMultMatrixf((GLfloat*)modelBall.rotationMatrix);

		// rotate to neutral z-directed start direction
		glRotatef(90, 0, 1, 0);
		CPrimitives::arrow();
		*/
		
		glEnable(GL_LIGHTING);
		glEnable(GL_CULL_FACE);
	glPopMatrix();

	/***************************
	 * RENDER SIMULATION
	 **************************/

	glScalef(scale[0], scale[1], scale[2]);

	glDisable(GL_TEXTURE_2D);

		// draw world border
		glColor3f(0.5, 0.5, 0.5);
		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);

		CPrimitives::box();

		glEnable(GL_LIGHTING);
		glEnable(GL_CULL_FACE);

	glColor3f(1,1,1);

checkGlError();
	if (render_particles)
		renderParticles();

checkGlError();
	if (render_volume)
		renderVolume();

	if (render_marching_cubes)
		renderMarchingCubes();



	/************************
	 * MODEL
	 ************************/
checkGlError();
	glMultMatrixf(modelview_matrix);

// draw object
	glEnable(GL_CULL_FACE);
	glFrontFace(GL_CCW);

	if (render_wireframe)
	{
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glDisable(GL_CULL_FACE);
	}

	/*
	switch(simulation->fluidObstacleType)
	{
		case 0:
			glCallList(objFileDisplayList);
			break;

		case 1:
			glColor3f(0, 0, 1);
			glCallList(sphereDisplayList);
			break;
	}
	*/

	if (render_wireframe)
	{
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}

	glFrontFace(GL_CW);
	glEnable(GL_CULL_FACE);


	/****************************
	 * TRACKBALL
	 ****************************/
	
	if (visible_trackball)
	{
		trackBall.render();
//		trackBallGravity.render();
	}

	/****************************
	 * INFORMATION OUTPUT
	 ****************************/
checkGlError();

	if (print_statistics)
	{
		// print statistics
		glColor3f(0,0,0);

		int pos = 5;

		font->printfxy(5, pos, "------------------------", last_fps);	pos += 16;
		font->printfxy(5, pos, "             res: (%i, %i, %i)", simulation->res_x, simulation->res_y, simulation->res_z);	pos += 16;
		font->printfxy(5, pos, "simulation steps: %i / %i\n", simulation_steps, simulation_steps + steps_to_do);	pos += 16;
		font->printfxy(5, pos, "------------------------", last_fps);	pos += 16;
		font->printfxy(5, pos, "           fps: %.2f", last_fps);	pos += 16;
		font->printfxy(5, pos, "------------------------", last_fps);	pos += 16;
#if GET_SIMULATION_TICKS
		font->printfxy(5, pos, "             percent / ticks per round / ticks total");	pos += 16;
		font->printfxy(5, pos, "  voxelisation: %.3f   %.3f   %i", percent_voxelisation,	ticks_one_voxelisation,	simulation->ticks_voxelisation);	pos += 16;
		font->printfxy(5, pos, "       collide: %.3f   %.3f   %i", percent_collide,		ticks_one_collide,	simulation->ticks_collide);	pos += 16;
		font->printfxy(5, pos, "   propagation: %.3f   %.3f   %i", percent_propagate,		ticks_one_propagate,	simulation->ticks_propagate);	pos += 16;
		font->printfxy(5, pos, "vel. and dens.: %.3f   %.3f   %i", percent_velocity_density,	ticks_one_velocity_density,	simulation->ticks_velocity_density);	pos += 16;
		font->printfxy(5, pos, "    cell state: %.3f   %.3f   %i", percent_update_state,	ticks_one_update_state,	simulation->ticks_update_state);	pos += 16;
		font->printfxy(5, pos, "        render: %.3f   %.3f   %i", percent_render,		ticks_one_render,	ticks_render);		pos += 16;
		font->printfxy(5, pos, "------------------------", last_fps);	pos += 16;
#endif
		pos += 16;
//		if (!guiVisible)
//			font->printfxy(5, pos, "press [space] to show gui");	pos += 16;
//		pos += 16;
	}

	if (render_single_marching_cube)
		renderMarchingCube(single_marching_cube_id);

	if (render_data_texture)
	{
		int texture_width, texture_height;
		CTexture *texture_to_render;
		const char *texture_description;
		char some_buffer[100];

		float render_factor = 1.0f;
		float render_add = 0.0f;
		int render_alpha_color = 0;
		switch(render_data_texture)
		{
		// VOXELISATION TEXTURES
			case 1:
				texture_to_render = simulation->voxel_texture;
				texture_description = "voxelisation texture - combined";
				break;

			case 2:
				texture_to_render = simulation->voxel_buffer_texture[2];
				texture_description = "voxelisation texture - Z Slices";
				break;

			case 3:
				texture_to_render = simulation->voxel_buffer_texture[1];
				texture_description = "voxelisation texture - Y Slices";
				break;

			case 4:
				texture_to_render = simulation->voxel_texture;
				texture_to_render = simulation->voxel_buffer_texture[0];
				texture_description = "voxelisation texture - X Slices";
				break;

		// VELOCITY TEXTURE
			case 5:
				texture_to_render = simulation->velocity_density_texture;
				texture_description = "velocity and density texture";
//				render_add = 0.5f;
				render_factor = 10;
				break;


		// DISTRIBUTION TEXTURES
			case 6:
			case 7:
			case 8:
			case 9:
			case 10:
				texture_to_render = simulation->distributions_texture[render_data_texture-6];
				render_factor = 10.0f;
				sprintf(some_buffer, "distribution texture # %i", render_data_texture-6);
				texture_description = some_buffer;
				break;

			case 11:
				texture_to_render = particle_texture;
				texture_description = "particle_texture";
				break;

			case 12:
				texture_to_render = particle_buf_texture;
				texture_description = "particle_buf_texture";
				break;

			case 13:
				texture_to_render = simulation->fluid_fraction_texture;
				texture_description = "fluid_fraction_texture - R: interface, G: obstacle, B: fluid, BLACK: gas, WHITE: error";
				render_alpha_color = 1;
				break;

			case 14:
				texture_to_render = simulation->fluid_mass_texture;
				texture_description = "fluid_mass_texture";
				render_factor = 0.5;
				break;

			default:
				texture_description = "ERROR!";
				texture_to_render = 0;
		}

		float scale = SCALE_TEXTURE;
		texture_width = texture_to_render->width;
		texture_height = texture_to_render->height;

		glColor4f(0,0,0,1);
		char another_buffer[100];
		sprintf(another_buffer, "(%i / %i): %s", render_data_texture, RENDER_DATA_TEXTURE_MAX, texture_description);
		font->printfxy(3, gui->screen_height - (int)(texture_height*scale) - 16-5-15, "%s", another_buffer);
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, gui->screen_width, 0, gui->screen_height, -1, 1);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		glDisable(GL_TEXTURE_2D);

		float tw = texture_width*scale;
		float th = texture_height*scale;

		glColor3f(1,0,0);
		glBegin(GL_QUADS);
			glVertex2f(0, 0);
			glVertex2f(tw+2, 0);
			glVertex2f(tw+2, th+2);
			glVertex2f(0, th+2);
		glEnd();
		texture_to_render->unbind();

		glColor3f(1,1,1);
		texture_to_render->bind();

		multiply_program->enable();
		multiply_program->setUniform1f("factor", render_factor);
		multiply_program->setUniform1f("add", render_add);
		multiply_program->setUniform1i("alpha_color", render_alpha_color);

		glDisable(GL_BLEND);
			glBegin(GL_QUADS);
				glTexCoord2f(0,0);	glVertex2f(1, 1);
				glTexCoord2f(texture_width,0);	glVertex2f(tw+1, 1);
				glTexCoord2f(texture_width,texture_height);	glVertex2f(tw+1, th+1);
				glTexCoord2f(0,texture_height);	glVertex2f(1, th+1);
			glEnd();
		multiply_program->disable();

		texture_to_render->unbind();

		glEnable(GL_DEPTH_TEST);

		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
	}

#if HAVE_LIBCEGUIBASE
// CEGUI Stuff
	static double last_time_pulse = 0;
	double t = 0.001*SDL_GetTicks();

	// inject the time that passed since the last call 
	CEGUI::System::getSingleton().injectTimePulse( float(t-last_time_pulse) );

	// store the new time as the last time
	last_time_pulse = t;

	// render the GUI :)
	if (guiVisible)
		CEGUI::System::getSingleton().renderGUI();
#endif

	if (run_simulation)	// always run simulation
		steps_to_do++;

	if (steps_to_do && take_screenshots)
	{
		char buffer[1024];
		sprintf(buffer, "screenshots/%08i.bmp", simulation_steps);
		string filename = buffer;
		gui->takeScreenshot(filename);
	}

#if GET_SIMULATION_TICKS
glFinish();
	int render_ticks_end = gui->get_ticks();
	ticks_render += render_ticks_end - render_ticks_start;
#endif

	return quit_program;
}


/**************
 * CALLBACKS
 **************/

void CGuiSimulation::_gui_callback_quit(CGuiSimulation *sim)
{
	sim->gui_callback_quit();
}

void CGuiSimulation::_gui_callback_key_down(CGuiSimulation *sim, int key, int mod, int scancode, int unicode)
{
	sim->gui_callback_key_down(key, mod, scancode, unicode);
}

void CGuiSimulation::_gui_callback_key_up(CGuiSimulation *sim, int key, int mod, int scancode, int unicode)
{
	sim->gui_callback_key_up(key, mod, scancode, unicode);
}

void CGuiSimulation::_gui_callback_mouse_motion(CGuiSimulation *sim, int x, int y)
{
	sim->gui_callback_mouse_motion(x, y);
}

void CGuiSimulation::_gui_callback_mouse_button_up(CGuiSimulation *sim, int button)
{
	sim->gui_callback_mouse_button_up(button);
}

void CGuiSimulation::_gui_callback_mouse_button_down(CGuiSimulation *sim, int button)
{
	sim->gui_callback_mouse_button_down(button);
}

void CGuiSimulation::_gui_callback_viewport_changed(CGuiSimulation *sim, int width, int height)
{
	sim->gui_callback_viewport_changed(width, height);
}






void CGuiSimulation::gui_callback_key_up(int key, int mod, int scancode, int unicode)
{
#if HAVE_LIBCEGUIBASE
	if (guiVisible)
	{
		CEGUI::System::getSingleton().injectKeyUp(scancode);
	}
#endif
}



void CGuiSimulation::gui_callback_mouse_motion(int x, int y)
{

#if HAVE_LIBCEGUIBASE
	if (guiVisible)
	        CEGUI::System::getSingleton().injectMousePosition((float)x, (float)y);
#endif

	if (mouse_button[1])	// left mouse button
	{	
		trackBall.dragMove(x, y);
	}

	if (mouse_button[3])	// right mouse button
	{
		trackBallGravity.dragMove(x, y);

		CVector<3,float> gravitation_dir_std(0, -1, 0);
		gravitation_dir = trackBallGravity.rotationMatrixCurrent*gravitation_dir_std;
		gravitation_dir.normalize();

		CVector<3,float> new_gravitation_dir = gravitation_dir*gravitation_length;

		simulation->setGravitation(new_gravitation_dir);

	}

	old_mouse_x = x;
	old_mouse_y = y;
}

void CGuiSimulation::gui_callback_mouse_button_down(int button)
{
	
#if HAVE_LIBCEGUIBASE
	// CEGUI STUFF
	if (guiVisible)
	switch(button)
	{
		case SDL_BUTTON_LEFT:
			CEGUI::System::getSingleton().injectMouseButtonDown(CEGUI::LeftButton);
			break;

		case SDL_BUTTON_MIDDLE:
			CEGUI::System::getSingleton().injectMouseButtonDown(CEGUI::MiddleButton);
			break;

		case SDL_BUTTON_RIGHT:
			CEGUI::System::getSingleton().injectMouseButtonDown(CEGUI::RightButton);
			break;

		case SDL_BUTTON_WHEELDOWN:
			CEGUI::System::getSingleton().injectMouseWheelChange( -1 );
			break;

		case SDL_BUTTON_WHEELUP:
			CEGUI::System::getSingleton().injectMouseWheelChange( +1 );
			break;
	}

	// IGNORE MOUSE BUTTON DOWN IF MOUSE IS IN GUI
	if (!myControlWindow->isHit(CEGUI::Vector2(old_mouse_x, old_mouse_y)) || !guiVisible)
	{
#endif
		if (button >= 10)	return;
		mouse_button[button] = true;

		switch(button)
		{
			case 1:	// left mouse button
				trackBall.dragStart(old_mouse_x, old_mouse_y);
				break;
			
			case 3:	// right mouse button
				trackBallGravity.dragStart(old_mouse_x, old_mouse_y);
				break;

			case 4:
				view_zoom -= 0.1;
				view_frustum_size = powf(M_E, view_zoom);
				break;

			case 5:
				view_zoom += 0.1;
				view_frustum_size = powf(M_E, view_zoom);
				break;
		}
#if HAVE_LIBCEGUIBASE
	}
#endif
}

void CGuiSimulation::gui_callback_mouse_button_up(int button)
{
#if HAVE_LIBCEGUIBASE
	if (guiVisible)
	switch(button)
	{
		case SDL_BUTTON_LEFT:
			CEGUI::System::getSingleton().injectMouseButtonUp(CEGUI::LeftButton);
			break;

		case SDL_BUTTON_MIDDLE:
			CEGUI::System::getSingleton().injectMouseButtonUp(CEGUI::MiddleButton);
			break;

		case SDL_BUTTON_RIGHT:
			CEGUI::System::getSingleton().injectMouseButtonUp(CEGUI::RightButton);
			break;
	}
#endif
	if (button >= 10)	return;
	mouse_button[button] = false;

	switch(button)
	{
		case 1:	// left mouse button
			trackBall.dragEnd();
			break;

		case 3:	// right mouse button
			trackBallGravity.dragEnd();
			break;
	}		
}




/*********************************************************
 * CEGUI STUFF
 *********************************************************/

#if HAVE_LIBCEGUIBASE
bool CGuiSimulation::renderBunnyClicked(const CEGUI::EventArgs&)
{
	simulation->fluidObstacleType = 0;
	return true;
}

bool CGuiSimulation::renderSphereClicked(const CEGUI::EventArgs&)
{
	simulation->fluidObstacleType = 1;
	return true;
}

bool CGuiSimulation::renderNothingClicked(const CEGUI::EventArgs&)
{
	simulation->fluidObstacleType = 2;
	return true;
}

bool CGuiSimulation::renderWireframe(const CEGUI::EventArgs& a)
{
	render_wireframe = myCheckboxWireframe->isSelected();
	return true;
}


bool CGuiSimulation::sliderEmissionPosition(const CEGUI::EventArgs& a)
{
	particle_emission_program->enable();
	particle_emission_program->setUniform1f("emission_position", mySliderEmissionPosition->getCurrentValue());
	particle_emission_program->disable();

	return true;
}

bool CGuiSimulation::toggleRunSimulation(const CEGUI::EventArgs& a)
{
	run_simulation = myRunSimulation->isSelected();
	return true;
}

bool CGuiSimulation::exitClicked(const CEGUI::EventArgs& a)
{
	quit_program = true;
	return true;
}

bool CGuiSimulation::resetClicked(const CEGUI::EventArgs& a)
{
	reset();
	return true;
}

void CGuiSimulation::reloadCEGui()
{
	using namespace CEGUI;

	try{
		WindowManager::getSingleton().destroyWindow(String("LBM"));
	}
	catch(...){}

	try{
		Window* myRoot = WindowManager::getSingleton().loadWindowLayout("lbm.layout");
		System::getSingleton().setGUISheet(myRoot);

		myControlWindow = myRoot->getChild("LBM/ControlWindow");
		setCEGUIWindowStyle();

		Window *w;
		w = myControlWindow->getChild("LBM/ControlWindow/RadioRenderBunny");
		w->subscribeEvent(	CEGUI::RadioButton::EventMouseClick, CEGUI::Event::Subscriber(	&CGuiSimulation::renderBunnyClicked, this	)	);

		w = myControlWindow->getChild("LBM/ControlWindow/RadioRenderSphere");
		w->subscribeEvent(	CEGUI::RadioButton::EventMouseClick, CEGUI::Event::Subscriber(	&CGuiSimulation::renderSphereClicked, this	)	);

		w = myControlWindow->getChild("LBM/ControlWindow/RadioRenderNothing");
		w->subscribeEvent(	CEGUI::RadioButton::EventMouseClick, CEGUI::Event::Subscriber(	&CGuiSimulation::renderNothingClicked, this	)	);

		myCheckboxWireframe = (Checkbox*)myControlWindow->getChild("LBM/ControlWindow/CheckboxWireframeMode");
		myCheckboxWireframe->subscribeEvent(	CEGUI::Checkbox::EventCheckStateChanged, CEGUI::Event::Subscriber(	&CGuiSimulation::renderWireframe, this	)	);
		mySliderEmissionPosition = (Slider*)myControlWindow->getChild("LBM/ControlWindow/SliderEmission");
		mySliderEmissionPosition->subscribeEvent(	CEGUI::Slider::EventValueChanged, CEGUI::Event::Subscriber(	&CGuiSimulation::sliderEmissionPosition, this	)	);

		myRunSimulation = (Checkbox*)myControlWindow->getChild("LBM/ControlWindow/CheckboxRunSimulation");
		myRunSimulation->subscribeEvent(	CEGUI::Checkbox::EventCheckStateChanged, CEGUI::Event::Subscriber(	&CGuiSimulation::toggleRunSimulation, this	)	);

		w = myControlWindow->getChild("LBM/ControlWindow/ButtonExit");
		w->subscribeEvent(	CEGUI::PushButton::EventClicked, CEGUI::Event::Subscriber(	&CGuiSimulation::exitClicked, this	)	);

		w = myControlWindow->getChild("LBM/ControlWindow/ButtonReset");
		w->subscribeEvent(	CEGUI::PushButton::EventClicked, CEGUI::Event::Subscriber(	&CGuiSimulation::resetClicked, this	)	);
	}
	catch (CEGUI::Exception& exc)
	{
		cerr << exc.getMessage().c_str() << endl;
	}
	catch (std::exception& exc)
	{
		cerr << exc.what() << endl;
	}
	catch(...)
	{
		cout << "unknown exception catched" << endl;
	}


}
#endif
