/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CGUISIMULATION_HH
#define CGUISIMULATION_HH

#include "CGui.hpp"
#include "CSimulation.hpp"
#include "libviz/CFont.hpp"
#include "libviz/CTrackBall.hh"

#if HAVE_LIBCEGUIBASE
	#include <CEGUI.h>
	#include <RendererModules/OpenGLGUIRenderer/openglrenderer.h>
#endif



class CGuiSimulation
{
	CGui *gui;
	CSimulation *simulation;

	CFont *font;

	CVector<3,float> scale;		// scaling of the scene if resolution is different at an axis

	CTrackBall<float> trackBall;	// trackball for rotation of fluid cube
	bool visible_trackball;		// toggle visibility of trackball

	bool mouse_button[10];		// which mouse button is down
	int old_mouse_x, old_mouse_y;		// old mouse position

	CTrackBall<float> trackBallGravity;
	float gravitation_length;
	CVector<3,float> gravitation_dir;

	float model_translate_x, model_translate_y, model_translate_z;
	float view_zoom;			// zooming
	float view_frustum_size;		// factor of which frustum is made widther

	// MODIFICATION OF MODEL MATRIX
	float model_rotate_axis_x, model_rotate_axis_y;	// object rotation

	bool enable_light;
	bool render_wireframe;
	bool print_statistics;
	bool quit_program;
	bool render_volume;
	bool render_marching_cubes;
	bool render_single_marching_cube;
	int single_marching_cube_id;
	bool render_particles;

	// fps statistics
	int last_fps_ticks;		// ticks on last fps change
	int fps_frame_count;	// counting frames for next fps display
	float last_fps;		// last fps number

	int render_data_texture;	// render some data texture (debug purposes)

	float modelview_matrix[16];	// modelview matrix to modify position of 3d object within simulation

	void gui_callback_quit();
	void gui_callback_key_down(int key, int mod, int scancode, int unicode);
	void gui_callback_key_up(int key, int mod, int scancode, int unicode);
	void gui_callback_mouse_motion(int x, int y);
	void gui_callback_mouse_button_down(int button);
	void gui_callback_mouse_button_up(int button);
	void gui_callback_viewport_changed(int width, int height);

	// GUI
	GLuint objFileDisplayList;	///< Displaylist to render Objectfile
	GLuint sphereDisplayList;	///< Displaylist to render wired model

	/**
	 * particle stuff
	 */
	CTexture *particle_texture;	CFbo *particle_fbo;	// storage for particle emission
	CTexture *particle_buf_texture;	CFbo *particle_buf_fbo;	// buffer for particle emission
	CProgram *particle_vector_program, *particle_emission_program;	// the programs
	CProgram *multiply_program;	// to enlight the texture
	GLuint particle_gpu_buffer;	// buffer within gpu
	int particle_emission_row;	// in which row should particles be emitted

	void renderParticles();
	void reset();

	void renderVolume();	// render fluid volume by fluid fraction texture
	void renderMarchingCubes();	// render fluid volume by using marching cubes (isosurface)
	void renderMarchingCube(int mc_index);	// render ONE marching cube surface given by mc_index
	GLuint displayListBox;	// box to render fluid volume

	// CE GUI
	void initCEGui();		// init crazy eddies gui
	void reloadCEGui();		// reload GUI and load GUI at the first time

#if HAVE_LIBCEGUIBASE
	bool renderBunnyClicked(const CEGUI::EventArgs&);
	bool renderSphereClicked(const CEGUI::EventArgs&);
	bool renderNothingClicked(const CEGUI::EventArgs&);
	bool renderWireframe(const CEGUI::EventArgs&);
	bool sliderEmissionPosition(const CEGUI::EventArgs&);
	bool toggleRunSimulation(const CEGUI::EventArgs&);
	bool exitClicked(const CEGUI::EventArgs&);
	bool resetClicked(const CEGUI::EventArgs&);

	bool mouseInGui;


	CEGUI::Window *myControlWindow;
	CEGUI::Checkbox *myCheckboxWireframe;
	CEGUI::Slider *mySliderEmissionPosition;
	CEGUI::Checkbox *myRunSimulation;
	CEGUI::OpenGLRenderer *myGuiRenderer;
#endif
	bool guiVisible;

	bool take_screenshots;
	bool run_simulation;	// run simulation (always increment simulation counter)

public:
	bool error;			// true if error happens
	int simulation_steps;		// count of steps already done
	int steps_to_do;		// count of steps to perform

	CGuiSimulation(CGui *p_gui, CSimulation *p_simulation, bool p_run_simulation = true, bool p_take_screenshots = false);
	~CGuiSimulation();

	static void _gui_callback_quit(CGuiSimulation *sim);
	static void _gui_callback_key_down(CGuiSimulation *sim, int key, int mod, int scancode, int unicode);
	static void _gui_callback_key_up(CGuiSimulation *sim, int key, int mod, int scancode, int unicode);
	static void _gui_callback_mouse_motion(CGuiSimulation *sim, int x, int y);
	static void _gui_callback_mouse_button_up(CGuiSimulation *sim, int button);
	static void _gui_callback_mouse_button_down(CGuiSimulation *sim, int button);

	static void _gui_callback_viewport_changed(CGuiSimulation *sim, int width, int height);

	bool render();

};

#endif
