/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "CSimulation.hpp"
#include <iostream>
#include "libviz/gltools.hpp"
#include "libviz/dspObjfile.hpp"
#include <math.h>
#include "libviz/CPrimitives.hpp"
#include "../shader/config.inc"
#include <fstream>


/**
 * reset the whole simulation to the initially state
 */
void CSimulation::reset()
{
	/*
	 * setup for writing to textures
	 */
	fbo_write_save_state();
	fbo_write_start(ft_z_width, ft_z_height);

checkGlError();

	/*
	 * use the clear program because the glColor() seems to truncate the float variables
	 */
	clear_program->enable();

	/**
	 * reset the DENSITY DISTRIBUTIONS to equilibrium values
	 * the density distributions are set to the solution of the according equilibrium distribution
	 * function with zero velocities
	 */

	clear_program->setUniform4f("color", 1.0f/18.0f, 1.0f/18.0f, 1.0f/18.0f, 1.0f/18.0f);
	distributions_fbo[0]->bind();
	renderComputationRectangleZ();

	clear_program->setUniform4f("color", 1.0f/36.0f, 1.0f/36.0f, 1.0f/36.0f, 1.0f/36.0f);
	distributions_fbo[1]->bind();
	renderComputationRectangleZ();
	distributions_fbo[2]->bind();
	renderComputationRectangleZ();
	distributions_fbo[3]->bind();
	renderComputationRectangleZ();

	clear_program->setUniform4f("color", 1.0f/18.0f, 1.0f/18.0f, 1.0f/3.0f, 0.0f);
	distributions_fbo[4]->bind();
	renderComputationRectangleZ();
	distributions_fbo[4]->unbind();

	/*
	 * reset the VELOCITY AND DENSITY texture
	 * put some speed at the z direction for initialization
	 */
	clear_program->setUniform4f("color", 0, 0, 0, 1);
	velocity_density_fbo->bind();	// test values
	renderComputationRectangleZ();
	velocity_density_fbo->unbind();

	/*
	 * update the FLUID MASS texture
	 *
	 * the fluid mass texture should be the same texture as the fluid fraction
	 * because the fluid mass is 0 on empty cells, between 0 and 1 if it's a interface cell
	 * and at least 1 if it's a fluid cell
	 *
	 * another reason is that that the first computation of the fluid fraction should be the same as the one with the values initialized above
	 * thus we just copy the fluid fraction texture
	 */

	fluid_mass_fbo->bind();

	/*
	 * 1st: set every cell to contain gas (0)
	 */
	clear_program->setUniform4f("color", 0.0, 0.0, 0.0, 1.0);
	glBegin(GL_QUADS);
		glVertex2f(0, 1);
		glVertex2f(1, 1);
		glVertex2f(1, 0);
		glVertex2f(0, 0);
	glEnd();

	/*
	 * 2nd: set the lowest row to be a fluid cell (+1)
	 * the relative heigh is given by interfaceHeight
	 */
	float y_blocks = floorf(floorf(ft_z_height / res_y)*0.25);
	float height = y_blocks*res_y/ft_z_height;
#if 0
	height = 1;
	float interfaceHeight = 0.9;
#else
	float interfaceHeight = 1.0;
#endif

	float blockHeight = (interfaceHeight*res_y)/ft_z_height;

	clear_program->setUniform4f("color", 1.0, 0.0, 0.0, 1.0);
	glBegin(GL_QUADS);
		glVertex2f(	0.0,	0.0	);
		glVertex2f(	1.0,	0.0	);
		glVertex2f(	1.0,	height + blockHeight	);
		glVertex2f(	0.0,	height + blockHeight	);
	glEnd();

	/*
	 * 3rd: setup the interface cells:
	 * the slide at the 1st row and 0 col is set to be an interface cell (0.5)
	 */
	clear_program->setUniform4f("color", 0.5, 0.0, 0.0, 1.0);
	glBegin(GL_QUADS);
		glVertex2f(	0.0,			height + (1.0*res_y)/ft_z_height	);
		glVertex2f(	(1.0*res_x)/ft_z_width,	height + (1.0*res_y)/ft_z_height	);
		glVertex2f(	(1.0*res_x)/ft_z_width,	height + ((1.0+interfaceHeight+(0.5/(float)res_y))*res_y)/ft_z_height	);
		glVertex2f(	0.0,			height + ((1.0+interfaceHeight+(0.5/(float)res_y))*res_y)/ft_z_height	);
	glEnd();
	/**
	 * one line is missing: the interface on the top is created by drawing a line above the "paintins of step 2"
	 */
	clear_program->setUniform4f("color", 0.5, 0.0, 0.0, 1.0);
	glBegin(GL_LINES);
		glVertex2f(	0.0,	height + blockHeight	);
		glVertex2f(	1.0,	height + blockHeight	);
	glEnd();

	fluid_mass_fbo->unbind();
checkGlError();

	/*
	 * reset the FLUID FRACTION/FLAG texture
	 * 
	 * only the red channel is used:
	 * 	-1000:	obstacle cell (always handle as gas)
	 * 	-100:	gas cell
	 * 	+100:	fluid cell
	 * 	else:	interface
	 *
	 * it's also necessary to setup the obstacle cells but we will ignore it here
	 * because they will be setup after the first loop
	 */
	fluid_fraction_fbo->bind();

	/*
	 * 1st: set every cell to contain gas (-100)
	 */
	clear_program->setUniform4f("color", 0.0, -1.0, 0.0, 0.0);
	glBegin(GL_QUADS);
		glVertex2f(0, 1);
		glVertex2f(1, 1);
		glVertex2f(1, 0);
		glVertex2f(0, 0);
	glEnd();

	/*
	 * 2nd: set the lowest row to be a fluid cell (+100)
	 * the relative heigh is given by interfaceHeight
	 */
	clear_program->setUniform4f("color", 1.0, 1.0, 0.0, 0.0);
	glBegin(GL_QUADS);
		glVertex2f(	0.0,	0.0	);
		glVertex2f(	1.0,	0.0	);
		glVertex2f(	1.0,	height + blockHeight	);
		glVertex2f(	0.0,	height + blockHeight	);
	glEnd();

	/*
	 * 3rd: setup the interface cells:
	 * the slide at the 1st row and 0 col is set to be an interface cell (0.5)
	 */
	clear_program->setUniform4f("color", 0.5, 0.0, 0.0, 0.0);
	glBegin(GL_QUADS);
		glVertex2f(	0.0,			height + (1.0*res_y)/ft_z_height	);
		glVertex2f(	(1.0*res_x)/ft_z_width,	height + (1.0*res_y)/ft_z_height	);
		glVertex2f(	(1.0*res_x)/ft_z_width,	height + ((1.0+interfaceHeight+(0.5/(float)res_y))*res_y)/ft_z_height	);
		glVertex2f(	0.0,			height + ((1.0+interfaceHeight+(0.5/(float)res_y))*res_y)/ft_z_height	);
	glEnd();

	/**
	 * one line is missing: the interface on the top is created by drawing a line above the "paintins of step 2"
	 */
	clear_program->setUniform4f("color", 0.5, 0.0, 0.0, 0.0);
	glBegin(GL_LINES);
		glVertex2f(	0.0,	height + blockHeight	);
		glVertex2f(	1.0,	height + blockHeight	);
	glEnd();

#if DEBUG_SHADER && false
	clear_program->setUniform4f("color", 0.5, 0.0, 0.0, 0.0);
	glBegin(GL_LINES);
		glVertex2f(	0.0,	height + blockHeight + 0.1	);
		glVertex2f(	1.0,	height + blockHeight + 0.1	);
	glEnd();
#endif

	fluid_fraction_fbo->unbind();

	clear_program->disable();
	
	/**
	 * now we have to call the subfunctions to initialize the correct border data
	 * otherwise maybe the borders will be loaded with wrong data!
	 */
	voxelisation();	// generate voxels
	collide();	// use voxels to setup feq[3] as obstacle information in texture4
	computate_velocity_density();	// use feq[3] from texture4 to copy information to fluid_fraction.g
	update_cell_state();	// update the cell state to init values

	// shutdown writing to textures
	fbo_write_end();
	fbo_write_restore_state();

#if GET_SIMULATION_TICKS
	ticks_reset();
#endif
}



/**
 * render objstacle (mesh or sphere)
 */
inline void CSimulation::renderObstacle()
{
return;
	switch(fluidObstacleType)
	{
		case 0:
			glCallList(objFileVoxelisationDisplayList);
			break;

		case 1:
			glCallList(sphereVoxelisationDisplayList);
			break;
	}
}

/**
 * update new graviation
 * also update the uniform variables in the collision programs
 */
void CSimulation::setGravitation(vec3f &new_gravitation)
{
	gravitation = new_gravitation*(d_timestep*d_timestep)/d_cell_size;

	int i;
	for (i = 0; i < 5; i++)
	{
		collision_program[i]->enable();
		collision_program[i]->setUniform3f("gravitation", gravitation.data);
		collision_program[i]->disable();
	}
	checkGlError();
}

/**
 * create the simulation
 * \param p_objFile	loaded .obj file
 * \param p_gui		handler for the gui class
 * \param p_res_?	simulation resolution
 * \param p_viscosity	viscosity of fluid in the simulation
 * \param p_x_length	length of x in meters
 */
CSimulation::CSimulation(	CObjFile *p_objFile,
				CGui *p_gui,
				int p_res_x,
				int p_res_y,
				int p_res_z,
				float p_viscosity,
				float p_length_x,
				float p_gravitation
)	:
	
	gui(p_gui),
	d_viscosity(p_viscosity),
	d_length_x(p_length_x),
	d_gravitation(vec3f(0, -p_gravitation, 0)),
	res_x(p_res_x), res_y(p_res_y), res_z(p_res_z),
	modelview_matrix(0),	// modelview matrix which is applied to the object within simulation
	objFile(p_objFile),
	fluidObstacleType(0)
{
	int i;

	d_cell_size = (p_length_x/(float)p_res_x);

/*
 * some dimensionality and normalization stuff
 */
	float compressibility = 0.0005;
	float gravitation_length = d_gravitation.length();
	if (gravitation_length == 0)
	{
		d_timestep = sqrt(	(compressibility*d_cell_size)	/	9.81);
	}
	else
	{
		d_timestep = sqrt(	(compressibility*d_cell_size)/gravitation_length	);
	}

	viscosity = (d_viscosity*d_timestep)/(d_cell_size*d_cell_size);
	gravitation = d_gravitation*(d_timestep*d_timestep)/d_cell_size;

	tau = 3.0*viscosity+0.5;
	omega = 1.0/tau;

/*
 * output new parametrized variables
 */
 /*
	ofstream outfile;
	outfile.open("/tmp/test");
	outfile << tau << "hello" << endl;
	
	outfile << "compressibility: " << compressibility << endl;
	outfile << "dim cell size: " << d_cell_size << endl;
	outfile << "dim length: " << d_length_x << endl;
	outfile << "dim viscosity: " << d_viscosity << endl;
	outfile << "dim gravitation: " << d_gravitation << endl;
	outfile << "---" << endl;
	outfile << "omega: " << omega << endl;
	outfile << "timestep: " << d_timestep << endl;
	outfile << "viscosity: " << viscosity << endl;
	outfile << "gravitation: " << gravitation << endl;
	outfile << "tau: " << tau << endl;
	
	outfile.close();
*/
	cout << "compressibility: " << compressibility << endl;
	cout << "dim cell size: " << d_cell_size << endl;
	cout << "dim length: " << d_length_x << endl;
	cout << "dim viscosity: " << d_viscosity << endl;
	cout << "dim gravitation: " << d_gravitation << endl;
	cout << "---" << endl;
	cout << "omega: " << omega << endl;
	cout << "timestep: " << d_timestep << endl;
	cout << "viscosity: " << viscosity << endl;
	cout << "gravitation: " << gravitation << endl;
	cout << "tau: " << tau << endl;

	if (tau < 0.51 || tau > 2.5)
	{
		cout << "tau has to be within the boundary [0.51; 2.5] - otherwise the sim becomes instable!" << endl;
		goto error_label;
	}

	cout << endl;
	cout << endl;



/*
 * store resolutions and inverse resolutions to array which makes the handling much easier
 * (specially for shader uniform variables)
 */
	fres[0] = (float)p_res_x;
	fres[1] = (float)p_res_y;
	fres[2] = (float)p_res_z;

	finv_res[0] = 1.0f/fres[0];
	finv_res[1] = 1.0f/fres[1];
	finv_res[2] = 1.0f/fres[2];

	/*
	 * initialize flat texture measurement data (resolution, etc.)
	 * ft_a_mod: specifies the amount of textures along axis a which are aligned together in the flat texture
	 * ft_a_height, ft_a_width:	width and size of flat-texture
	 */
	int tmp;	// mod in y direction to init variables

	tmp = ft_z_mod = (int)sqrt(res_z);
	while (ft_z_mod*tmp < res_z)		// search for the best fitting texture sizes
	{
		ft_z_mod++;
		if (ft_z_mod*tmp >= res_z)
			break;
		tmp++;
	}
	ft_z_height = res_y * tmp;
	ft_z_width = res_x * ft_z_mod;

	tmp = ft_y_mod = (int)sqrt(res_y);
	while (ft_y_mod*tmp < res_y)
	{
		ft_y_mod++;
		if (ft_y_mod*tmp >= res_y)
			break;
		tmp++;
	}
	ft_y_height = res_z * tmp;
	ft_y_width = res_x * ft_y_mod;

	tmp = ft_x_mod = (int)sqrt(res_x);
	while (ft_x_mod*tmp < res_x)
	{
		ft_x_mod++;
		if (ft_x_mod*tmp >= res_x)
			break;
		tmp++;
	}
	ft_x_height = res_z * tmp;
	ft_x_width = res_y * ft_x_mod;

#ifdef DEBUG
	cout << "Flat Texture Z: ( " << ft_z_width << " x " << ft_z_height << " )   modulo: " << ft_z_mod << endl;
	cout << "Flat Texture Y: ( " << ft_y_width << " x " << ft_y_height << " )   modulo: " << ft_y_mod << endl;
	cout << "Flat Texture X: ( " << ft_x_width << " x " << ft_x_height << " )   modulo: " << ft_x_mod << endl;
#endif

	/*
	 * CREATE CALLLIST TO WRITE VERY FAST TO TEXTURES
	 */
	write_to_texture_call_list_x = glGenLists(1);
	glNewList(write_to_texture_call_list_x, GL_COMPILE);
	glBegin(GL_QUADS);
		glTexCoord2f(0, 0);		glVertex2f(0, 0);
		glTexCoord2f(ft_x_width, 0);		glVertex2f(1, 0);
		glTexCoord2f(ft_x_width, ft_x_height);	glVertex2f(1, 1);
		glTexCoord2f(0, ft_x_height);		glVertex2f(0, 1);
	glEnd();
	glEndList();

	write_to_texture_call_list_y = glGenLists(1);
	glNewList(write_to_texture_call_list_y, GL_COMPILE);
	glBegin(GL_QUADS);
		glTexCoord2f(0, 0);		glVertex2f(0, 0);
		glTexCoord2f(ft_y_width, 0);		glVertex2f(1, 0);
		glTexCoord2f(ft_y_width, ft_y_height);	glVertex2f(1, 1);
		glTexCoord2f(0, ft_y_height);		glVertex2f(0, 1);
	glEnd();
	glEndList();

	write_to_texture_call_list_z = glGenLists(1);
	glNewList(write_to_texture_call_list_z, GL_COMPILE);
	glBegin(GL_QUADS);
		glTexCoord2f(0, 0);		glVertex2f(0, 0);
		glTexCoord2f(ft_z_width, 0);		glVertex2f(1, 0);
		glTexCoord2f(ft_z_width, ft_z_height);	glVertex2f(1, 1);
		glTexCoord2f(0, ft_z_height);		glVertex2f(0, 1);
	glEnd();
	glEndList();

/*
 * here comes the stuff which we need for the VOXELISATION
 */
#ifdef DEBUG
	cout << "init voxelisation" << endl;	
#endif

	voxel_texture = new CTexture(ft_z_width, ft_z_height, GL_RGBA, GL_RED, GL_BYTE, GL_TEXTURE_RECTANGLE_ARB);
	voxel_buffer_texture[0] = new CTexture(ft_x_width, ft_x_height, GL_RGBA, GL_RED, GL_BYTE, GL_TEXTURE_RECTANGLE_ARB);
	voxel_buffer_texture[1] = new CTexture(ft_y_width, ft_y_height, GL_RGBA, GL_RED, GL_BYTE, GL_TEXTURE_RECTANGLE_ARB);
	voxel_buffer_texture[2] = new CTexture(ft_z_width, ft_z_height, GL_RGBA, GL_RED, GL_BYTE, GL_TEXTURE_RECTANGLE_ARB);

#if CSIMULATION_USE_DEEPTH_PEELING
// TODO: use only one texture for all axes
	voxel_buffer_depth_texture[0] = new CTexture(ft_x_width, ft_x_height, GL_DEPTH_COMPONENT32, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, GL_TEXTURE_RECTANGLE_ARB);
	voxel_buffer_depth_texture[1] = new CTexture(ft_y_width, ft_y_height, GL_DEPTH_COMPONENT32, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, GL_TEXTURE_RECTANGLE_ARB);
	voxel_buffer_depth_texture[2] = new CTexture(ft_z_width, ft_z_height, GL_DEPTH_COMPONENT32, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, GL_TEXTURE_RECTANGLE_ARB);

#if CSIMULATION_USE_DEEPTH_PEELING_2_DEPTH_BUFFER
	voxel_buffer_depth_texture_[0] = new CTexture(ft_x_width, ft_x_height, GL_DEPTH_COMPONENT32, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, GL_TEXTURE_RECTANGLE_ARB);
	voxel_buffer_depth_texture_[1] = new CTexture(ft_y_width, ft_y_height, GL_DEPTH_COMPONENT32, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, GL_TEXTURE_RECTANGLE_ARB);
	voxel_buffer_depth_texture_[2] = new CTexture(ft_z_width, ft_z_height, GL_DEPTH_COMPONENT32, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, GL_TEXTURE_RECTANGLE_ARB);
#endif //CSIMULATION_USE_DEEPTH_PEELING_2_DEPTH_BUFFER

#endif

	if (!voxel_texture || !voxel_buffer_texture[0]->valid || !voxel_buffer_texture[1]->valid || !voxel_buffer_texture[2]->valid)
	{
		cerr << "ERROR after creating voxelisation textures" << endl;
		goto error_voxelisation_textures;
	}
	// setup voxelisation fbos
	voxel_fbo = new CFbo();
	voxel_fbo->bind();	voxel_fbo->bindFramebufferTexture(voxel_texture);	voxel_fbo->unbind();

	voxel_buffer_fbo[0] = new CFbo();
	voxel_buffer_fbo[0]->bind();
		voxel_buffer_fbo[0]->bindFramebufferTexture(voxel_buffer_texture[0]);
#if CSIMULATION_USE_DEEPTH_PEELING
		voxel_buffer_fbo[0]->bindFramebufferDepthTexture(voxel_buffer_depth_texture[0]);
#endif
	voxel_buffer_fbo[0]->unbind();

	voxel_buffer_fbo[1] = new CFbo();
	voxel_buffer_fbo[1]->bind();
		voxel_buffer_fbo[1]->bindFramebufferTexture(voxel_buffer_texture[1]);
#if CSIMULATION_USE_DEEPTH_PEELING
		voxel_buffer_fbo[1]->bindFramebufferDepthTexture(voxel_buffer_depth_texture[1]);
#endif
	voxel_buffer_fbo[1]->unbind();

	voxel_buffer_fbo[2] = new CFbo();
	voxel_buffer_fbo[2]->bind();
		voxel_buffer_fbo[2]->bindFramebufferTexture(voxel_buffer_texture[2]);
#if CSIMULATION_USE_DEEPTH_PEELING
		voxel_buffer_fbo[2]->bindFramebufferDepthTexture(voxel_buffer_depth_texture[2]);
#endif
	voxel_buffer_fbo[2]->unbind();

	// load voxelisation program
#if CSIMULATION_USE_DEEPTH_PEELING
	voxel_program_combine = new CProgram("shader/voxel_combine_depth_peeling.frag", NULL);
#else
	voxel_program_combine = new CProgram("shader/voxel_combine_slices.frag", NULL);
#endif
	if (voxel_program_combine->error)
	{
		cerr << "ERROR after creating voxelisation program" << endl;
		cerr << voxel_program_combine->errorLog << endl;
		goto error_voxelisation_program;
	}

	voxel_program_combine->enable();

	voxel_program_combine->setUniform3f("res", fres.data);

#if CSIMULATION_USE_DEEPTH_PEELING
	voxel_program_combine->setUniform3f("inv_res", finv_res);
	voxel_program_combine->setUniform1f("tex_z_mod", ft_z_mod);

	voxel_program_combine->setUniform2f("texture_size_x", (float)voxel_buffer_texture[0]->width, (float)voxel_buffer_texture[0]->height);
	voxel_program_combine->setUniform2f("texture_size_y", (float)voxel_buffer_texture[1]->width, (float)voxel_buffer_texture[1]->height);
	voxel_program_combine->setUniform2f("texture_size_z", (float)voxel_buffer_texture[2]->width, (float)voxel_buffer_texture[2]->height);
#else
	voxel_program_combine->setUniform1f("tex_x_mod", ft_x_mod);
	voxel_program_combine->setUniform1f("tex_y_mod", ft_y_mod);
	voxel_program_combine->setUniform1f("tex_z_mod", ft_z_mod);
#endif

	voxel_program_combine->setUniform1i("TextureZ", 0);
	voxel_program_combine->setUniform1i("TextureY", 1);
	voxel_program_combine->setUniform1i("TextureX", 2);

	voxel_program_combine->disable();
checkGlError();

#if CSIMULATION_USE_DEEPTH_PEELING
	glGenQueries(1, &voxel_query_object);

	// load voxel peeling program for x depth peeling
	voxel_peeling_program_first = new CProgram("shader/voxel_depth_peeling.vert", "shader/voxel_depth_peeling_first.frag", NULL);
	if (voxel_peeling_program_first->error)
	{
		cerr << "ERROR after creating voxel depth peeling program_first" << endl;
		cerr << voxel_peeling_program_first->errorLog << endl;
		goto error_voxel_peeling_program_first;
	}

	// load voxel peeling program for x depth peeling
	voxel_peeling_program[0] = new CProgram("shader/voxel_depth_peeling.vert", "shader/voxel_depth_peeling_x.frag", NULL);
	if (voxel_peeling_program[0]->error)
	{
		cerr << "ERROR after creating voxel depth peeling program[0]" << endl;
		cerr << voxel_peeling_program[0]->errorLog << endl;
		goto error_voxel_peeling_program_0;
	}

	// load voxel peeling program for y depth peeling
	voxel_peeling_program[1] = new CProgram("shader/voxel_depth_peeling.vert", "shader/voxel_depth_peeling_y.frag", NULL);
	if (voxel_peeling_program[1]->error)
	{
		cerr << "ERROR after creating voxel depth peeling program[0]" << endl;
		cerr << voxel_peeling_program[1]->errorLog << endl;
		goto error_voxel_peeling_program_1;
	}

	// load voxel peeling program for z depth peeling
	voxel_peeling_program[2] = new CProgram("shader/voxel_depth_peeling.vert", "shader/voxel_depth_peeling_z.frag", NULL);
	if (voxel_peeling_program[2]->error)
	{
		cerr << "ERROR after creating voxel depth peeling program[2]" << endl;
		cerr << voxel_peeling_program[2]->errorLog << endl;
		goto error_voxel_peeling_program_2;
	}

	for (i = 0; i <= 2; i++)
	{
		voxel_peeling_program[i]->enable();
		voxel_peeling_program[i]->setUniform3f("res", fres);
		voxel_peeling_program[i]->setUniform1i("TextureDepth", 0);
		voxel_peeling_program[i]->setUniform1f("texture_width", (float)voxel_buffer_depth_texture[i]->width);

		float inv_texture_size[2] = {1.0f/(float)voxel_buffer_depth_texture[i]->width, 1.0f/(float)voxel_buffer_depth_texture[i]->height};

		switch(i)
		{
			case 0:	voxel_peeling_program[0]->setUniform1f("inv_res_x_mul_0_5", finv_res[0]*0.5f);
				break;
			case 1:	voxel_peeling_program[1]->setUniform1f("inv_res_y_mul_0_5", finv_res[1]*0.5f);
				break;
			case 2:	voxel_peeling_program[2]->setUniform1f("inv_res_z_mul_0_5", finv_res[2]*0.5f);
				break;
		}

		voxel_peeling_program[i]->disable();
	}

checkGlError();

#endif
checkGlError();

/*
 * some BUFFER TEXTURES
 */
#ifdef DEBUG
	cout << "init buffer texture" << endl;	
#endif
	buffer_texture = new CTexture(ft_z_width, ft_z_height, GL_RGBA32F_ARB, GL_RGBA, GL_FLOAT, GL_TEXTURE_RECTANGLE_ARB);
	buffer2_texture = new CTexture(ft_z_width, ft_z_height, GL_RGBA32F_ARB, GL_RGBA, GL_FLOAT, GL_TEXTURE_RECTANGLE_ARB);
	if (!buffer_texture->valid || !buffer2_texture->valid)
	{
		cerr << "ERROR after creating buffer texture" << endl;
		goto error_buffer_texture;
	}
checkGlError();

	buffer_fbo = new CFbo(buffer_texture);
	buffer2_fbo = new CFbo(buffer2_texture);


/**
 * VELOCITY AND DENSITY STUFF texture and shader
 *
 * the velocity and density shader also computates the new fluid fraction
 */
#ifdef DEBUG
	cout << "init velocity and density texture" << endl;	
#endif
	velocity_density_texture = new CTexture(ft_z_width, ft_z_height, GL_RGBA32F_ARB, GL_RGBA, GL_FLOAT, GL_TEXTURE_RECTANGLE_ARB);
	if (!velocity_density_texture->valid)
	{
		cerr << "ERROR after creating velocity and density texture" << endl;
		goto error_velocity_density_texture;
	}

	velocity_density_fbo = new CFbo(velocity_density_texture);

checkGlError();

	velocity_density_program = new CProgram("shader/velocity_density.frag", NULL);
	if (velocity_density_program->error)
	{
		cerr << "ERROR after creating velocity_density program" << endl;
		cerr << velocity_density_program->errorLog << endl;
		goto error_velocity_density_program;
	}
	velocity_density_program->enable();
	velocity_density_program->setUniform1i("Texture0", 0);
	velocity_density_program->setUniform1i("Texture1", 1);
	velocity_density_program->setUniform1i("Texture2", 2);
	velocity_density_program->setUniform1i("Texture3", 3);
	velocity_density_program->setUniform1i("Texture4", 4);
	velocity_density_program->setUniform1i("TextureFluidMass", 5);
	velocity_density_program->setUniform1i("TextureFluidFraction", 6);
	velocity_density_program->disable();

/*
 * DISTRIBUTION TEXTURES and PROGRAMS
 */
#ifdef DEBUG
	cout << "init distribution texture" << endl;	
#endif
	for (i = 0; i < 5; i++)
	{
		distributions_texture[i] = new CTexture(ft_z_width, ft_z_height, GL_RGBA32F_ARB, GL_RGBA, GL_FLOAT, GL_TEXTURE_RECTANGLE_ARB);
		distributions_fbo[i] = new CFbo(distributions_texture[i]);
	}

	for (i = 0; i < 5; i++)
	{
		if (!distributions_texture[i]->valid)
		{
			cerr << "ERROR after creating distribution texture" << endl;
			goto error_distributions;
		}
	}

/*
 * texture for FLUID FRACTION
 */
#ifdef DEBUG
	cout << "fluid fraction texture" << endl;	
#endif
	fluid_fraction_texture = new CTexture(ft_z_width, ft_z_height, GL_RGBA32F_ARB, GL_RGBA, GL_FLOAT, GL_TEXTURE_RECTANGLE_ARB);
	fluid_fraction_fbo = new CFbo(fluid_fraction_texture);

	if (!fluid_fraction_texture->valid)
	{
		cerr << "ERROR after creating fluid fraction texture" << endl;
		goto error_fluid_fraction;
	}

/*
 * update the CELL STATE according to new and previous fluid fraction
 */
	/*
	 * STEP 1
	 * search for interface cells which are on the change and set change flag
	 */
	update_cell_state_first_program = new CProgram("shader/update_cell_state_first.frag", NULL);
	if (update_cell_state_first_program->error)
	{
		cerr << "ERROR after creating update_cell_first_state program" << endl;
		cerr << update_cell_state_first_program->errorLog << endl;
		goto error_update_cell_state_first_program;
	}
	update_cell_state_first_program->enable();
	update_cell_state_first_program->setUniform1i("TextureFluidFraction", 0);
	update_cell_state_first_program->setUniform3f("res", fres.data);
	update_cell_state_first_program->setUniform3f("res_sub_1", fres[0]-1.0f, fres[1]-1.0f, fres[2]-1.0f);
	update_cell_state_first_program->setUniform1f("ft_z_mod", ft_z_mod);
	update_cell_state_first_program->disable();

	/*
	 * STEP 2
	 * for fluid and gas cells: search for neighbored change flags
	 * and change to interface cell
	 */
	update_cell_state_second_program = new CProgram("shader/update_cell_state_second.frag", NULL);
	if (update_cell_state_second_program->error)
	{
		cerr << "ERROR after creating update_cell_state_second program" << endl;
		cerr << update_cell_state_second_program->errorLog << endl;
		goto error_update_cell_state_second_program;
	}
	update_cell_state_second_program->enable();
	update_cell_state_second_program->setUniform1i("TextureFluidFraction", 0);
	update_cell_state_second_program->setUniform3f("res", fres.data);
	update_cell_state_second_program->setUniform3f("res_sub_1", fres[0]-1.0f, fres[1]-1.0f, fres[2]-1.0f);
	update_cell_state_second_program->setUniform1f("ft_z_mod", ft_z_mod);


	/*
	 * STEP 3
	 * for fluid and gas cells: search for neighbored change flags
	 * and change to interface cell
	 */
	update_cell_state_third_program = new CProgram("shader/update_cell_state_third.frag", NULL);
	if (update_cell_state_third_program->error)
	{
		cerr << "ERROR after creating update_cell_state_third program" << endl;
		cerr << update_cell_state_third_program->errorLog << endl;
		goto error_update_cell_state_third_program;
	}
	update_cell_state_third_program->enable();
	update_cell_state_third_program->setUniform1i("TextureFluidFraction", 0);
	update_cell_state_third_program->setUniform1i("TextureVelocityDensity", 1);
	update_cell_state_third_program->setUniform3f("res", fres.data);
	update_cell_state_third_program->setUniform3f("res_sub_1", fres[0]-1.0f, fres[1]-1.0f, fres[2]-1.0f);
	update_cell_state_third_program->setUniform1f("ft_z_mod", ft_z_mod);

	/*
	 * STEP 4
	 * search for new gas cell and setup the DF's
	 */
	update_cell_state_fourth_program[0] = new CProgram("shader/update_cell_state_fourth_0.frag", NULL);
	update_cell_state_fourth_program[1] = new CProgram("shader/update_cell_state_fourth_1.frag", NULL);
	update_cell_state_fourth_program[2] = new CProgram("shader/update_cell_state_fourth_2.frag", NULL);
	update_cell_state_fourth_program[3] = new CProgram("shader/update_cell_state_fourth_3.frag", NULL);
	update_cell_state_fourth_program[4] = new CProgram("shader/update_cell_state_fourth_4.frag", NULL);
	if (	update_cell_state_fourth_program[0]->error	||
		update_cell_state_fourth_program[1]->error	||
		update_cell_state_fourth_program[2]->error	||
		update_cell_state_fourth_program[3]->error	||
		update_cell_state_fourth_program[4]->error
	)
	{
		cerr << "ERROR after creating update_cell_fourth_state program" << endl;
		for (i = 0; i < 5; i++)
			if (update_cell_state_fourth_program[i]->error)
				cout << update_cell_state_fourth_program[i]->errorLog << endl;
		goto error_update_cell_state_fourth_program;
	}

	for (i = 0; i < 5; i++)
	{
		update_cell_state_fourth_program[i]->enable();
		update_cell_state_fourth_program[i]->setUniform1i("TextureDF", 0);
		update_cell_state_fourth_program[i]->setUniform1i("TextureFluidFraction", 1);
		update_cell_state_fourth_program[i]->setUniform1i("TextureVelocityDensity", 2);
		update_cell_state_fourth_program[i]->setUniform3f("res", fres.data);
		update_cell_state_fourth_program[i]->setUniform3f("res_sub_1", fres[0]-1.0f, fres[1]-1.0f, fres[2]-1.0f);
		update_cell_state_fourth_program[i]->setUniform1f("ft_z_mod", ft_z_mod);
		update_cell_state_fourth_program[i]->disable();
	}


/*
 * texture for FLUID MASS
 */
#ifdef DEBUG
	cout << "fluid fraction texture" << endl;	
#endif
	fluid_mass_texture = new CTexture(ft_z_width, ft_z_height, GL_RGBA32F_ARB, GL_RGBA, GL_FLOAT, GL_TEXTURE_RECTANGLE_ARB);
	fluid_mass_fbo = new CFbo(fluid_mass_texture);

	if (!fluid_mass_texture->valid)
	{
		cerr << "ERROR after creating fluid mass texture" << endl;
		goto error_fluid_mass;
	}


/*
 * COLLISION & PROPAGATION PROGRAM
 */
#ifdef DEBUG
	cout << "init collision program" << endl;	
#endif
	collision_program[0] = new CProgram("shader/collision_tex_0.frag", NULL);
	collision_program[1] = new CProgram("shader/collision_tex_1.frag", NULL);
	collision_program[2] = new CProgram("shader/collision_tex_2.frag", NULL);
	collision_program[3] = new CProgram("shader/collision_tex_3.frag", NULL);
	collision_program[4] = new CProgram("shader/collision_tex_4.frag", NULL);

	if (	collision_program[0]->error	||
		collision_program[1]->error	||
		collision_program[2]->error	||
		collision_program[3]->error	||
		collision_program[4]->error
	)
	{
		cerr << "ERROR after creating collision programs" << endl;
		for (i = 0; i < 5; i++)
			if (collision_program[i]->error)
				cout << collision_program[i]->errorLog << endl;
		goto error_collision_program;
	}

	for (i = 0; i < 5; i++)
	{
		collision_program[i]->enable();
		collision_program[i]->setUniform1i("TextureDF", 0);
		collision_program[i]->setUniform1i("TextureVector", 1);
		collision_program[i]->setUniform1i("TextureVoxel", 2);
		collision_program[i]->setUniform3f("res", fres.data);
		collision_program[i]->setUniform1f("ft_z_mod", ft_z_mod);
		collision_program[i]->setUniform1f("omega", omega);
		collision_program[i]->setUniform3f("gravitation", gravitation.data);
		collision_program[i]->disable();
	checkGlError();
	}

#ifdef DEBUG
	cout << "init propagation program" << endl;	
#endif

	propagation_program[0] = new CProgram("shader/propagation_tex_0.frag", NULL);
	propagation_program[1] = new CProgram("shader/propagation_tex_1.frag", NULL);
	propagation_program[2] = new CProgram("shader/propagation_tex_2.frag", NULL);
	propagation_program[3] = new CProgram("shader/propagation_tex_3.frag", NULL);
	propagation_program[4] = new CProgram("shader/propagation_tex_4.frag", NULL);

	if (	propagation_program[0]->error	||
		propagation_program[1]->error	||
		propagation_program[2]->error	||
		propagation_program[3]->error	||
		propagation_program[4]->error
	)
	{
		cerr << "ERROR after creating propagation programs" << endl;
		for (i = 0; i < 5; i++)
			if (propagation_program[i]->error)
				cout << propagation_program[i]->errorLog << endl;

		goto error_propagation_program;
	}

	for (i = 0; i < 5; i++)
	{
		propagation_program[i]->enable();
		propagation_program[i]->setUniform1i("TextureDF", 0);
		propagation_program[i]->setUniform1i("TextureFluidMass", 1);
		propagation_program[i]->setUniform1i("TextureFluidFraction", 2);
		propagation_program[i]->setUniform1i("TextureVelocityDensity", 3);
		propagation_program[i]->setUniform3f("res", fres.data);
		propagation_program[i]->setUniform3f("res_sub_1", fres[0]-1.0f, fres[1]-1.0f, fres[2]-1.0f);
		propagation_program[i]->setUniform1f("ft_z_mod", ft_z_mod);
		propagation_program[i]->disable();
	checkGlError();
	}

/*
 * we also need a CLEAR PROGRAM to initialize textures with specific values
 */
	clear_program = new CProgram("shader/clear_texture.frag", NULL);
	if (clear_program->error)
	{
		cerr << "ERROR after creating clear program" << endl;
		cerr << clear_program->errorLog << endl;
		goto error_clear_program;
	}
checkGlError();

// OBJECT
#ifdef DEBUG
	cout << "init object display lists" << endl;	
#endif
	// create display list for objectfile
	objFileVoxelisationDisplayList = glGenLists(1);
	glNewList(objFileVoxelisationDisplayList, GL_COMPILE);

#if !CSIMULATION_USE_DEEPTH_PEELING
	renderObjfileVoxel(*objFile, false);
#else
	renderObjfileVoxelWithNormals(*objFile, false);

#endif
	glEndList();

// create display list for sphere
	sphereVoxelisationDisplayList = glGenLists(1);
	glNewList(sphereVoxelisationDisplayList, GL_COMPILE);
	CPrimitives::sphere(0.4, 0, 0, 0, CSIMULATION_SPHERE_LOD);
	glEndList();


	worldBoxDisplayList = glGenLists(1);
	glNewList(worldBoxDisplayList, GL_COMPILE);
	float borderPos[3];
	/*
	borderPos[0] = 0.5f + 0.1f/fres[0];	// slightly move the border to the center of the cell for accurate voxelisation
	borderPos[1] = 0.5f + 0.5f/fres[1];
	borderPos[2] = 0.5f + 0.5f/fres[2];
	*/
	borderPos[0] = 0.5f + 0;	// slightly move the border to the center of the cell for accurate voxelisation
	borderPos[1] = 0.5f + 0;
	borderPos[2] = 0.5f + 0;

	glBegin(GL_QUADS);
		glNormal3f(1, 0, 0);
			glVertex3f(-borderPos[0], -0.5, -0.5);
			glVertex3f(-borderPos[0], 0.5, -0.5);
			glVertex3f(-borderPos[0], 0.5, 0.5);
			glVertex3f(-borderPos[0], -0.5, 0.5);
		glNormal3f(1, 0, 0);
			glVertex3f(borderPos[0], -0.5, -0.5);
			glVertex3f(borderPos[0], 0.5, -0.5);
			glVertex3f(borderPos[0], 0.5, 0.5);
			glVertex3f(borderPos[0], -0.5, 0.5);

		glNormal3f(0, 1, 0);
			glVertex3f(-0.5, -borderPos[1], -0.5);
			glVertex3f(0.5, -borderPos[1], -0.5);
			glVertex3f(0.5, -borderPos[1], 0.5);
			glVertex3f(-0.5, -borderPos[1], 0.5);
		glNormal3f(0, 1, 0);
			glVertex3f(-0.5, borderPos[1], -0.5);
			glVertex3f(0.5, borderPos[1], -0.5);
			glVertex3f(0.5,	borderPos[1], 0.5);
			glVertex3f(-0.5, borderPos[1], 0.5);

		glNormal3f(0, 0, 1);
			glVertex3f(-0.5, -0.5, -borderPos[2]);
			glVertex3f(0.5, -0.5, -borderPos[2]);
			glVertex3f(0.5, 0.5, -borderPos[2]);
			glVertex3f(-0.5, 0.5, -borderPos[2]);
		glNormal3f(0, 0, 1);
			glVertex3f(-0.5, -0.5, borderPos[2]);
			glVertex3f(0.5, -0.5, borderPos[2]);
			glVertex3f(0.5,	0.5, borderPos[2]);
			glVertex3f(-0.5, 0.5, borderPos[2]);

	glEnd();

	glEndList();


	checkGlError();

	reset();	// reset everything

	ready = true;
	return;			// RETURN


//error_objfile_displaylists:
	glDeleteLists(objFileVoxelisationDisplayList, 1);
	glDeleteLists(sphereVoxelisationDisplayList, 1);
	delete voxel_program_combine;

error_clear_program:
	delete clear_program;

//error_fluid_fraction_program:
//	delete fluid_fraction_program;

error_fluid_mass:
	delete fluid_mass_texture;
	delete fluid_mass_fbo;


error_fluid_fraction:
	delete fluid_fraction_texture;
	delete fluid_fraction_fbo;

error_distributions:
	for (i = 0; i < 5; i++)
	{
		delete distributions_fbo[i];
		delete distributions_texture[i];
	}

	delete velocity_density_fbo;

error_propagation_program:
	for (i = 0; i < 5; i++)
		delete propagation_program[i];

error_collision_program:
	for (i = 0; i < 5; i++)
		delete collision_program[i];

error_update_cell_state_fourth_program:
	for (i = 0; i < 5; i++)
		delete update_cell_state_fourth_program[i];
error_update_cell_state_third_program:
	delete update_cell_state_third_program;
error_update_cell_state_second_program:
	delete update_cell_state_second_program;
error_update_cell_state_first_program:
	delete update_cell_state_first_program;

error_velocity_density_program:
	delete velocity_density_program;

error_velocity_density_texture:
	delete velocity_density_texture;

	delete buffer_fbo;
	delete buffer2_fbo;

error_buffer_texture:
	delete buffer_texture;
	delete buffer2_texture;

#if CSIMULATION_USE_DEEPTH_PEELING
error_voxel_peeling_program_2:
	delete voxel_peeling_program[2];
error_voxel_peeling_program_1:
	delete voxel_peeling_program[1];
error_voxel_peeling_program_0:
	delete voxel_peeling_program[0];
error_voxel_peeling_program_first:
	delete voxel_peeling_program_first;
#endif
	delete voxel_program_combine;

error_voxelisation_program:
	delete voxel_fbo;
	delete voxel_buffer_fbo[0];
	delete voxel_buffer_fbo[1];
	delete voxel_buffer_fbo[2];

error_voxelisation_textures:

	delete voxel_texture;
	delete voxel_buffer_texture[0];
	delete voxel_buffer_texture[1];
	delete voxel_buffer_texture[2];

error_label:
	ready = false;
}


inline void CSimulation::renderComputationRectangleZ()
{
#if 0
	glCallList(write_to_texture_call_list_z);
#else
	glBegin(GL_QUADS);
		glTexCoord2f(0, 0);		glVertex2f(0, 0);
		glTexCoord2f(ft_z_width, 0);		glVertex2f(1, 0);
		glTexCoord2f(ft_z_width, ft_z_height);	glVertex2f(1, 1);
		glTexCoord2f(0, ft_z_height);		glVertex2f(0, 1);
	glEnd();
#endif
}


/******************************************
 * VOXELISATION
 *
 * lighting, blending, depth text and cull face should be deactivated
 * for performance reasons
 ******************************************/
void CSimulation::voxelisation()
{
#if CSIMULATION_USE_DEEPTH_PEELING
	int x, y, z;
checkGlError();
	// initialize values
	glEnable(GL_TEXTURE_2D);

	glClearColor(0, 0, 0, 0);	// clear every face with black color
	glColor4f(1, 1, 1, 1);		// turn on white color for voxelisation

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();

checkGlError();

	int samples_passed = -1;

	/**
	 * here it's going to be a little bit tricky
	 * we use only one depth buffer for depth peeling:
	 * because we create flat textures, we can read the depth
	 * from the depth buffer to which the recent depth is written to
	 */
	glEnable(GL_DEPTH_TEST);

	glMatrixMode(GL_MODELVIEW);	// reset modelview matrix
	glLoadIdentity();

	// rotate the model - this is the only operation which is done on the modelview matrix for every depth peeling
	if (modelview_matrix)	glMultMatrixf(modelview_matrix);

// Z AXIS (res_x, res_y)
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-0.5, 0.5, -0.5, 0.5, 1.0f, -1.0f);

	// prepare framebuffer & clear depth buffers
	voxel_buffer_fbo[2]->bind();

	glViewport(0, 0, voxel_buffer_depth_texture[2]->width, voxel_buffer_depth_texture[2]->height); 

#if CSIMULATION_USE_DEEPTH_PEELING_2_DEPTH_BUFFER
	voxel_buffer_fbo[2]->bindFramebufferDepthTexture(voxel_buffer_depth_texture_[2]);
	glClear(GL_DEPTH_BUFFER_BIT);
	voxel_buffer_fbo[2]->bindFramebufferDepthTexture(voxel_buffer_depth_texture[2]);
#endif
	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

	// draw on the first slice without depth peeling
	glViewport(0, 0, res_x, res_y);
	glBeginQuery(GL_SAMPLES_PASSED, voxel_query_object);
	voxel_peeling_program_first->enable();
		renderObstacle();

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();		glLoadIdentity();
			glCallList(worldBoxDisplayList);
		glPopMatrix();

	voxel_peeling_program_first->disable();
	glEndQuery(GL_SAMPLES_PASSED);
	glGetQueryObjectiv(voxel_query_object, GL_QUERY_RESULT, &samples_passed);

	voxel_buffer_depth_texture[2]->bind();	// bind depth texture from framebuffer
#if !CSIMULATION_USE_DEEPTH_PEELING_2_DEPTH_BUFFER
	voxel_buffer_fbo[2]->bindFramebufferDepthTexture(voxel_buffer_depth_texture[2]);
#endif
	// start the real part of depth peeling
	voxel_peeling_program[2]->enable();
	for (z = 1; z < res_z && samples_passed; z++)
	{
#if CSIMULATION_USE_DEEPTH_PEELING_2_DEPTH_BUFFER
		if (z % 2)
		{
			voxel_buffer_depth_texture_[2]->unbind();	// bind depth texture from framebuffer
			voxel_buffer_fbo[2]->bindFramebufferDepthTexture(voxel_buffer_depth_texture_[2]);
			voxel_buffer_depth_texture[2]->bind();	// bind depth texture from framebuffer
		}
		else
		{
			voxel_buffer_depth_texture[2]->unbind();	// bind depth texture from framebuffer
			voxel_buffer_fbo[2]->bindFramebufferDepthTexture(voxel_buffer_depth_texture[2]);
			voxel_buffer_depth_texture_[2]->bind();	// bind depth texture from framebuffer
		}
#endif
		glViewport((z % ft_z_mod) * res_x, (z / ft_z_mod) * res_y, res_x, res_y);
		glBeginQuery(GL_SAMPLES_PASSED, voxel_query_object);
			renderObstacle();

			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();		glLoadIdentity();
				glCallList(worldBoxDisplayList);
			glPopMatrix();

		glEndQuery(GL_SAMPLES_PASSED);
		glGetQueryObjectiv(voxel_query_object, GL_QUERY_RESULT, &samples_passed);
		glFinish();	// call this because we read and write to the same buffer
	}

	voxel_peeling_program[2]->disable();
	voxel_buffer_depth_texture[2]->unbind();
	voxel_buffer_fbo[2]->unbind();

	voxel_program_combine->enable();
	voxel_program_combine->setUniform1f("max_z_slices", (float)z);
	voxel_program_combine->disable();

// Y AXIS (res_x, res_z)
	glMatrixMode(GL_PROJECTION);	// perform moving operation
	glLoadIdentity();
	glOrtho(0.5, -0.5, -0.5, 0.5, 1.0f, -1.0f);	// mirror the result on vertical axis! (for faster combination)

	// we dont rotate the modelview because we wont change the normals!
//	glMatrixMode(GL_MODELVIEW);	// perform moving operation
//	glLoadIdentity();
	glRotatef(90, 1, 0, 0);
	glRotatef(180, 0, 1, 0);

	// prepare framebuffer & clear depth buffers
	voxel_buffer_fbo[1]->bind();

	glViewport(0, 0, voxel_buffer_depth_texture[1]->width, voxel_buffer_depth_texture[1]->height); 

#if CSIMULATION_USE_DEEPTH_PEELING_2_DEPTH_BUFFER
	voxel_buffer_fbo[1]->bindFramebufferDepthTexture(voxel_buffer_depth_texture_[1]);
	glClear(GL_DEPTH_BUFFER_BIT);
	voxel_buffer_fbo[1]->bindFramebufferDepthTexture(voxel_buffer_depth_texture[1]);
#endif
	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

	// draw on the first slice without depth peeling
	glViewport(0, 0, res_x, res_z);
	glBeginQuery(GL_SAMPLES_PASSED, voxel_query_object);
	voxel_peeling_program_first->enable();
		renderObstacle();

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();		glLoadIdentity();
			glCallList(worldBoxDisplayList);
		glPopMatrix();

	voxel_peeling_program_first->disable();
	glEndQuery(GL_SAMPLES_PASSED);
	glGetQueryObjectiv(voxel_query_object, GL_QUERY_RESULT, &samples_passed);

	voxel_buffer_depth_texture[1]->bind();	// bind depth texture from framebuffer
#if !CSIMULATION_USE_DEEPTH_PEELING_2_DEPTH_BUFFER
	voxel_buffer_fbo[1]->bindFramebufferDepthTexture(voxel_buffer_depth_texture[1]);
#endif
	// start the real part of depth peeling
	voxel_peeling_program[1]->enable();
	for (y = 1; y < res_y && samples_passed; y++)
	{
#if CSIMULATION_USE_DEEPTH_PEELING_2_DEPTH_BUFFER
		if (y % 2)
		{
			voxel_buffer_depth_texture_[1]->unbind();	// bind depth texture from framebuffer
			voxel_buffer_fbo[1]->bindFramebufferDepthTexture(voxel_buffer_depth_texture_[1]);
			voxel_buffer_depth_texture[1]->bind();	// bind depth texture from framebuffer
		}
		else
		{
			voxel_buffer_depth_texture[1]->unbind();	// bind depth texture from framebuffer
			voxel_buffer_fbo[1]->bindFramebufferDepthTexture(voxel_buffer_depth_texture[1]);
			voxel_buffer_depth_texture_[1]->bind();	// bind depth texture from framebuffer
		}
#endif
		glViewport((y % ft_y_mod) * res_x, (y / ft_y_mod) * res_z, res_x, res_z);
		glBeginQuery(GL_SAMPLES_PASSED, voxel_query_object);
			renderObstacle();

			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();		glLoadIdentity();
				glCallList(worldBoxDisplayList);
			glPopMatrix();

		glEndQuery(GL_SAMPLES_PASSED);
		glGetQueryObjectiv(voxel_query_object, GL_QUERY_RESULT, &samples_passed);
		glFinish();	// call this because we read and write to the same buffer
	}
	voxel_peeling_program[1]->disable();
	voxel_buffer_depth_texture[1]->unbind();

	voxel_buffer_fbo[1]->unbind();

	voxel_program_combine->enable();
	voxel_program_combine->setUniform1f("max_y_slices", (float)y);
	voxel_program_combine->disable();


// X AXIS (res_y, res_z)
	glMatrixMode(GL_PROJECTION);	// perform moving operation
	glLoadIdentity();
	glOrtho(-0.5, 0.5, -0.5, 0.5, 1.0f, -1.0f);	// mirror the result on vertical axis! (for faster combination)

	// dont use modelview because of normal
//	glMatrixMode(GL_MODELVIEW);	// perform moving operation
//	glLoadIdentity();
	glRotatef(-90, 0, 0, 1);
	glRotatef(-90, 0, 1, 0);

	// prepare framebuffer & clear depth buffers
	voxel_buffer_fbo[0]->bind();

	glViewport(0, 0, voxel_buffer_depth_texture[0]->width, voxel_buffer_depth_texture[0]->height); 
#if CSIMULATION_USE_DEEPTH_PEELING_2_DEPTH_BUFFER
	voxel_buffer_fbo[0]->bindFramebufferDepthTexture(voxel_buffer_depth_texture_[0]);
	glClear(GL_DEPTH_BUFFER_BIT);
	voxel_buffer_fbo[0]->bindFramebufferDepthTexture(voxel_buffer_depth_texture[0]);
#endif
	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

	// draw on the first slice without depth peeling
	glViewport(0, 0, res_y, res_z);
	glBeginQuery(GL_SAMPLES_PASSED, voxel_query_object);
	voxel_peeling_program_first->enable();
		renderObstacle();

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();		glLoadIdentity();
			glCallList(worldBoxDisplayList);
		glPopMatrix();

	voxel_peeling_program_first->disable();
	glEndQuery(GL_SAMPLES_PASSED);
	glGetQueryObjectiv(voxel_query_object, GL_QUERY_RESULT, &samples_passed);

	voxel_buffer_depth_texture[0]->bind();	// bind depth texture from framebuffer
#if !CSIMULATION_USE_DEEPTH_PEELING_2_DEPTH_BUFFER
	voxel_buffer_fbo[0]->bindFramebufferDepthTexture(voxel_buffer_depth_texture[0]);
#endif
	// start the real part of depth peeling
	voxel_peeling_program[0]->enable();
	for (x = 1; x < res_x && samples_passed; x++)
	{
#if CSIMULATION_USE_DEEPTH_PEELING_2_DEPTH_BUFFER
		if (x % 2)
		{
			voxel_buffer_depth_texture_[0]->unbind();	// bind depth texture from framebuffer
			voxel_buffer_fbo[0]->bindFramebufferDepthTexture(voxel_buffer_depth_texture_[0]);
			voxel_buffer_depth_texture[0]->bind();	// bind depth texture from framebuffer
		}
		else
		{
			voxel_buffer_depth_texture[0]->unbind();	// bind depth texture from framebuffer
			voxel_buffer_fbo[0]->bindFramebufferDepthTexture(voxel_buffer_depth_texture[0]);
			voxel_buffer_depth_texture_[0]->bind();	// bind depth texture from framebuffer
		}
#endif
		glViewport((x % ft_x_mod) * res_y, (x / ft_x_mod) * res_z, res_y, res_z);
		glBeginQuery(GL_SAMPLES_PASSED, voxel_query_object);
			renderObstacle();

			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();		glLoadIdentity();
				glCallList(worldBoxDisplayList);
			glPopMatrix();

		glEndQuery(GL_SAMPLES_PASSED);
		glGetQueryObjectiv(voxel_query_object, GL_QUERY_RESULT, &samples_passed);
		glFinish();	// call this because we read and write to the same buffer
	}
	voxel_peeling_program[0]->disable();
	voxel_buffer_depth_texture[0]->unbind();

	voxel_buffer_fbo[0]->unbind();

	voxel_program_combine->enable();
	voxel_program_combine->setUniform1f("max_x_slices", (float)x);
	voxel_program_combine->disable();


// AXIS END

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	// COMBINE THE DEPTH CULLED SLICES

	voxel_fbo->bind();

checkGlError();
// prepare rendering to texture
	fbo_write_start(voxel_texture->width, voxel_texture->height);

	glEnable(GL_TEXTURE_2D);	// draw 2d textures

		voxel_buffer_texture[2]->bind();	// load texture units
		glActiveTexture(GL_TEXTURE1);
		voxel_buffer_texture[1]->bind();
		glActiveTexture(GL_TEXTURE2);
		voxel_buffer_texture[0]->bind();
		glActiveTexture(GL_TEXTURE0);
checkGlError();

			voxel_program_combine->enable();
			renderComputationRectangleZ()
			voxel_program_combine->disable();

		glActiveTexture(GL_TEXTURE2);
		voxel_buffer_texture[0]->unbind();
		glActiveTexture(GL_TEXTURE1);
		voxel_buffer_texture[1]->unbind();
		glActiveTexture(GL_TEXTURE0);
		voxel_buffer_texture[2]->unbind();
		checkGlError();

checkGlError();
	voxel_fbo->unbind();

	fbo_write_end();
#else
/************************* SIMPLE VOXELISATION **************************/
/**
 * STEP 1) VOXELISATION
 */

	// initialize values
	glDisable(GL_TEXTURE_2D);

	glClearColor(0, 0, 0, 0);	// clear every face with black color
	glColor4f(1, 1, 1, 1);		// turn on white color for voxelisation

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

checkGlError();
// LOOP FOR Z AXIS (res_x, res_y)
	voxel_buffer_fbo[2]->bind();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	if (modelview_matrix)	glMultMatrixf(modelview_matrix);
	glClear(GL_COLOR_BUFFER_BIT);
	for (int z = 0; z < res_z; z++)
	{
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		// IMPORTANT!
		// we have to use an offset of 0.5 on BOTH sides to avoid diagonal GAPS!!!
		glOrtho(-0.5, 0.5, -0.5, 0.5, 0.5f - (float(z-0.5))/float(res_z) + 0.0005, 0.5 - (float(z+1.5))/float(res_z) - 0.0005);

		// TODO: remove mod and div operations
		glViewport((z % ft_z_mod) * res_x, (z / ft_z_mod) * res_y, res_x, res_y);

		renderObstacle();

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();		glLoadIdentity();
			glCallList(worldBoxDisplayList);
		glPopMatrix();
	}
checkGlError();

// LOOP FOR Y AXIS (res_x, res_z)
	voxel_buffer_fbo[1]->bind();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glRotatef(90, 1, 0, 0);

	if (modelview_matrix)	glMultMatrixf(modelview_matrix);
	glClear(GL_COLOR_BUFFER_BIT);
	for (int y = 0; y < res_y; y++)
	{
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(-0.5, 0.5, 0.5, -0.5, 0.5f - (float(y-0.5))/float(res_y) + 0.0005, 0.5 - (float(y+1.5))/float(res_y) - 0.0005);

		glViewport((y % ft_y_mod) * res_x, (y / ft_y_mod) * res_z, res_x, res_z);

		renderObstacle();

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();		glLoadIdentity();
			glCallList(worldBoxDisplayList);
		glPopMatrix();
	}

// LOOP FOR X AXIS (res_y, res_z)
	voxel_buffer_fbo[0]->bind();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glRotatef(-90, 0, 0, 1);
	glRotatef(-90, 0, 1, 0);
	if (modelview_matrix)	glMultMatrixf(modelview_matrix);

	glClear(GL_COLOR_BUFFER_BIT);
	for (int x = 0; x < res_x; x++)
	{
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(-0.5, 0.5, -0.5, 0.5, 0.5f - (float(x-0.5))/float(res_x) + 0.0005, 0.5 - (float(x+1.5))/float(res_x) - 0.0005);

		glViewport((x % ft_x_mod) * res_y, (x / ft_x_mod) * res_z, res_y, res_z);

		renderObstacle();

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();		glLoadIdentity();
			glCallList(worldBoxDisplayList);
		glPopMatrix();
	}
	voxel_buffer_fbo[0]->unbind();
checkGlError();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	checkGlError();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

// reset viewport
	gui->setup_viewport();

// now, we draw everything to the buffer_Ztex
	voxel_fbo->bind();

checkGlError();
// prepare rendering to texture
	fbo_write_start(voxel_texture->width, voxel_texture->height);

		voxel_buffer_texture[2]->bind();	// load texture units
		glActiveTexture(GL_TEXTURE1);
		voxel_buffer_texture[1]->bind();
		glActiveTexture(GL_TEXTURE2);
		voxel_buffer_texture[0]->bind();
		glActiveTexture(GL_TEXTURE0);
checkGlError();

			voxel_program_combine->enable();
			renderComputationRectangleZ();
			voxel_program_combine->disable();

		glActiveTexture(GL_TEXTURE2);
		voxel_buffer_texture[0]->unbind();
		glActiveTexture(GL_TEXTURE1);
		voxel_buffer_texture[1]->unbind();
		glActiveTexture(GL_TEXTURE0);
		voxel_buffer_texture[2]->unbind();
		checkGlError();

checkGlError();
	voxel_fbo->unbind();

	fbo_write_end();
#endif

	checkGlError();

}


/******************************************
 * COMPUTATE VELOCITY AND DENSITY
 *
 * input:
 * distributions_texture[n]
 *
 * output:
 * velocity_density_texture
 * fluid_fraction_texture: the new fluid fractions
 ******************************************/

void CSimulation::computate_velocity_density()
{
	GLenum buffers_start[2] = {GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT};
	GLenum buffers_finish[2] = {GL_COLOR_ATTACHMENT0_EXT, GL_NONE};

	// initialize values
	fbo_write_start(velocity_density_texture->width, velocity_density_texture->height);
checkGlError();

	velocity_density_fbo->bind();
	velocity_density_fbo->bindFramebufferTexture(buffer_texture, 1);
	glDrawBuffers(2, buffers_start);

checkGlError();

		distributions_texture[0]->bind();
		glActiveTexture(GL_TEXTURE1);
		distributions_texture[1]->bind();
		glActiveTexture(GL_TEXTURE2);
		distributions_texture[2]->bind();
		glActiveTexture(GL_TEXTURE3);
		distributions_texture[3]->bind();
		glActiveTexture(GL_TEXTURE4);
		distributions_texture[4]->bind();

		glActiveTexture(GL_TEXTURE5);
		fluid_mass_texture->bind();
		glActiveTexture(GL_TEXTURE6);
		fluid_fraction_texture->bind();

			velocity_density_program->enable();
			renderComputationRectangleZ();
			velocity_density_program->disable();

		glActiveTexture(GL_TEXTURE6);
		fluid_fraction_texture->unbind();
		glActiveTexture(GL_TEXTURE5);
		fluid_mass_texture->unbind();

		glActiveTexture(GL_TEXTURE4);
		distributions_texture[4]->unbind();
		glActiveTexture(GL_TEXTURE3);
		distributions_texture[3]->unbind();
		glActiveTexture(GL_TEXTURE2);
		distributions_texture[2]->unbind();
		glActiveTexture(GL_TEXTURE1);
		distributions_texture[1]->unbind();
		glActiveTexture(GL_TEXTURE0);
		distributions_texture[0]->unbind();

	glDrawBuffers(2, buffers_finish);
	velocity_density_fbo->unbindFramebufferTexture(1);
	velocity_density_fbo->unbind();

	fbo_write_end();
checkGlError();

	swap(&buffer_fbo, &fluid_fraction_fbo);
	swap(&buffer_texture, &fluid_fraction_texture);
}


/******************************************
 * UPDATE CELL STATE
 *
 * update cell state according to the new fraction texture
 * this compares the new with the old fraction texture and activates
 * or deactivates interface, gas or fluid cells according to the data
 *
 * input:
 * buffer_texture: the recently calculated fluid fractions
 * fluid_fraction_texture: the old fluid fractions
 *
 * output:
 * fluid_fraction_texture: the new fluid fractions
 ******************************************/
void CSimulation::update_cell_state()
{
	/**
	 * FIRST PASS (CHANGE INTERFACE STATE)
	 * 
	 * obstacle and fluid cells are not touched in this pass
	 * 
	 * - INTERFACE cells are changed to FLUID cells if the fluid fraction is over a value
	 * - GAS cells change to INTERFACE cells if an adjacent cell was changed during the step above  
	 */
	fbo_write_start(fluid_fraction_texture->width, fluid_fraction_texture->height);

	buffer_fbo->bind();

		fluid_fraction_texture->bind();	// old fluid fraction to texture unit 0

			update_cell_state_first_program->enable();
			renderComputationRectangleZ();
			update_cell_state_first_program->disable();

		fluid_fraction_texture->unbind();

	buffer_fbo->unbind();
checkGlError();

	/**
	 * SECOND PASS
	 * 
	 * gas and new gas cells (-10) are not changed
	 * 
	 * - INTERFACE CELLS are changed to GAS cells if the fluid fraction is below 0
	 * - FLUID cells change to interface cell if an above mentioned interface cell was converted to a gas cell 
	 */
	fluid_fraction_fbo->bind();

		buffer_texture->bind();

			update_cell_state_second_program->enable();
			renderComputationRectangleZ();
			update_cell_state_second_program->disable();

		buffer_texture->unbind();

	fluid_fraction_fbo->unbind();
checkGlError();


#if HANDLE_GAS

	/**
	 * THIRD PASS
	 * computate velocities for new gas cells
	 */

	fluid_fraction_texture->bind();
	glActiveTexture(GL_TEXTURE1);
	velocity_density_texture->bind();

		buffer_fbo->bind();

			update_cell_state_third_program->enable();
			renderComputationRectangleZ();
			update_cell_state_third_program->disable();

		buffer_fbo->unbind();

	velocity_density_texture->unbind();
	glActiveTexture(GL_TEXTURE0);
	fluid_fraction_texture->unbind();

	swap(&buffer_fbo, &velocity_density_fbo);
	swap(&buffer_texture, &velocity_density_texture);
checkGlError();


	/**
	 * FOURTH PASS
	 * update DF's
	 */
checkGlError();

	glActiveTexture(GL_TEXTURE2);
	velocity_density_texture->bind();
	glActiveTexture(GL_TEXTURE1);
	fluid_fraction_texture->bind();
	glActiveTexture(GL_TEXTURE0);

	for (int i = 0; i < 4; i++)
	{
		buffer2_fbo->bind();

			update_cell_state_fourth_program[i]->enable();

				distributions_texture[i]->bind();
				renderComputationRectangleZ();
				distributions_texture[i]->unbind();

			update_cell_state_fourth_program[i]->disable();

		buffer2_fbo->unbind();

		// swap the textures
		swap(&buffer2_fbo, &distributions_fbo[i]);
		swap(&buffer2_texture, &distributions_texture[i]);
	}

	velocity_density_texture->unbind();
	glActiveTexture(GL_TEXTURE1);
	fluid_fraction_texture->unbind();
	glActiveTexture(GL_TEXTURE0);
checkGlError();

	// render the last texture and update the fluid fraction texture (new gass cell flag to 'gas cell')
	GLenum buffers_start[2] = {GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT};
	GLenum buffers_finish[2] = {GL_COLOR_ATTACHMENT0_EXT, GL_NONE};

	buffer_fbo->bind();
	buffer_fbo->bindFramebufferTexture(buffer2_texture, 1);
	glDrawBuffers(2, buffers_start);

	update_cell_state_fourth_program[4]->enable();

		distributions_texture[4]->bind();
		glActiveTexture(GL_TEXTURE1);
		fluid_fraction_texture->bind();
		glActiveTexture(GL_TEXTURE2);
		velocity_density_texture->bind();

		renderComputationRectangleZ();

		velocity_density_texture->unbind();
		glActiveTexture(GL_TEXTURE1);
		fluid_fraction_texture->unbind();
		glActiveTexture(GL_TEXTURE0);
		distributions_texture[4]->unbind();

	update_cell_state_fourth_program[4]->disable();

	glDrawBuffers(2, buffers_finish);
	buffer_fbo->unbindFramebufferTexture(1);
	buffer_fbo->unbind();

	// swap the textures
	swap(&buffer_fbo, &distributions_fbo[4]);
	swap(&buffer_texture, &distributions_texture[4]);

	swap(&buffer2_fbo, &fluid_fraction_fbo);
	swap(&buffer2_texture, &fluid_fraction_texture);

#endif	// HANDLE_GAS


checkGlError();

	fbo_write_end();
}



/******************************************
 * COLLISION
 *
 * previously, the voxel texture and the velocity_density texture has to be computated!
 ******************************************/
void CSimulation::collide()
{
	// initialize values
	fbo_write_start(velocity_density_texture->width, velocity_density_texture->height);

	glActiveTexture(GL_TEXTURE1);
	velocity_density_texture->bind();
	glActiveTexture(GL_TEXTURE2);
	voxel_texture->bind();
	glActiveTexture(GL_TEXTURE0);

	for (int i = 0; i < 5; i++)	// TODO: collide 4 textures at once
	{
		collision_program[i]->enable();
		buffer_fbo->bind();

			distributions_texture[i]->bind();
			renderComputationRectangleZ();
			distributions_texture[i]->unbind();

		buffer_fbo->unbind();
		collision_program[i]->disable();

		// swap the textures
		swap(&buffer_fbo, &(distributions_fbo[i]));
		swap(&buffer_texture, &(distributions_texture[i]));
	}

	// TODO: perform in/output conditions with rendering 2 rectangles to the flat texture
	buffer_fbo->unbind();

	glActiveTexture(GL_TEXTURE2);
	voxel_texture->unbind();
	glActiveTexture(GL_TEXTURE1);
	velocity_density_texture->unbind();
	glActiveTexture(GL_TEXTURE0);

	fbo_write_end();
}

/******************************************
 * PROPAGATION
 *
 * just the propagation operation,
 * nothing else,
 * no magic,
 * no bunny inside,
 * no beer available
 *
 * input: density_distributions_texture[n], fluid_fraction_texture, fluid_mass_texture
 * output: density_distributions_texture[n], fluid_mass_texture
 ******************************************/

void CSimulation::propagate()
{
	GLenum buffers_start[2] = {GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT};
	GLenum buffers_finish[2] = {GL_COLOR_ATTACHMENT0_EXT, GL_NONE};

	fbo_write_start(buffer_texture->width, buffer_texture->height);
checkGlError();

	glActiveTexture(GL_TEXTURE3);
	velocity_density_texture->bind();
	glActiveTexture(GL_TEXTURE2);
	fluid_fraction_texture->bind();

checkGlError();
	for (int i = 0; i < 5; i++)
	{
		buffer_fbo->bind();
		buffer_fbo->bindFramebufferTexture(buffer2_texture, 1);
		glDrawBuffers(2, buffers_start);	// we have change the drawbuffers for every activation!

		propagation_program[i]->enable();

			glActiveTexture(GL_TEXTURE1);
			fluid_mass_texture->bind();
			glActiveTexture(GL_TEXTURE0);
			distributions_texture[i]->bind();

		renderComputationRectangleZ();

			distributions_texture[i]->unbind();

			glActiveTexture(GL_TEXTURE1);
			fluid_mass_texture->unbind();
			glActiveTexture(GL_TEXTURE0);
			distributions_texture[i]->unbind();

		propagation_program[i]->disable();

		glDrawBuffers(2, buffers_finish);
		buffer_fbo->unbindFramebufferTexture(1);

		buffer_fbo->unbind();

	checkGlError();

		// swap the textures
		swap(&buffer_fbo, &(distributions_fbo[i]));
		swap(&buffer_texture, &(distributions_texture[i]));

		swap(&buffer2_fbo, &fluid_mass_fbo);
		swap(&buffer2_texture, &fluid_mass_texture);

	checkGlError();
	}

	glActiveTexture(GL_TEXTURE3);
	velocity_density_texture->unbind();
	glActiveTexture(GL_TEXTURE2);
	fluid_fraction_texture->unbind();

	glActiveTexture(GL_TEXTURE0);
	distributions_texture[4]->unbind();

	fbo_write_end();
}


/******************************************
 * COMPUTATIONS (public function)
 *
 * perform the computations
 * 1. voxelisation
 * 2. velocity + density + fluid fractions
 * 3. collide
 * 4. propagate + mass exchange
 */

#if DEBUG_SHADER && false
void CSimulation::computate()
{
static int flag = 0;

	fbo_write_save_state();

	glColor4f(1, 1, 1, 1);

	switch(flag)
	{
		case 0:
			cout << "DEBUG: voxelisation" << endl;
			voxelisation();			// VOXELISATION
			break;

		case 1:
			cout << "DEBUG: collide" << endl;
			collide();			// COLLIDE
			break;

		case 2:
			cout << "DEBUG: propagate" << endl;
			propagate();			// PROPAGATE
			break;
	
		case 3:
			cout << "DEBUG: velocity - density" << endl;
			computate_velocity_density();	// VELOCITY / DENSITY + FLUID FRACTION
			break;

		case 4:
			cout << "DEBUG: update cell state" << endl;
			update_cell_state();		// UPDATE CELL STATE
			break;

		default:
			cout << "DEBUG: invalid computation flag" << endl;

	}
	flag = (flag+1)%5;

	fbo_write_restore_state();
}

#else

#if GET_SIMULATION_TICKS
void CSimulation::computate()
{
	int ticks, ticks_new;

	fbo_write_save_state();

	glColor4f(1, 1, 1, 1);

	glFinish();	// call finish to get correct statistics
	ticks = gui->get_ticks();
		voxelisation();			// VOXELISATION
	glFinish();
	ticks_new = gui->get_ticks();
	ticks_voxelisation += ticks_new - ticks;


	ticks = gui->get_ticks();
		collide();			// COLLIDE
	ticks_new = gui->get_ticks();
	ticks_collide += ticks_new - ticks;

	ticks = gui->get_ticks();
		propagate();			// PROPAGATE
	glFinish();
	ticks_new = gui->get_ticks();
	ticks_propagate += ticks_new - ticks;

	ticks = gui->get_ticks();
		computate_velocity_density();	// VELOCITY / DENSITY + FLUID FRACTION
	glFinish();
	ticks_new = gui->get_ticks();
	ticks_velocity_density += ticks_new - ticks;

	ticks = gui->get_ticks();
		update_cell_state();		// UPDATE CELL STATE
	glFinish();
	ticks_new = gui->get_ticks();
	ticks_update_state += ticks_new - ticks;

	fbo_write_restore_state();
}

#else	// GET_SIMULATION_TICKS

void CSimulation::computate()
{
	fbo_write_save_state();

	glColor4f(1, 1, 1, 1);

		voxelisation();			// VOXELISATION
		collide();			// COLLIDE
		propagate();			// PROPAGATE
		computate_velocity_density();	// VELOCITY / DENSITY + FLUID FRACTION
		update_cell_state();		// UPDATE CELL STATE

	fbo_write_restore_state();
}

#endif	// GET_SIMULATION_TICKS

#endif



CSimulation::~CSimulation()
{
	if (!ready)	return;

	int i;
	delete clear_program;

	for (i = 0; i < 5; i++)
	{
		delete distributions_fbo[i];
		delete distributions_texture[i];
	}

	delete fluid_fraction_texture;
	delete fluid_fraction_fbo;

	delete fluid_mass_texture;
	delete fluid_mass_fbo;

	delete buffer_texture;

	delete velocity_density_fbo;
	delete velocity_density_texture;

	delete voxel_program_combine;

	delete voxel_fbo;
	delete voxel_buffer_fbo[0];
	delete voxel_buffer_fbo[1];
	delete voxel_buffer_fbo[2];

	delete voxel_texture;
	delete voxel_buffer_texture[0];
	delete voxel_buffer_texture[1];
	delete voxel_buffer_texture[2];

	for (i = 0; i < 5; i++)
		delete propagation_program[i];

	for (i = 0; i < 5; i++)
		delete collision_program[i];

	delete update_cell_state_second_program;
	delete update_cell_state_first_program;

	delete velocity_density_program;


#if CSIMULATION_USE_DEEPTH_PEELING
	glDeleteQueries(1, &voxel_query_object);
#endif

	glDeleteLists(objFileVoxelisationDisplayList, 1);
	glDeleteLists(sphereVoxelisationDisplayList, 1);
	glDeleteLists(worldBoxDisplayList, 1);
	glDeleteLists(write_to_texture_call_list_x, 1);
	glDeleteLists(write_to_texture_call_list_y, 1);
	glDeleteLists(write_to_texture_call_list_z, 1);
}
