/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __CSIMULATION_H
#define __CSIMULATION_H

#include "libviz/CObjFile.hpp"
#include "libviz/CFbo.hpp"
#include "libviz/CTexture.hpp"
#include "libviz/CProgram.hpp"
#include "CGui.hpp"
#include "libmath/CVector.hh"

#define DEBUG	1

/**
 * \file CSimulation.hpp
 * LBM Simulation on GPU
 *
 * \class CSimulation
 * \brief this class handles everything for the LBM simulation on gpus
 *
 * starting from creation of textures up to the simulation
 *
 *

 * <pre>
 * D3Q19 Vector enumerations + directions
 *
 * GPU Gems 2 / Page 752
 *
 * 		| R         G           B          A
 * --------------------------------------------------------
 * tex0:	| f(1,0,0), f(-1,0,0),  f(0,1,0),  f(0,-1,0)
 * tex1:	| f(1,1,0), f(-1,-1,0), f(1,-1,0), f(-1,1,0)
 * tex2:	| f(1,0,1), f(-1,0,-1), f(1,0,-1), f(-1,0,1)
 * tex3:	| f(0,1,1), f(0,-1,-1), f(0,1,-1), f(0,-1,1)
 * tex4:	| f(0,0,1), f(0,0,-1),  f(0,0,0),  voxel
 *		| 
 * texv:	| vx,       vy,         vz,        phi
 *
 * <pre>
 *
 *
 *       14,12
 *     15  |  18
 *       \ | /
 *        \|/
 * 7,9----- ----- 8,10
 *        /|\
 *       / | \
 *     17  |  16
 *       11,13
 *
 * voxel:	if greater than 0, this node has an obstacle voxel
 * vx, vy, vz:	the velocity components
 * phi:		the density
 *
 * ---
 *
 * To simulate open surfaces, we need an additional textures
 *
 * mass in cell texture (mass_texture):
 *	R: mass in cell
 *
 * fluid fraction texture (fluid_fraction_texture):
 * 	R: fluid fraction e = m/density
 * 	G: cell state:
 * 		-2: obstacle
 * 		-1: gas
 * 		0: interface
 * 		1: fluid
 *
 * 		update_cell_state.frag:
 * 		-10: transition state from interface cell to gas cell
 * 		+10: transition state from interface cell to fluid cell
 *
 * 	the borders are always seen as obstacles and thus as empty cells
 * 	so the neighbored cells of obstacles and the borders can be only interface or empty cells
 *
 * the flags in the cells change according to the fluid fraction:
 *	< 0:	=> empty cell
 *	> 1:	=> fluid cell
 *	else:	=> gas cell
 *
 *
 * </pre>
 */


#define CSIMULATION_SPHERE_LOD	50

class CSimulation
{
	CGui *gui;			///< handler to gui CGui class
	void collide();		///< collision step (1)
	void propagate();	///< propagation step (2)
	void voxelisation();	///< voxelisation step (3)
	void computate_velocity_density();
	void update_cell_state();

	float d_viscosity;		///< dimensional parameters
	float d_length_x;
	vec3f d_gravitation;
	float d_cell_size;
	float d_timestep;

	float viscosity;		///< dimensionless parameters for simulation
	float cell_size;		///< parameters
	float timestep;
	float omega;			///< inverse of tau
	float tau;

public:
	vec3f gravitation;	/// direction + strength of gravitation
	void setGravitation(vec3f &new_gravitation);
	inline vec3f getGravitation()	{	return d_gravitation;	};

	/**
	 * resolutions of textures in different dimensions
	 * ft_A_width, ft_A_height: resolution of flat textures produced through axis A
	 */
	int ft_x_width, ft_x_height, ft_x_mod;
	int ft_y_width, ft_y_height, ft_y_mod;
	int ft_z_width, ft_z_height, ft_z_mod;

	// resolution of flow domain
	int res_x, res_y, res_z;		///< resolution of grid
	CVector<3,float> fres;			///< resolution as float array
	CVector<3,float> finv_res;		///< inverted resolution as float array

	// stored modelview matrix to move boundary object within flow domain
	float *modelview_matrix;

	// voxelisation stuff
	// framebuffer objects, textures, etc.
	CFbo *voxel_fbo;		CTexture *voxel_texture;
	CFbo *voxel_buffer_fbo[3];	CTexture *voxel_buffer_texture[3];	// x,y,z voxelisation with slices
	CProgram *voxel_program_combine;

#if	CSIMULATION_USE_DEEPTH_PEELING
	GLuint voxel_query_object;			// object for glBegin/EndQuery
	GLuint voxel_peeling_depth[3];		// depth (rendering times) of voxelisation
	CProgram *voxel_peeling_program[3];	// program to perform depth peeling operation
	CProgram *voxel_peeling_program_first;	// peeling program for first render call
	CTexture *voxel_buffer_depth_texture[3];	///< depth buffer if peeling is used
#endif

#if	CSIMULATION_USE_DEEPTH_PEELING_2_DEPTH_BUFFER
	CTexture *voxel_buffer_depth_texture_[3];
#endif

// lbm specific textures/fbos

	/// storage for distribution values
	CFbo *distributions_fbo[5];	CTexture *distributions_texture[5];

	/**
	 * the mass is modified in the propagation step
	 */
	CFbo *fluid_mass_fbo;		CTexture *fluid_mass_texture;
	CProgram *fluid_fraction_program;

	// velocity and dentity
	CFbo *velocity_density_fbo;	CTexture *velocity_density_texture;
	CProgram *velocity_density_program;

	/**
	 * we use fluid fraction texture also to differ between fluid, interface and empty cells!
	 * this texture is calculated during the collision step (mass texture has to be bounded, too
	 * fluid > 1 >= interface > 0 >= empty
	 *
	 * fluid_faction = mass / density
	 */
	CFbo *fluid_fraction_fbo;	CTexture *fluid_fraction_texture;

private:
	/// shader: update cell state
	CProgram *update_cell_state_first_program;
	CProgram *update_cell_state_second_program;
	CProgram *update_cell_state_third_program;
	CProgram *update_cell_state_fourth_program[5];

	// shader: collision & propagation programs
	CProgram *collision_program[5], *propagation_program[5];

	GLuint objFileVoxelisationDisplayList;
	GLuint sphereVoxelisationDisplayList;	// voxelisation list for a sphere (alternative for objFile)
	void renderObstacle();

	GLuint worldBoxDisplayList;	// this list draws a quad around the flow domain

	// buffer
	CFbo *buffer_fbo;	CTexture *buffer_texture;	///< buffer texture which has the same dimensions as the distribution textures
	CFbo *buffer2_fbo;	CTexture *buffer2_texture;

public:
	CObjFile *objFile;		///< ObjectFile handler to draw content of loaded objectfile

	// texture clear program
	CProgram *clear_program;

	bool ready;


	CSimulation(	CObjFile *p_objFile,		// obj file handler to render mesh (boundary)
			CGui *p_gui,			// pointer to gui interface
			int res_x = 50, int res_y = 50, int res_z = 50,	// resolution of flow domain
			float p_viscosity = 0.012f,		// viscosity of fluid within domain [m^2/s]
			float p_x_length = 1.0,		// length of one size of simulation box [m]
			float p_gravitation = 9.81	// [m/s^2]
			);
	~CSimulation();
	void reset();			///< reset the simulation
	void computate();		///< perform one computation step


	inline void renderComputationRectangleX();
	inline void renderComputationRectangleY();
	inline void renderComputationRectangleZ();

/**
 * used by simulation gui
 */

	// writing to fbo texture call list to make things faster
	GLuint write_to_texture_call_list_x;
	GLuint write_to_texture_call_list_y;
	GLuint write_to_texture_call_list_z;

	int fluidObstacleType;	/// 0 for mesh, 1 for sphere, otherwise rendering an obstacle is disabled

#if GET_SIMULATION_TICKS
	int ticks_voxelisation, ticks_collide, ticks_propagate, ticks_velocity_density, ticks_update_state;

	void ticks_reset()
	{
		ticks_voxelisation = 0;
		ticks_collide = 0;
		ticks_propagate = 0;
		ticks_velocity_density = 0;
		ticks_update_state = 0;
	}
#endif

inline void fbo_write_start(int width, int height)
{
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, 1, 0, 1, -1, 1);

	glViewport(0, 0, width, height);
checkGlError();
}

inline void fbo_write_end()
{
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	gui->setup_viewport();
checkGlError();
}


void inline swap(CTexture **t1, CTexture **t2)
{
	CTexture *a = *t1;
	*t1 = *t2;
	*t2 = a;
}

void inline swap(CFbo **fbo1, CFbo **fbo2)
{
	CFbo *a = *fbo1;
	*fbo1 = *fbo2;
	*fbo2 = a;
}



};

#endif
