/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef TMATRIX4_HH
#define TMATRIX4_HH

/**
 * CHANGELOG:
 *
 * 2008-10-28: martin schreiber
 * 	created
 *
 */

#include <limits>
#include <iostream>
#include "CMath.hh"
#include "CVector.hh"

/// 4x4 matrix handler
template <typename T>
class CMatrix4
{
public:
	// vector data
	CVector<4,T> matrix[4];

	/*******************
	 * CONSTRUCTURS
	 *******************/
	inline CMatrix4()
	{
		setZero();
	}

	/**
	 * set values to 0
	 */
	inline void setZero()
	{
		matrix[0][0] = (T)0;
		matrix[0][1] = (T)0;
		matrix[0][2] = (T)0;
		matrix[0][3] = (T)0;
		matrix[1][0] = (T)0;
		matrix[1][1] = (T)0;
		matrix[1][2] = (T)0;
		matrix[1][3] = (T)0;
		matrix[2][0] = (T)0;
		matrix[2][1] = (T)0;
		matrix[2][2] = (T)0;
		matrix[2][3] = (T)0;
		matrix[3][0] = (T)0;
		matrix[3][1] = (T)0;
		matrix[3][2] = (T)0;
		matrix[3][3] = (T)0;
	}


	/**
	 * copy data from matrix m
	 * \param m	source matrix
	 */
	inline CMatrix4<T>& operator=(CMatrix4<T> m)
	{
		matrix[0] = m[0];
		matrix[1] = m[1];
		matrix[2] = m[2];
		matrix[3] = m[3];

		return *this;
	};

	/**
	 * load identity matrix
	 */
	inline void loadIdentity()
	{
		matrix[0][0] = (T)1;
		matrix[0][1] = (T)0;
		matrix[0][2] = (T)0;
		matrix[0][3] = (T)0;
		matrix[1][0] = (T)0;
		matrix[1][1] = (T)1;
		matrix[1][2] = (T)0;
		matrix[1][3] = (T)0;
		matrix[2][0] = (T)0;
		matrix[2][1] = (T)0;
		matrix[2][2] = (T)1;
		matrix[2][3] = (T)0;
		matrix[3][0] = (T)0;
		matrix[3][1] = (T)0;
		matrix[3][2] = (T)0;
		matrix[3][3] = (T)1;
	}

	/**
	 * matrix vector product
	 * use 1 as 4th component of input vector v
	 * avoid using parameters by references because this cannot be handled by
	 * subsequenced  orders like a+(o*i)
	 */
	inline CVector<3,T>	operator*(CVector<3,T> v)
	{
		return CVector<3,T>(	matrix[0][0]*v[0] + matrix[0][1]*v[1] + matrix[0][2]*v[2] + matrix[0][3],
					matrix[1][0]*v[0] + matrix[1][1]*v[1] + matrix[1][2]*v[2] + matrix[1][3],
					matrix[2][0]*v[0] + matrix[2][1]*v[1] + matrix[2][2]*v[2] + matrix[2][3]
				);
	}

	/**
	 * matrix vector product
	 */
	inline CVector<4,T>	operator*(CVector<4,T> v)
	{
		return CVector<4,T>(	matrix[0][0]*v[0] + matrix[0][1]*v[1] + matrix[0][2]*v[2] + matrix[0][3]*v[3],
					matrix[1][0]*v[0] + matrix[1][1]*v[1] + matrix[1][2]*v[2] + matrix[1][3]*v[3],
					matrix[2][0]*v[0] + matrix[2][1]*v[1] + matrix[2][2]*v[2] + matrix[2][3]*v[3],
					matrix[3][0]*v[0] + matrix[3][1]*v[1] + matrix[3][2]*v[2] + matrix[3][3]*v[3]
				);
	}

	/**
	 * multiply this matrix with m2 and return result
	 * \return this->matrix * m2
	 */
	inline CMatrix4<T>	operator*(CMatrix4<T> &m2)
	{
		CMatrix4 m;

		m[0][0] = matrix[0][0]*m2[0][0] + matrix[0][1]*m2[1][0] + matrix[0][2]*m2[2][0] + matrix[0][3]*m2[3][0];
		m[0][1] = matrix[0][0]*m2[0][1] + matrix[0][1]*m2[1][1] + matrix[0][2]*m2[2][1] + matrix[0][3]*m2[3][1];
		m[0][2] = matrix[0][0]*m2[0][2] + matrix[0][1]*m2[1][2] + matrix[0][2]*m2[2][2] + matrix[0][3]*m2[3][2];
		m[0][3] = matrix[0][0]*m2[0][3] + matrix[0][1]*m2[1][3] + matrix[0][2]*m2[2][3] + matrix[0][3]*m2[3][3];

		m[1][0] = matrix[1][0]*m2[0][0] + matrix[1][1]*m2[1][0] + matrix[1][2]*m2[2][0] + matrix[1][3]*m2[3][0];
		m[1][1] = matrix[1][0]*m2[0][1] + matrix[1][1]*m2[1][1] + matrix[1][2]*m2[2][1] + matrix[1][3]*m2[3][1];
		m[1][2] = matrix[1][0]*m2[0][2] + matrix[1][1]*m2[1][2] + matrix[1][2]*m2[2][2] + matrix[1][3]*m2[3][2];
		m[1][3] = matrix[1][0]*m2[0][3] + matrix[1][1]*m2[1][3] + matrix[1][2]*m2[2][3] + matrix[1][3]*m2[3][3];

		m[2][0] = matrix[2][0]*m2[0][0] + matrix[2][1]*m2[1][0] + matrix[2][2]*m2[2][0] + matrix[2][3]*m2[3][0];
		m[2][1] = matrix[2][0]*m2[0][1] + matrix[2][1]*m2[1][1] + matrix[2][2]*m2[2][1] + matrix[2][3]*m2[3][1];
		m[2][2] = matrix[2][0]*m2[0][2] + matrix[2][1]*m2[1][2] + matrix[2][2]*m2[2][2] + matrix[2][3]*m2[3][2];
		m[2][3] = matrix[2][0]*m2[0][3] + matrix[2][1]*m2[1][3] + matrix[2][2]*m2[2][3] + matrix[2][3]*m2[3][3];

		m[3][0] = matrix[3][0]*m2[0][0] + matrix[3][1]*m2[1][0] + matrix[3][2]*m2[2][0] + matrix[3][3]*m2[3][0];
		m[3][1] = matrix[3][0]*m2[0][1] + matrix[3][1]*m2[1][1] + matrix[3][2]*m2[2][1] + matrix[3][3]*m2[3][1];
		m[3][2] = matrix[3][0]*m2[0][2] + matrix[3][1]*m2[1][2] + matrix[3][2]*m2[2][2] + matrix[3][3]*m2[3][2];
		m[3][3] = matrix[3][0]*m2[0][3] + matrix[3][1]*m2[1][3] + matrix[3][2]*m2[2][3] + matrix[3][3]*m2[3][3];

		return m;
	}

private:
	// switch 4 colums and vector elements
	inline void solve_switch_row(CMatrix4<T> &m, CVector<4,T> &b, int ra, int rb)
	{
		T a;
		// switch 1st and 3nd row
		a = m[ra][0];	m[ra][0] = m[rb][0];	m[rb][0] = a;
		a = m[ra][1];	m[ra][1] = m[rb][1];	m[rb][1] = a;
		a = m[ra][2];	m[ra][2] = m[rb][2];	m[rb][2] = a;
		a = m[ra][3];	m[ra][3] = m[rb][3];	m[rb][3] = a;
		a = b[ra];	b[ra] = b[rb];		b[rb] = a;
	}
	// switch the rightmost 3 columns and vector elements
	inline void solve_switch_row3(CMatrix4<T> &m, CVector<4,T> &b, int ra, int rb)
	{
		T a;
		a = m[ra][1];	m[ra][1] = m[rb][1];	m[rb][1] = a;
		a = m[ra][2];	m[ra][2] = m[rb][2];	m[rb][2] = a;
		a = m[ra][3];	m[ra][3] = m[rb][3];	m[rb][3] = a;
		a = b[ra];	b[ra] = b[rb];		b[rb] = a;
	}

	// switch the rightmost 2 columns and vector elements
	inline void solve_switch_row2(CMatrix4<T> &m, CVector<4,T> &b, int ra, int rb)
	{
		T a;
		// switch 1st and 3nd row
		a = m[ra][2];	m[ra][2] = m[rb][2];	m[rb][2] = a;
		a = m[ra][3];	m[ra][3] = m[rb][3];	m[rb][3] = a;
		a = b[ra];	b[ra] = b[rb];		b[rb] = a;
	}
public:
#define SMALL_EPSILON	0.0001
	/**
	 * solve matrix system
	 *
	 * M*f = b
	 *
	 * return f as result
	 *
	 * f[0] is set to nan if no unique solution exists!
	 * the matrix this->matrix is not modified!!!
	 *
	 * \param b	vector from right side of equation
	 */
	inline CVector<4,T> solve(CVector<4,T> b)
	{
		CMatrix4<T> m = *this;
		T a;

		// use gaussian elimination

		// TODO: if m[0][0] is 0, switch with a row where m[][0] is not zero or smaller than a specific value
		if (fabs(m[0][0]) < SMALL_EPSILON)
		{
			if (fabs(m[1][0]) < SMALL_EPSILON)
			{
				if (fabs(m[2][0]) < SMALL_EPSILON)
				{
					if (fabs(m[3][0]) < SMALL_EPSILON)
					{
						// first row seems to be 0 => infinite solutions for b[0]
//						std::cerr << "MATRIX NOT SOLVABLE (ROW1)!" << std::endl;
						return CVector<4,T>(std::numeric_limits<float>::quiet_NaN());
					}

					// switch 1st and 3rd row
					solve_switch_row(m, b, 0, 3);
				}

				// switch 1st and 3rd row
				solve_switch_row(m, b, 0, 2);
			}
			else
			{
				// switch 1st and 2nd row
				solve_switch_row(m, b, 0, 1);
			}
		}

		/************************
		 * ROW: 1
		 */

		a =1.0/m[0][0];	// normalize 1st row
		m[0][1] *= a;
		m[0][2] *= a;
		m[0][3] *= a;
		b[0] *= a;

		// handle 2nd row
//		m[1][0] = 0;	// use m[1][0] as pivot element!
		m[1][1] -= m[1][0]*m[0][1];
		m[1][2] -= m[1][0]*m[0][2];
		m[1][3] -= m[1][0]*m[0][3];
		b[1] -= m[1][0]*b[0];

		// handle 3rd row
//		m[2][0] = 0;	// use m[2][0] as pivot element!
		m[2][1] -= m[2][0]*m[0][1];
		m[2][2] -= m[2][0]*m[0][2];
		m[2][3] -= m[2][0]*m[0][3];
		b[2] -= m[2][0]*b[0];

		// handle 4th row
//		m[2][0] = 0;	// use m[2][0] as pivot element!
		m[3][1] -= m[3][0]*m[0][1];
		m[3][2] -= m[3][0]*m[0][2];
		m[3][3] -= m[3][0]*m[0][3];
		b[3] -= m[3][0]*b[0];

		/************************
		 * ROW: 2
		 *
		 * now we have 2 zero entries at m[1][0] and m[2][0]
		 */

		if (fabs(m[1][1]) < SMALL_EPSILON)
		{
			if (fabs(m[2][1]) < SMALL_EPSILON)
			{
				if (fabs(m[3][1]) < SMALL_EPSILON)
				{
					// second element of row seems to be 0 => infinite solutions for b[1]
	//				std::cerr << "MATRIX NOT SOLVABLE (ROW2)!" << std::endl;
					return CVector<4,T>(std::numeric_limits<float>::quiet_NaN());
				}

				// switch 1st and 3rd row
				solve_switch_row3(m, b, 1, 3);
			}

			// switch 1st and 3rd row
			solve_switch_row3(m, b, 1, 2);
		}

		a = 1.0/m[1][1];	// normalize 2nd row
//		m[1][1] = (T)1;
		m[1][2] *= a;
		m[1][3] *= a;
		b[1] *= a;

		// handle 3rd row
//		m[2][1] = 0;	// use m[2][1] as pivot element!
//		m[2][1] -= m[2][1]*m[1][1];
		m[2][2] -= m[2][1]*m[1][2];
		m[2][3] -= m[2][1]*m[1][3];
		b[2] -= m[2][1]*b[1];

		// handle 4th row
//		m[3][1] = 0;	// use m[2][1] as pivot element!
//		m[3][1] -= m[3][1]*m[1][1];
		m[3][2] -= m[3][1]*m[1][2];
		m[3][3] -= m[3][1]*m[1][3];
		b[3] -= m[3][1]*b[1];

		/************************
		 * ROW: 3
		 */

		if (fabs(m[2][2]) < SMALL_EPSILON)
		{
			if (fabs(m[3][2]) < SMALL_EPSILON)
				return CVector<4,T>(std::numeric_limits<float>::quiet_NaN());

			// switch 1st and 2rd row
			solve_switch_row2(m, b, 2, 3);
		}

		a = 1.0/m[2][2];	// normalize 3nd row
//		m[2][2] = (T)1;
		m[2][3] *= a;
		b[2] *= a;

		// handle 4th row
//		m[3][1] = 0;	// use m[2][1] as pivot element!
//		m[3][2] -= m[3][1]*m[2][2];
		m[3][3] -= m[3][2]*m[2][3];
		b[3] -= m[3][2]*b[2];

		/************************
		 * ROW: 4
		 */

		if (fabs(m[3][3]) < SMALL_EPSILON)
			return CVector<4,T>(std::numeric_limits<float>::quiet_NaN());

//		a = 1.0/m[3][3];	// normalize 4th row
//		m[2][2] = (T)1;
//		m[2][3] *= a;
		b[3] /= m[3][3];


		/************************
		 * eliminate UPPER DIAGONAL
		 */

		// use 4th row
		b[2] -= b[3]*m[2][3];
		b[1] -= b[3]*m[1][3];
		b[0] -= b[3]*m[0][3];

		// use 3rd row
		b[1] -= b[2]*m[1][2];
//		m[1][2] = (T)0;
		b[0] -= b[2]*m[0][2];
//		m[0][2] = (T)0;

		// use 2nd row
		b[0] -= b[1]*m[0][1];
//		m[0][1] = (T)0;

		return b;
	}

private:
	// switch 4 colums and vector elements
	inline void inverse_solve_switch_row(CMatrix4<T> &m, CMatrix4<T> &m2, int ra, int rb)
	{
		T a;
		// switch 1st and 3nd row
		a = m[ra][0];	m[ra][0] = m[rb][0];	m[rb][0] = a;
		a = m[ra][1];	m[ra][1] = m[rb][1];	m[rb][1] = a;
		a = m[ra][2];	m[ra][2] = m[rb][2];	m[rb][2] = a;
		a = m[ra][3];	m[ra][3] = m[rb][3];	m[rb][3] = a;

		a = m2[ra][0];	m2[ra][0] = m2[rb][0];	m2[rb][0] = a;
		a = m2[ra][1];	m2[ra][1] = m2[rb][1];	m2[rb][1] = a;
		a = m2[ra][2];	m2[ra][2] = m2[rb][2];	m2[rb][2] = a;
		a = m2[ra][3];	m2[ra][3] = m2[rb][3];	m2[rb][3] = a;
	}
	// switch the rightmost 3 columns and vector elements
	inline void inverse_solve_switch_row3(CMatrix4<T> &m, CMatrix4<T> &m2, int ra, int rb)
	{
		T a;
		a = m[ra][1];	m[ra][1] = m[rb][1];	m[rb][1] = a;
		a = m[ra][2];	m[ra][2] = m[rb][2];	m[rb][2] = a;
		a = m[ra][3];	m[ra][3] = m[rb][3];	m[rb][3] = a;

		a = m2[ra][1];	m2[ra][1] = m2[rb][1];	m2[rb][1] = a;
		a = m2[ra][2];	m2[ra][2] = m2[rb][2];	m2[rb][2] = a;
		a = m2[ra][3];	m2[ra][3] = m2[rb][3];	m2[rb][3] = a;
	}

	// switch the rightmost 2 columns and vector elements
	inline void inverse_solve_switch_row2(CMatrix4<T> &m, CMatrix4<T> &m2, int ra, int rb)
	{
		T a;
		// switch 1st and 3nd row
		a = m[ra][2];	m[ra][2] = m[rb][2];	m[rb][2] = a;
		a = m[ra][3];	m[ra][3] = m[rb][3];	m[rb][3] = a;

		a = m2[ra][2];	m2[ra][2] = m2[rb][2];	m2[rb][2] = a;
		a = m2[ra][3];	m2[ra][3] = m2[rb][3];	m2[rb][3] = a;
	}

public:
	/**
	 * compute the inverse of a matrix
	 */
	inline bool compute_inverse(CMatrix4<T> &inv_matrix)
	{
		CMatrix4<T> m = *this;
		inv_matrix.loadIdentity();
		T a;

		// use gaussian elimination

		// TODO: if m[0][0] is 0, switch with a row where m[][0] is not zero or smaller than a specific value
		if (fabs(m[0][0]) < SMALL_EPSILON)
		{
			if (fabs(m[1][0]) < SMALL_EPSILON)
			{
				if (fabs(m[2][0]) < SMALL_EPSILON)
				{
					if (fabs(m[3][0]) < SMALL_EPSILON)
					{
						// first row seems to be 0 => infinite solutions for b[0]
//						std::cerr << "MATRIX NOT SOLVABLE (ROW1)!" << std::endl;
						return false;
					}

					// switch 1st and 3rd row
					inverse_solve_switch_row(m, inv_matrix, 0, 3);
				}

				// switch 1st and 3rd row
				inverse_solve_switch_row(m, inv_matrix, 0, 2);
			}
			else
			{
				// switch 1st and 2nd row
				inverse_solve_switch_row(m, inv_matrix, 0, 1);
			}
		}

		/************************
		 * ROW: 1
		 */

		a =1.0/m[0][0];	// normalize 1st row
		m[0][1] *= a;
		m[0][2] *= a;
		m[0][3] *= a;

		inv_matrix[0][0] *= a;
		inv_matrix[0][1] *= a;
		inv_matrix[0][2] *= a;
		inv_matrix[0][3] *= a;

		// handle 2nd row
//		m[1][0] = 0;	// use m[1][0] as pivot element!
		m[1][1] -= m[1][0]*m[0][1];
		m[1][2] -= m[1][0]*m[0][2];
		m[1][3] -= m[1][0]*m[0][3];

		inv_matrix[1][0] -= m[1][0]*inv_matrix[0][0];
		inv_matrix[1][1] -= m[1][0]*inv_matrix[0][1];
		inv_matrix[1][2] -= m[1][0]*inv_matrix[0][2];
		inv_matrix[1][3] -= m[1][0]*inv_matrix[0][3];

		// handle 3rd row
//		m[2][0] = 0;	// use m[2][0] as pivot element!
		m[2][1] -= m[2][0]*m[0][1];
		m[2][2] -= m[2][0]*m[0][2];
		m[2][3] -= m[2][0]*m[0][3];

		inv_matrix[2][0] -= m[2][0]*inv_matrix[0][0];
		inv_matrix[2][1] -= m[2][0]*inv_matrix[0][1];
		inv_matrix[2][2] -= m[2][0]*inv_matrix[0][2];
		inv_matrix[2][3] -= m[2][0]*inv_matrix[0][3];

		// handle 4th row
//		m[2][0] = 0;	// use m[2][0] as pivot element!
		m[3][1] -= m[3][0]*m[0][1];
		m[3][2] -= m[3][0]*m[0][2];
		m[3][3] -= m[3][0]*m[0][3];

		inv_matrix[3][0] -= m[3][0]*inv_matrix[0][0];
		inv_matrix[3][1] -= m[3][0]*inv_matrix[0][1];
		inv_matrix[3][2] -= m[3][0]*inv_matrix[0][2];
		inv_matrix[3][3] -= m[3][0]*inv_matrix[0][3];

		/************************
		 * ROW: 2
		 *
		 * now we have 2 zero entries at m[1][0] and m[2][0]
		 */

		if (fabs(m[1][1]) < SMALL_EPSILON)
		{
			if (fabs(m[2][1]) < SMALL_EPSILON)
			{
				if (fabs(m[3][1]) < SMALL_EPSILON)
				{
					// second element of row seems to be 0 => infinite solutions for b[1]
					return false;
				}

				// switch 1st and 3rd row
				inverse_solve_switch_row3(m, inv_matrix, 1, 3);
			}

			// switch 1st and 3rd row
			inverse_solve_switch_row3(m, inv_matrix, 1, 2);
		}

		a = 1.0/m[1][1];	// normalize 2nd row
//		m[1][1] = (T)1;
		m[1][2] *= a;
		m[1][3] *= a;

		inv_matrix[1][0] *= a;
		inv_matrix[1][1] *= a;
		inv_matrix[1][2] *= a;
		inv_matrix[1][3] *= a;

		// handle 3rd row
//		m[2][1] = 0;	// use m[2][1] as pivot element!
//		m[2][1] -= m[2][1]*m[1][1];
		m[2][2] -= m[2][1]*m[1][2];
		m[2][3] -= m[2][1]*m[1][3];

		inv_matrix[2][0] -= m[2][1]*inv_matrix[1][0];
		inv_matrix[2][1] -= m[2][1]*inv_matrix[1][1];
		inv_matrix[2][2] -= m[2][1]*inv_matrix[1][2];
		inv_matrix[2][3] -= m[2][1]*inv_matrix[1][3];

		// handle 4th row
//		m[3][1] = 0;	// use m[2][1] as pivot element!
//		m[3][1] -= m[3][1]*m[1][1];
		m[3][2] -= m[3][1]*m[1][2];
		m[3][3] -= m[3][1]*m[1][3];

		inv_matrix[3][0] -= m[3][1]*inv_matrix[1][0];
		inv_matrix[3][1] -= m[3][1]*inv_matrix[1][1];
		inv_matrix[3][2] -= m[3][1]*inv_matrix[1][2];
		inv_matrix[3][3] -= m[3][1]*inv_matrix[1][3];

		/************************
		 * ROW: 3
		 */

		if (fabs(m[2][2]) < SMALL_EPSILON)
		{
			if (fabs(m[3][2]) < SMALL_EPSILON)
				return false;

			// switch 1st and 2rd row
			inverse_solve_switch_row2(m, inv_matrix, 2, 3);
		}

		a = 1.0/m[2][2];	// normalize 3nd row
//		m[2][2] = (T)1;
		m[2][3] *= a;

		inv_matrix[2][0] *= a;
		inv_matrix[2][1] *= a;
		inv_matrix[2][2] *= a;
		inv_matrix[2][3] *= a;

		// handle 4th row
//		m[3][1] = 0;	// use m[2][1] as pivot element!
//		m[3][2] -= m[3][1]*m[2][2];
		m[3][3] -= m[3][2]*m[2][3];

		inv_matrix[3][0] -= m[3][2]*inv_matrix[2][0];
		inv_matrix[3][1] -= m[3][2]*inv_matrix[2][1];
		inv_matrix[3][2] -= m[3][2]*inv_matrix[2][2];
		inv_matrix[3][3] -= m[3][2]*inv_matrix[2][3];

		/************************
		 * ROW: 4
		 */

		if (fabs(m[3][3]) < SMALL_EPSILON)
			return false;

		a = 1.0/m[3][3];	// normalize 4th row
//		m[2][2] = (T)1;
//		m[2][3] *= a;

		inv_matrix[3][0] *= a;
		inv_matrix[3][1] *= a;
		inv_matrix[3][2] *= a;
		inv_matrix[3][3] *= a;


		/************************
		 * eliminate UPPER DIAGONAL
		 */

		// use 4th row to eliminate upper diagonal
		inv_matrix[2][0] -= m[2][3]*inv_matrix[3][0];
		inv_matrix[2][1] -= m[2][3]*inv_matrix[3][1];
		inv_matrix[2][2] -= m[2][3]*inv_matrix[3][2];
		inv_matrix[2][3] -= m[2][3]*inv_matrix[3][3];

		inv_matrix[1][0] -= m[1][3]*inv_matrix[3][0];
		inv_matrix[1][1] -= m[1][3]*inv_matrix[3][1];
		inv_matrix[1][2] -= m[1][3]*inv_matrix[3][2];
		inv_matrix[1][3] -= m[1][3]*inv_matrix[3][3];

		inv_matrix[0][0] -= m[0][3]*inv_matrix[3][0];
		inv_matrix[0][1] -= m[0][3]*inv_matrix[3][1];
		inv_matrix[0][2] -= m[0][3]*inv_matrix[3][2];
		inv_matrix[0][3] -= m[0][3]*inv_matrix[3][3];

		// use 3rd row
		inv_matrix[1][0] -= m[1][2]*inv_matrix[2][0];
		inv_matrix[1][1] -= m[1][2]*inv_matrix[2][1];
		inv_matrix[1][2] -= m[1][2]*inv_matrix[2][2];
		inv_matrix[1][3] -= m[1][2]*inv_matrix[2][3];

		inv_matrix[0][0] -= m[0][2]*inv_matrix[2][0];
		inv_matrix[0][1] -= m[0][2]*inv_matrix[2][1];
		inv_matrix[0][2] -= m[0][2]*inv_matrix[2][2];
		inv_matrix[0][3] -= m[0][2]*inv_matrix[2][3];

		// use 2nd row
		inv_matrix[0][0] -= m[0][1]*inv_matrix[1][0];
		inv_matrix[0][1] -= m[0][1]*inv_matrix[1][1];
		inv_matrix[0][2] -= m[0][1]*inv_matrix[1][2];
		inv_matrix[0][3] -= m[0][1]*inv_matrix[1][3];

		return true;
	}
	
	/**
	 * store to array
	 */
	inline void store(T *m)
	{
		m[0] = matrix[0][0];
		m[1] = matrix[0][1];
		m[2] = matrix[0][2];
		m[3] = matrix[0][3];

		m[4] = matrix[1][0];
		m[5] = matrix[1][1];
		m[6] = matrix[1][2];
		m[7] = matrix[1][3];
		
		m[8] = matrix[2][0];
		m[9] = matrix[2][1];
		m[10] = matrix[2][2];
		m[11] = matrix[2][3];
		
		m[12] = matrix[3][0];
		m[13] = matrix[3][1];
		m[14] = matrix[3][2];
		m[15] = matrix[3][3];
	}


	/**
	 * linear array access to matrix components
	 */
	inline CVector<4,T>& operator[](const int i)
	{
		return matrix[i];
	}
};



typedef CMatrix4<float>	mat4f;


template <class T>
inline
std::ostream&
operator<<(std::ostream &co, CMatrix4<T> &m)
{
	return co	<< "[" << m[0][0] << ", " << m[0][1] << ", " << m[0][2] << ", " << m[0][3] << "]" << std::endl
			<< "[" << m[1][0] << ", " << m[1][1] << ", " << m[1][2] << ", " << m[1][3] << "]" << std::endl
			<< "[" << m[2][0] << ", " << m[2][1] << ", " << m[2][2] << ", " << m[2][3] << "]" << std::endl
			<< "[" << m[3][0] << ", " << m[3][1] << ", " << m[3][2] << ", " << m[3][3] << "]" << std::endl	;
}



#endif
