#ifndef CQUATERNION_HH_
#define CQUATERNION_HH_

#include <complex>

/**
 * http://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation
 */

template <typename T>
class CQuaternion
{
public:
		T a, b, c, d;

		CQuaternion()
		{
			loadIdentity();
		}
		
		void loadIdentity()
		{
			a = 1;
			b = 0;
			c = 0;
			d= 0;
		}


		CQuaternion(	T p_a,
						T p_b,
						T p_c,
						T p_d)
		{
			a = p_a;
			b = p_b;
			c = p_c;
			d = p_d;
		}

		CQuaternion<T> operator+(CQuaternion<T> &q)
		{
			return CQuaternion(
					a + q.a,
					b + q.b,
					c + q.c,
					d + q.d
			);
		}

		/**
		 * multiply 2 quaternions
		 */
		CQuaternion<T> operator*(CQuaternion<T> &q)
		{
			return CQuaternion(
					a*q.a - b*q.b - c*q.c - d*q.d,
					a*q.b + b*q.a + c*q.d - d*q.c,
					a*q.c + c*q.a + d*q.b - a*q.d,
					a*q.d + d*q.a + b*q.c - c*q.b
			);
		}
		
		/**
		 * compute rotation quaternion rotation around `axis`
		 */
		void createRotation(CVector<3,T> axis, T angle)
		{
			a = CMath<T>::cos(angle*0.5);

			// abuse q3 as temporary variable
			d = CMath<T>::sin(angle*0.5);

			b = d*axis[0];
			c = d*axis[1];
			d *= axis[2];
		}
		
		/**
		 * load point
		 */
		void loadPoint(CVector<3,T> p)
		{
			a = 0;
			b = p[0];
			c = p[1];
			d = p[2];
		}
		
		/**
		 * compute rotation matrix
		 */
		void getRotationMatrix(CMatrix4<T> &matrix)
		{
			matrix[0][0] = a*a + b*b - c*c - d*d;
			matrix[0][1] = 2*b*c - 2*a*d;
			matrix[0][2] = 2*a*c + 2*b*d;
			matrix[0][3] = 0;

			matrix[1][0] = 2*a*d + 2*b*c;
			matrix[1][1] = a*a - b*b + c*c - d*d;
			matrix[1][2] = 2*c*d - 2*a*b;
			matrix[1][3] = 0;

			matrix[2][0] = 2*b*d - 2*a*c;
			matrix[2][1] = 2*a*b + 2*c*d;
			matrix[2][2] = a*a - b*b - c*c + d*d;
			matrix[2][3] = 0;

			matrix[3][0] = 0;
			matrix[3][1] = 0;
			matrix[3][2] = 0;
			matrix[3][3] = 1;

		}
		
		
		/**
		 * normalize the quaterion to a unit length of 1
		 */
		void normalize()
		{
			T fac = (T)1.0 / CMath<T>::sqrt(a*a + b*b + c*c + d*d);
			a *= fac;
			b *= fac;
			c *= fac;
			d *= fac;
		}
};

#endif /*CQUATERNION_HH_*/
