/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __CVECTOR_HH
	#error "dont include CVector3.hh directly!"
#endif

#ifndef __CVECTOR3_HH
#define __CVECTOR3_HH
/*
 * CHANGELOG:
 *
 * 2008-02-27: martin schreiber
 * 	dist2();
 *
 */

#include <iostream>
#include "CMath.hh"

/// 3D Vector handler
template <typename T>
class CVector<3,T>
{
public:
	T data[3];	///< vector data

	/*******************
	 * CONSTRUCTURS
	 *******************/
	inline CVector()
	{
		setZero();
	}

	/**
	 * initialize vector with (x0, x1, x2)
	 */
	inline CVector(const T x0, const T x1, const T x2)
	{
		data[0] = x0;
		data[1] = x1;
		data[2] = x2;
	}

	/**
	 * initialize all vector components with the scalar value 'x'
	 */
	inline CVector(const T x)
	{
		data[0] = x;
		data[1] = x;
		data[2] = x;
	}

	/**
	 * initialize integer vector components with the array 'v' converting to the type (T)
	 */
	inline CVector(const CVector<3,int> &v)
	{
		data[0] = (T)v.data[0];
		data[1] = (T)v.data[1];
		data[2] = (T)v.data[2];
	}

	/**
	 * initialize vector components with the array 'v'
	 */
	inline CVector(const CVector<3,float> &v)
	{
		data[0] = v.data[0];
		data[1] = v.data[1];
		data[2] = v.data[2];
	}

	/**
	 * initialize vector components with the array 'v'
	 */
	inline CVector(const T v[3])
	{
		data[0] = v[0];
		data[1] = v[1];
		data[2] = v[2];
	}


	/**
	 * set all components of vector to 0
	 */
	inline void setZero()
	{
		data[0] = T(0);
		data[1] = T(0);
		data[2] = T(0);	
	}


	/**
	 * return normalized (length=1) vector
	 */
	inline CVector<3,T> getNormalized()
	{
		CVector<3,T> v = *this;
		float inv_length = 1.0f/sqrtf(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
		v[0] *= inv_length;
		v[1] *= inv_length;
		v[2] *= inv_length;
		return v;
	}

	/**
	 * compute and return the dot product of this vector and 'v'
	 * \return dot product
	 */
	inline T getDotProd(const CVector<3,T> &v)
	{
		return v.data[0]*data[0] + v.data[1]*data[1] + v.data[2]*data[2];
	}

	/**
	 * compute and return the cross product of this vector and 'a'
	 * \return cross product
	 */
	inline CVector<3,T> getCrossProd(CVector<3,T> &a)
	{
		return CVector<3,T>(	data[1]*a.data[2] - data[2]*a.data[1],
					data[2]*a.data[0] - data[0]*a.data[2],
					data[0]*a.data[1] - data[1]*a.data[0]
				);
	}

	/**
	 * return square distance to other point
	 */
	inline T dist2(const CVector<3,T> &v)
	{
		CVector<3,T> d = CVector<3,T>(v.data[0] - data[0], v.data[1] - data[1], v.data[2] - data[2]);
		return (d[0]*d[0] + d[1]*d[1] + d[2]*d[2]);
	}

	/**
	 * return distance to other point
	 */
	inline T dist(const CVector<3,T> &v)
	{
		CVector<3,T> d = CVector<3,T>(v.data[0] - data[0], v.data[1] - data[1], v.data[2] - data[2]);
		return CMath<T>::sqrt(d[0]*d[0] + d[1]*d[1] + d[2]*d[2]);
	}

	/**
	 * return elements in cube.
	 * e.g. if the vector components give the resolution of a cube, this function returns the number of cells
	 */
	inline T elements()
	{
		return (data[0]*data[1]*data[2]);
	}

	/**
	 * return length of vector
	 */
	inline T length()
	{
		return CMath<T>::sqrt(data[0]*data[0] + data[1]*data[1] + data[2]*data[2]);
	}

	/**
	 * return (length*length) of vector
	 */
	inline T length2()
	{
		return data[0]*data[0] + data[1]*data[1] + data[2]*data[2];
	}

	/**
	 * normalize the vector
	 */
	inline void normalize()	
	{
		T il = 1/length();
		data[0] *= il;
		data[1] *= il;
		data[2] *= il;
	}

	/**
	 * clamp to values -1 and 1
	 */
	inline void clamp1_1()
	{
		data[0] = (data[0] < -1 ? -1 : (data[0] > 1 ? 1 : data[0]));
		data[1] = (data[1] < -1 ? -1 : (data[1] > 1 ? 1 : data[1]));
		data[2] = (data[2] < -1 ? -1 : (data[2] > 1 ? 1 : data[2]));
	}

//	inline void		setVector(const T a[3])				{	data[0] = a[0]; data[1] = a[1]; data[2] = a[2];	};
//	inline void		setVector(const T a0, const T a1, const T a2)	{	data[0] = a0; data[1] = a1; data[2] = a2;	};

	/*******************
	 * ACCESSIBILITY
	 *******************/
	/*
	inline T operator.x+=(T value)
	{
		data[0] += value;
		return data[0];
	}
	*/


	/*******************
	 * OPERATORS
	 *******************/
	inline CVector<3,T>&	operator=(const T a[3])	{	data[0] = a[0]; data[1] = a[1]; data[2] = a[2];	return *this;	};
	inline CVector<3,T>&	operator=(CVector<3,T> const & a)	{	data[0] = a.data[0]; data[1] = a.data[1]; data[2] = a.data[2];	return *this;	};


// T
	inline CVector<3,T>	operator+(const T a)	{	return CVector<3,T>(data[0]+a, data[1]+a, data[2]+a);	}
	inline CVector<3,T>	operator-(const T a)	{	return CVector<3,T>(data[0]-a, data[1]-a, data[2]-a);	}
	inline CVector<3,T>	operator*(const T a)	{	return CVector<3,T>(data[0]*a, data[1]*a, data[2]*a);	}
	inline CVector<3,T>	operator/(const T a)	{	return CVector<3,T>(data[0]/a, data[1]/a, data[2]/a);	}
	inline CVector<3,T>&	operator+=(const T a)	{	data[0] += a; data[1] += a; data[2] += a;	return *this;	}
	inline CVector<3,T>&	operator-=(const T a)	{	data[0] -= a; data[1] -= a; data[2] -= a;	return *this;	}
	inline CVector<3,T>&	operator*=(const T a)	{	data[0] *= a; data[1] *= a; data[2] *= a;	return *this;	}
	inline CVector<3,T>&	operator/=(const T a)	{	data[0] /= a; data[1] /= a; data[2] /= a;	return *this;	}

	inline CVector<3,T>	operator%(const CVector<3,T> &a)	{	return CVector<3,T>(data[0]%a.data[0], data[1]%a.data[1], data[2]%a.data[2]);	}

// CVector<3>
	inline CVector<3,T>	operator+(const CVector<3,T> &v)	{	return CVector<3,T>(data[0]+v.data[0], data[1]+v.data[1], data[2]+v.data[2]);	}
	inline CVector<3,T>	operator-(const CVector<3,T> &v)	{	return CVector<3,T>(data[0]-v.data[0], data[1]-v.data[1], data[2]-v.data[2]);	}
	inline CVector<3,T>	operator*(const CVector<3,T> &v)	{	return CVector<3,T>(data[0]*v.data[0], data[1]*v.data[1], data[2]*v.data[2]);	}
	inline CVector<3,T>	operator/(const CVector<3,T> &v)	{	return CVector<3,T>(data[0]/v.data[0], data[1]/v.data[1], data[2]/v.data[2]);	}

	inline CVector<3,T>&	operator+=(const CVector<3,T> &v)	{	data[0] += v.data[0]; data[1] += v.data[1]; data[2] += v.data[2];	return *this;	}
	inline CVector<3,T>&	operator-=(const CVector<3,T> &v)	{	data[0] -= v.data[0]; data[1] -= v.data[1]; data[2] -= v.data[2];	return *this;	}

	inline bool	operator==(const CVector<3,T> &v)	{	return bool(data[0] == v.data[0] && data[1] == v.data[1] && data[2] == v.data[2]);	}
	inline bool	operator!=(const CVector<3,T> &v)	{	return bool(data[0] != v.data[0] || data[1] != v.data[1] || data[2] != v.data[2]);	}


	inline T& operator[](const int i)
	{
		return data[i];
	}

	/// compare set for sort operation
	struct compareSet
	{

		inline bool operator()(CVector<3,float> *v1, CVector<3,float> *v2)
		{
			if ((*v1)[0] != (*v2)[0])
				return (*v1)[0] < (*v2)[0];
			if ((*v1)[1] != (*v2)[1])
				return (*v1)[1] < (*v2)[1];
			if ((*v1)[2] != (*v2)[2])
				return (*v1)[2] < (*v2)[2];
			return false;
		}

		inline bool operator()(const CVector<3,T> &v1, const CVector<3,T> &v2)
		{
			if (v1.data[0] != v2.data[0])
				return v1.data[0] < v2.data[0];
			if (v1.data[1] != v2.data[1])
				return v1.data[1] < v2.data[1];
			if (v1.data[2] != v2.data[2])
				return v1.data[2] < v2.data[2];
			return false;
		}
	};
};


typedef CVector<3,float>	vec3f;
typedef CVector<3,int>	vec3i;


template <class T>
inline
std::ostream&
operator<<(std::ostream &co, CVector<3,T> &v)
{
	return co << "[" << v[0] << ", " << v[1] << ", " << v[2] << "]";
}


#endif
