/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef FBO_HPP
#define FBO_HPP

#include <GL/glew.h>
#include <GL/gl.h>
#include "gltools.hpp"
#include "CTexture.hpp"
#include <iostream>
using namespace std;


static const GLuint fbo_attachments__[] = {
		GL_COLOR_ATTACHMENT0_EXT,
		GL_COLOR_ATTACHMENT1_EXT,
		GL_COLOR_ATTACHMENT2_EXT,
		GL_COLOR_ATTACHMENT3_EXT,
		GL_COLOR_ATTACHMENT4_EXT,
		GL_COLOR_ATTACHMENT5_EXT,
		GL_COLOR_ATTACHMENT6_EXT,
		GL_COLOR_ATTACHMENT7_EXT,
		GL_COLOR_ATTACHMENT8_EXT,
		GL_COLOR_ATTACHMENT9_EXT,
		GL_COLOR_ATTACHMENT10_EXT,
		GL_COLOR_ATTACHMENT11_EXT,
		GL_COLOR_ATTACHMENT12_EXT,
		GL_COLOR_ATTACHMENT13_EXT,
		GL_COLOR_ATTACHMENT14_EXT,
		GL_COLOR_ATTACHMENT15_EXT,
	};

class CFbo
{
	GLuint fbo;		// framebuffer object itself
	GLint nTextureUnits;	// number of available texture units
	GLuint mode;

public:
	/**
	 * create new fbo
	 * \param CTexture	texture to bind to framebuffer
	 */
	CFbo(CTexture *texture)
	{
		glGenFramebuffersEXT(1, &fbo);					// create fbo
		glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &nTextureUnits);	// get available texture units
		mode = texture->target;

		bind();
		bindFramebufferTexture(texture);
		unbind();
		checkGlError();
	}


	/**
	 * create new fbo
	 * \param p_mode	texture mode (maybe you want rectangle textures)
	 */
	CFbo(GLuint p_mode = GL_TEXTURE_2D)
	{
		glGenFramebuffersEXT(1, &fbo);		// create fbo

		glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &nTextureUnits);	// get available texture units
//		if (nTextureUnits > 16)	nTextureUnits = 16;			// limit texture units to 16

		mode = p_mode;

		checkGlError();
	}

	~CFbo()
	{
		glDeleteFramebuffersEXT(1, &fbo);
	}

	/**
	 * bind texture to framebuffer
	 * \param textureUnit	texture unit to bind texture to
	 */
	void bindFramebufferTexture(CTexture *texture, GLint textureUnit = 0)
	{
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, fbo_attachments__[textureUnit], texture->target, texture->textureid, 0);
		checkStatus();
		checkGlError();
	}

	/**
	 * bind texture to framebuffer
	 * \param textureUnit	texture unit to bind texture to
	 */
	/*
	void bindFramebufferTexture(GLuint textureid, GLuint target, GLint textureUnit = 0)
	{
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, fbo_attachments[textureUnit], target, textureid, 0);
		checkStatus();
		checkGlError();
	}
	*/

	/**
	 * unbind texture from framebuffer
	 */
	void unbindFramebufferTexture(GLint textureUnit)
	{
//		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, fbo_attachments__[textureUnit], GL_TEXTURE_2D, 0, 0);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, fbo_attachments__[textureUnit], mode, 0, 0);
		checkGlError();
	}


	/**
	 * bind depth texture to framebuffer
	 * \param textureUnit	texture unit to bind texture to
	 */
	void bindFramebufferDepthTexture(CTexture *texture, GLint textureUnit = 0)
	{
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, texture->target, texture->textureid, 0);
		checkStatus();
		checkGlError();
	}

	/**
	 * unbind depth texture to framebuffer
	 */
	void bindFramebufferDepthTexture()
	{
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, mode, 0, 0);
		checkStatus();
		checkGlError();
	}

	/**
	 * bind framebuffer, so we can work with it or initialize it
	 */
	void bind()
	{
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
		checkGlError();
	}

	void unbind()
	{
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
		checkGlError();
	}

	bool checkStatus()
	{
		GLint status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
		switch (status)
		{
			case GL_FRAMEBUFFER_COMPLETE_EXT:
				return true;

			case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
				cout << "FBO: incomplete attachment" << endl;
				return false;

			case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
				cout << "FBO: incomplete missing attachment" << endl;
				return false;
/*
			case GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT:
				cout << "FBO: incomplete duplicate attachment" << endl;
				return false;
*/
			case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
				cout << "FBO: incomplete dimensions" << endl;
				return false;

			case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
				cout << "FBO: incomplete formats" << endl;
				return false;

			case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
				cout << "FBO: incomplete draw buffer" << endl;
				return false;

			case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
				cout << "FBO: incomplete read buffer" << endl;
				return false;

			case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
				cout << "FBO: unsupported" << endl;
				return false;

			default:
				fprintf(stderr, "FBO: unknown error 0x%x\n", status);
				return false;
		}
	}


};

#endif
