/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "CPrimitives.hpp"
#include <GL/gl.h>
#include <math.h>


void CPrimitives::circle(int fans)
{
	glBegin(GL_LINE_STRIP);
	for (int i = 0; i < fans+1; i++)
	{
		float x = sinf(	(M_PI*2*(float)i)/(float)fans);
		float y = cosf(	(M_PI*2*(float)i)/(float)fans);
		glTexCoord2f(1.0f+x, 1.0f+y);
		glVertex3f(x, 0, y);
	}
	glEnd();

	glBegin(GL_POINTS);
	  glVertex3f(0,0,0);
	glEnd();
}


void CPrimitives::round_plate(int fans, float scale)
{
	glBegin(GL_TRIANGLE_FAN);
	glNormal3f(0, 1, 0);

	glTexCoord2f(0.5f, 0.5f);
	glVertex2f(0,0);
	for (int i = 0; i < fans+1; i++)
	{
		float y = sinf(	-(M_PI*2*(float)i)/(float)fans);
		float x = cosf(	-(M_PI*2*(float)i)/(float)fans);
		glTexCoord2f(x*0.5f + 0.5f, y*0.5f + 0.5f);
		glVertex3f(x*scale, 0, y*scale);
	}
	glEnd();
}

void CPrimitives::barrel(int parts, float scale_height, float scale_radius)
{
//glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	float px1 = 0;
	float pz1 = 1;

//	glEnable(GL_NORMALIZE);
	glBegin(GL_QUADS);
	for (int a = 1; a < parts+1; a++)
	{
		float px2 = sinf(-(M_PI*2*(float)a)/(float)parts);
		float pz2 = cosf(-(M_PI*2*(float)a)/(float)parts);

		float py1 = -1;
		for (int y = 1; y < parts+1; y++)
		{
			float py2 = -1+(float)(y*2)/(float)parts;

			// parabolic distortion
			float d1 = 1.2f-0.2f*py1*py1;
			float d2 = 1.2f-0.2f*py2*py2;

			glNormal3f(px1*d2, 0, pz1*d2);	glTexCoord2f((float)(a)/(float)parts, py2);	glVertex3f(px1*d2*scale_radius, py2*scale_height, pz1*d2*scale_radius);
			glNormal3f(px2*d2, 0, pz2*d2);	glTexCoord2f((float)(a+1)/(float)parts, py2);	glVertex3f(px2*d2*scale_radius, py2*scale_height, pz2*d2*scale_radius);
			glNormal3f(px2*d1, 0, pz2*d1);	glTexCoord2f((float)(a+1)/(float)parts, py1);	glVertex3f(px2*d1*scale_radius, py1*scale_height, pz2*d1*scale_radius);
			glNormal3f(px1*d1, 0, pz1*d1);	glTexCoord2f((float)(a)/(float)parts, py1);	glVertex3f(px1*d1*scale_radius, py1*scale_height, pz1*d1*scale_radius);

			py1 = py2;
		}

		px1 = px2;
		pz1 = pz2;
	}
	glEnd();
//	glDisable(GL_NORMALIZE);
}


/* 
 * BUILD A SPHERE. Sphere is centered at c with radius r.
 * LOD specifies the level of detail, i.e. the number of quads is LOD*LOD/2.
 * Sphere does not actually draw a sphere, but records it in a so called 
 * "display-list". Display-lists are used to cache vertices and offer
 * great rendering speed, but only for static geometry. Sphere returns
 * an unsigned integer that serves as a handle to access the geometry. See 
 * functions initialize() and displayfunc() for more details.
 */
void CPrimitives::sphere(float r, float cx,float cy,float cz, unsigned int LODi, float texScale)
{
	unsigned int quads_per_layer;
	GLuint sphere_int = 0;
	int layers;
	sphere_int++;

	// quads_per_layer * layers should be about LODi value
	quads_per_layer = (unsigned int)sqrt((float)(LODi*LODi/2));
	layers = quads_per_layer & ~1;

	float nx, ny, nz;	// normals
	float h_nx, h_ny, h_nz;	// higher normals

	float angle;

	float layer_nr, higher_layer_nr;
	float layer_angle, higher_layer_angle;

	for (int layer = -(layers/2)+1; layer < (layers/2) - 1; layer++)
	{
		layer_angle = (M_PI_2*layer)/((layers >> 1) - 1);
		higher_layer_angle = (M_PI_2*(layer + 1))/((layers >> 1) - 1);

		ny = sin(layer_angle);
		h_ny = sin(higher_layer_angle);

		layer_nr = cos(layer_angle);
		higher_layer_nr = cos(higher_layer_angle);

		glBegin(GL_QUAD_STRIP);
		for (unsigned int quad = 0; quad < quads_per_layer + 1; quad++)
		{
			angle = (2*M_PI*quad)/quads_per_layer;

			nx = cos(angle) * layer_nr;
			nz = sin(angle) * layer_nr;

			h_nx = cos(angle) * higher_layer_nr;
			h_nz = sin(angle) * higher_layer_nr;

			glTexCoord2f((1 - angle/M_PI/2) * texScale, (h_ny + 1) / 2);
			glNormal3f(h_nx, h_ny, h_nz);
			glVertex3f(h_nx*r + cx,	h_ny*r + cy,	h_nz*r + cz);

			glTexCoord2f((1 - angle/M_PI/2) * texScale, (ny + 1) / 2);
			glNormal3f(nx, ny, nz);
			glVertex3f(nx*r + cx,	ny*r + cy,	nz*r + cz);
		}
		glEnd();
	}
}


void CPrimitives::plane(int lod, float size)
{
	float x0, x1;

	x1 = 0;
	for (int x = 1; x <= lod; x++)
	{
		x0 = x1;
		x1 = (float)x/(float)lod;

		float yp = 0;

		glBegin(GL_QUAD_STRIP);
		glTexCoord2f(x1, yp);	glVertex3f((x1-0.5f)*size, 0, (yp-0.5f)*size);
		glTexCoord2f(x0, yp);	glVertex3f((x0-0.5f)*size, 0, (yp-0.5f)*size);

		for (int y = 1; y <= lod; y++)
		{
			yp = (float)y/(float)lod;

			if (y & 1)
			{
				glTexCoord2f(x1, yp);	glVertex3f((x1-0.5f)*size, 0, (yp-0.5f)*size);
				glTexCoord2f(x0, yp);	glVertex3f((x0-0.5f)*size, 0, (yp-0.5f)*size);
			}
			else
			{
				glTexCoord2f(x1, yp);	glVertex3f((x1-0.5f)*size, 0, (yp-0.5f)*size);
				glTexCoord2f(x0, yp);	glVertex3f((x0-0.5f)*size, 0, (yp-0.5f)*size);
			}
		}
		glEnd();
	}
}
