/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include "CProgram.hpp"
#include "gltools.hpp"
#include <string.h>
#include <stdarg.h>
#include "../lib/CFile.hpp"


using namespace std;

/**
 * Default Contructor
 */
void CProgram::init(bool checkExtension)
{
	if (checkExtension)
	{
		if (!strstr((char*)glGetString(GL_EXTENSIONS), "_fragment_program"))
		{
			shaderAvailable = false;
			return;
		}
	}
	else
	{
		shaderAvailable = false;
	}

	GLProgram = 0;
	shaderAvailable = true;
	error = false;
}


CProgram::CProgram(bool checkExtension)
{
	init(checkExtension);
	return;
}

/**
 * initialize with files
 * file endings .vert are vertex shaders
 * those with .frag are fragment shaders
 */
CProgram::CProgram(const char *files, ...)
{
	init();
	if (!shaderAvailable)
		return;

	va_list ap;
	list<SShaderFile> shaderFiles;

	va_start(ap, files);
	const char *arg = files;
	while (arg)
	{
		shaderFiles.push_back(SShaderFile());
		SShaderFile &shaderFile = shaderFiles.back();

		shaderFile.fileName = arg;

		int l = shaderFile.fileName.length();
		if (l < 5)
		{
			cerr << "filename to short" << endl;
			error = true;
			return;
		}

		if (shaderFile.fileName.compare(l-5, 5, ".vert") == 0)
		{
			shaderFile.type = GL_VERTEX_SHADER;
		}
		else if (shaderFile.fileName.compare(l-5, 5, ".frag") == 0)
		{
			shaderFile.type = GL_FRAGMENT_SHADER;
		}
		else
		{
			cerr << "filetype of file '" << shaderFile.fileName << "' not recognized - must end with .vert or .frag" << endl;
			error = true;
			return;
		}
	
		arg = va_arg(ap, const char*);
	}
	va_end(ap);

	load(shaderFiles);

}

/**
 * Destructor
 * Cleanup the ressources
 */
CProgram::~CProgram()
{
	if (!shaderAvailable)	return;

	cleanup();

#ifdef _TC_DEBUG_
	checkGlError();
#endif
}



/**
 * Load Shader
 * Compile the Shader with type ShaderType given by shaderCode
 * \param shaderCode - 0 terminated shadercode string
 * \param ShaderType - defines the GL Shader Type (GL_VERTEX_SHADER or GL_FRAGMENT_SHADER)
 */
bool CProgram::loadShader(SShaderCode &shaderCode)
{
	if (!shaderAvailable)	return false;

	GLuint Shader = glCreateShader(shaderCode.type);
	shaderCode.glShader = Shader;

	const char *code = shaderCode.code.c_str();
	glShaderSource(Shader, 1, (const GLchar**)&code, 0);
	glCompileShader(Shader);

	GLint param = GL_TRUE;
	glGetShaderiv(Shader, GL_COMPILE_STATUS, &param);

	// always read out info log
	int logLength = 0;
	glGetShaderiv(Shader, GL_INFO_LOG_LENGTH, &logLength);
	
	if (logLength == 0)
	{
		errorLog = "no shader log available";
	}
	else
	{
		char *tmp_log = new GLchar[logLength];
		if (!tmp_log)
		{
			error = true;
			errorLog = "out of memory";
			return false;
		}
		GLsizei length = 0;
		glGetShaderInfoLog(Shader, logLength, &length, tmp_log);
		if (length == 0)
		{
			((char*)tmp_log)[0] = '\0';
		}
		errorLog = tmp_log;
	}

	checkGlError();

	if (param == GL_FALSE)
	{
		errorFile = shaderCode.fileName;
		errorLog = "('" + errorFile + "'): '" + errorLog + "'";
		shaderCode.output();
		error = true;
		return false;
	}
#ifdef PROGRAM_OUTPUT_WARNINGS
	else
	{
		errorFile = shaderCode.fileName;
		if (errorLog.length() > 0)
		{
			if (errorFile.length())
				cout << errorFile << endl;
			cout << errorLog << endl;
			shaderCode.output();
		}
	}
#endif

	return true;
}


/*
 * Link CProgram
 * Create a new program and link every shader connected to this class the program
 * the gl program id will be in stored [CProgram]
 * \return true, if the linkage was sucessful
 */
bool CProgram::linkProgram(list<SShaderCode> &shaderCodes)
{
	if (!shaderAvailable)	return false;

	GLProgram = glCreateProgram();

	if (GLProgram == 0)
	{
		fprintf(stderr, "Cannot create GL CProgram Handler!\n");
		return false;
	}

	if (shaderCodes.size())
		identifier = shaderCodes.front().fileName;

	for (list<SShaderCode>::iterator iter = shaderCodes.begin(); iter != shaderCodes.end(); iter++)
		glAttachShader(GLProgram, (*iter).glShader);

	glLinkProgram(GLProgram);

	GLint param = GL_TRUE;
	glGetProgramiv(GLProgram, GL_LINK_STATUS, &param);

	if (param == GL_TRUE)
	{
		errorLog = "";
	}
	else
	{
		int logLength = 0;
		glGetProgramiv(GLProgram, GL_INFO_LOG_LENGTH, &logLength);

		GLchar *tmp_log = new GLchar[logLength];
		if (!tmp_log)
		{
			cout << "out of memory" << endl;
			goto link_error;
		}

		GLsizei length = 0;
		glGetProgramInfoLog(GLProgram, logLength, &length, tmp_log);

		cerr << "Error Log (identifier: '" << identifier << "'): \"" << errorLog << "\"" << endl;

		errorLog = tmp_log;
#ifdef PROGRAM_OUTPUT_WARNINGS
		if (errorLog.length() > 0)
			cout << errorLog << endl;
#endif

		goto link_error;
	}

	return true;


link_error:
	for (list<SShaderCode>::iterator iter = shaderCodes.begin(); iter != shaderCodes.end(); iter++)
	{
		glDetachShader(GLProgram, (*iter).glShader);
		glDeleteShader((*iter).glShader);
	}

	glDeleteProgram(GLProgram);

	GLProgram = 0;

	error = true;
	return false;
}


/*
 * Cleanup
 * If an error occurs or the user will clean up the ressources, this function is called to
 * clean up the existing memory and free the gl handlers
 */
void CProgram::cleanup()
{
	if (!shaderAvailable)	return;

	// clear shader code ids!
//	shaderCodes.clear();

	if (GLProgram)
	{
		glDeleteProgram(GLProgram);
		GLProgram = 0;
	}
}


/**
 * Create a new program from memory
 * Load "length" shaders from shader_code, compile and link them to a program
 * \param shader_code - array with 0 terminated shader codes and shader types
 * 			end of array is shader_code_s with ShaderCode == 0
 */
bool CProgram::load(list<SShaderCode> &shaderCodes)
{
	if (!shaderAvailable)	return false;

	if (shaderCodes.size() == 0)
	{
		cerr << "no shaders found (cannot create empty program)!" << endl;
		error = true;
		return false;
	}

	cleanup();	// clean up existing ressources

	for (list<SShaderCode>::iterator code_iter = shaderCodes.begin(); code_iter != shaderCodes.end(); code_iter++)
	{
		SShaderCode &shaderCode = *code_iter;

		if (!loadShader(shaderCode))	// load and init shader
		{
			cerr << "error in file '" << shaderCode.fileName << "'" << endl;

			for (list<SShaderCode>::iterator code_iter = shaderCodes.begin(); code_iter != shaderCodes.end(); code_iter++)
			{
				SShaderCode &shaderCode = *code_iter;
				if (shaderCode.glShader != 0)
					glDeleteShader(shaderCode.glShader);
			}

			error = true;
			cleanup();
			return false;
		}
	}

	bool retval = linkProgram(shaderCodes);
	if (retval)
		return true;

	for (list<SShaderCode>::iterator code_iter = shaderCodes.begin(); code_iter != shaderCodes.end(); code_iter++)
		(*code_iter).output();

	return false;
}

/**
 * load the shader specified in file
 * here we use a precompilation to make it possible that shaders can include other shader code!
 */
bool CProgram::loadAndPrecompile(const string &file, string &code, int depth)
{
	if (depth > 10)
	{
		errorLog = "too much resursions";
		error = true;
		return false;
	}

	if (!CFile::fileContents(file, code, errorLog))
	{
		error = true;
		errorLog = "unable to read file '" + file + "'";
		return false;
	}

	size_t pos = 0;
	do
	{
		if (code.length() <= pos)
			break;

		// strlen("#include ") == 9
		if (code.compare(pos, 9, "#include ") == 0)
		{
			int pos_eol = code.find('\n', pos);	// search for end of line

			if (pos_eol < 0)
				pos_eol = code.length();	// if no eol is found, this is the last line in the file

			int start_pos = pos+9;
			int end_pos = pos_eol;
			int inc_file_length = pos_eol-pos-9;

			if (code[start_pos] == '"' && code[pos_eol-1] == '"')
			{
				start_pos++;
				end_pos--;
				inc_file_length -= 2;
			}

			string incfile = code.substr(start_pos, inc_file_length);

			string inc_code;
			if (!loadAndPrecompile(incfile, inc_code, depth+1))
				return false;

			code = code.substr(0, pos) + inc_code + code.substr(pos_eol);
			pos += inc_code.length();
		}

	} while ((pos = code.find('\n', pos) + 1) >= 1);

	return true;
}


/**
 * Create a new program from file
 * Load "length" shaders from file and call Load
 * \param shader_file - array with 0 terminated shader files and shader types
 * 			end of array is shader_code_s with shaderFile == 0
 * \see program.h for details about shader_file_s
 */
bool CProgram::load(list<SShaderFile> &shaderFiles)
{
	if (!shaderAvailable)	return false;

	cleanup();	// clean up existing ressources

	list<SShaderCode> shaderCodes;

	for (list<SShaderFile>::iterator files_iter = shaderFiles.begin(); files_iter != shaderFiles.end(); files_iter++)
	{
		SShaderFile &shaderFile = *files_iter;

		shaderCodes.push_back(SShaderCode());
		SShaderCode &shaderCode = shaderCodes.back();

		shaderCode.fileName = shaderFile.fileName;
		shaderCode.type = shaderFile.type;

		if (!loadAndPrecompile(shaderFile.fileName, shaderCode.code))
			return false;
//		shaderCode.code = "#error\n"+shaderCode.code;
	}

	if (!load(shaderCodes))
	{
		error = true;
		return false;
	}
	return true;
}

void SShaderCode::output(bool extraInformation)
{
	string output = code;

	int line = 2;
	output.insert(0, "  1: ");

	char buffer[100];
	size_t pos = output.find('\n');

	while(pos != string::npos)
	{
		sprintf(buffer, "% 3i: ", line);
		output.insert(pos+1, buffer);
		pos += strlen(buffer);

		pos = output.find('\n', pos);
		line++;
	}

	if (extraInformation)
		cout << "--- " << fileName << " (" << (type == GL_VERTEX_SHADER ? "VERTEX SHADER" : "FRAGMENT SHADER") << ") ---" << endl;
	cout << output << endl;

	if (extraInformation)
		cout << "--- --- ---" << endl;
}

