/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * \file CProgram.h
 * \brief GLSL Handling
 *
 * \class CProgram
 * \brief GLSL Handling
 *
 * provide functions to load and activate opengl shaders
 */
#ifndef CPROGRAM_HPP
#define CPROGRAM_HPP

#include <iostream>
#include <list>
using namespace std;

#include <GL/glew.h>
#include <GL/gl.h>
#include <string>
#include "gltools.hpp"

/**
 * \brief struct to shader code strings
 * create an array with this struct to initialize the shader program
 */
struct SShaderCode
{
	string code;		// shader code
	string fileName;	// filename for debug purposes (or any other information)
	GLuint type;		// gl type of shader code (fragment / vertex)
	GLuint glShader;

	SShaderCode()
	{
		glShader = 0;
	}

	void output(bool extraInformation = false);
};


/**
 * \brief struct describing the location of the files for a shader code
 * create an array with this struct to initialize the shader program with the filenames
 */
struct SShaderFile
{
	string fileName;	// filename
	GLuint type;
};



class CProgram
{
	private:
		string identifier;
		GLuint GLProgram;	///< GL program Handler
		bool shaderAvailable;	///< GL ShaderLanguage Availalbe

		void init(bool checkExtension = true);
		bool linkProgram(list<SShaderCode> &shaderCodes);

		bool loadShader(SShaderCode &shaderCode);

	public:
		string errorLog;		///< last error string
		string errorFile;	///< filename of last error
		bool error;		///< error happened

		CProgram(bool checkExtension);
		CProgram(const char *, ...);	// 0 terminated file list with shaders (vertex and fragment shaders)
		~CProgram();

		bool loadAndPrecompile(const string &file, string &code, int depth = 0);

		/**
		 * load and initialize the program according to the code given by shaderCodes
		 */
		bool load(list<SShaderCode> &shaderCodes);

		/**
		 * load the shader code from shaderFiles and initialize the shader program
		 */
		bool load(list<SShaderFile> &shaderFiles);

		void cleanup();

		GLuint getFileType(string fileName)
		{
			size_t l = fileName.length();

			if (l < 5)
			{
				error = true;
				errorLog = "filetype not recognized - endings .frag or .vert needed";
				return 0;
			}

			if (fileName.compare(l-5, 5, ".vert") == 0)
			{
				return GL_VERTEX_SHADER;
			}
			else if (fileName.compare(l-5, 5, ".frag") == 0)
			{
				return GL_FRAGMENT_SHADER;
			}

			error = true;
			errorLog = "filetype not recognized - endings .frag or .vert needed";
			return 0;
		}


/**
 * return the location of an uniform in the compiled shader program
 * \param name	uniform name
 */
inline GLint getUniformLocation(const GLchar *name)
{
#ifndef DEBUG
	return glGetUniformLocation(GLProgram, name);
#else
	if (!shaderAvailable)	return 0;

	GLint location;

	location = glGetUniformLocation(GLProgram, name);

	checkGlError();
	return location;
#endif
}




/**
 * set a float uniform value
 * \param name	name of uniform variable
 * \param x	value to set
 */
inline void setUniform1f(const GLchar *name, float x)
{
#ifndef DEBUG
	glUniform1f(glGetUniformLocation(GLProgram, name), x);
#else
	if (!shaderAvailable)	return;

	GLint location;
	location = glGetUniformLocation(GLProgram, name);

	if (location >= 0)	glUniform1f(location, x);
	else			cerr << identifier << ": unknown location '" << name << "'" << endl;

	checkGlError();
#endif
}



/**
 * set a 2 float uniform values
 * \param name	name of uniform variable
 * \param x	x value to set
 * \param y	y value to set
 */
inline void setUniform2f(const GLchar *name, float x, float y)
{
#ifndef DEBUG
	glUniform2f(glGetUniformLocation(GLProgram, name), x, y);
#else
	if (!shaderAvailable)	return;

	GLint location;
	location = glGetUniformLocation(GLProgram, name);

	if (location >= 0)	glUniform2f(location, x, y);
	else			cerr << identifier << ": unknown location '" << name << "'" << endl;

	checkGlError();
#endif
}

inline void setUniform2f(const GLchar *name, float *v)
{
#ifndef DEBUG
	glUniform2fv(glGetUniformLocation(GLProgram, name), 1, v);
#else
	if (!shaderAvailable)	return;

	GLint location;
	location = glGetUniformLocation(GLProgram, name);

	if (location >= 0)	glUniform2fv(location, 1, v);
	else			cerr << identifier << ": unknown location '" << name << "'" << endl;

	checkGlError();
#endif
}

inline void setUniform3f(const GLchar *name, float x, float y, float z)
{
#ifndef DEBUG
	glUniform3f(glGetUniformLocation(GLProgram, name), x, y, z);
#else
	if (!shaderAvailable)	return;

	GLint location;
	location = glGetUniformLocation(GLProgram, name);

	if (location >= 0)	glUniform3f(location, x, y, z);
	else			cerr << identifier << ": unknown location '" << name << "'" << endl;
#endif
}

inline void setUniform3f(const GLchar *name, float *v)
{
#ifndef DEBUG
	glUniform3fv(glGetUniformLocation(GLProgram, name), 1, v);
#else
	if (!shaderAvailable)	return;

	GLint location;
	location = glGetUniformLocation(GLProgram, name);

	if (location >= 0)	glUniform3fv(location, 1, v);
	else			cerr << identifier << ": unknown location '" << name << "'" << endl;
#endif
}


inline void setUniform4f(const GLchar *name, float x, float y, float z, float w)
{
#ifndef DEBUG
	glUniform4f(glGetUniformLocation(GLProgram, name), x, y, z, w);
#else
	if (!shaderAvailable)	return;

	GLint location;
	location = glGetUniformLocation(GLProgram, name);

	if (location >= 0)	glUniform4f(location, x, y, z, w);
	else			cerr << identifier << ": unknown location '" << name << "'" << endl;
#endif
}


inline void setUniform1i(const GLchar *name, int x)
{
#ifndef DEBUG
	glUniform1i(glGetUniformLocation(GLProgram, name), x);
#else
	if (!shaderAvailable)	return;

	GLint location;
	location = glGetUniformLocation(GLProgram, name);

	if (location >= 0)	glUniform1i(location, x);
	else			cerr << identifier << ": unknown location '" << name << "'" << endl;
	checkGlError();
#endif
}


/*

inline void setUniform2i(const GLchar *name, int x, int y)
{
	if (!shaderAvailable)	return;

	GLint location;
	location = glGetUniformLocation(GLProgram, name);

	if (location >= 0)	glUniform2i(location, x, y);
	else			fprintf(stderr, "unknown location '%s'\n", name);

	checkGlError();
}


inline void setUniform3i(const GLchar *name, int x, int y, int z)
{
	if (!shaderAvailable)	return;

	GLint location;
	location = glGetUniformLocation(GLProgram, name);

	if (location >= 0)	glUniform3i(location, x, y, z);
	else			fprintf(stderr, "unknown location '%s'\n", name);
}


inline void setUniform4i(const GLchar *name, int x, int y, int z, int w)
{
	if (!shaderAvailable)	return;

	GLint location;
	location = glGetUniformLocation(GLProgram, name);

	if (location >= 0)
		glUniform4i(location, x, y, z, w);
}
*/


/**
 * CProgram Gl id
 * returns the gl handler id (e.g. to set uniform variables)
 * \return program gl handler
 */
inline GLuint id()
{
	if (!shaderAvailable)	return false;

	return GLProgram;
}


/**
 * Enable Shader CProgram
 */
inline void enable()
{
	if (!shaderAvailable)	return;

	glUseProgram(GLProgram);
#ifdef _TC_DEBUG_
	checkGlError();
#endif
}


/**
 * Disable Shader CProgram
 */
inline void disable()
{
	if (!shaderAvailable)	return;

	glUseProgram(0);
	checkGlError();
}


};

#endif
