/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 
#ifndef CTRACKBALL_HH
#define CTRACKBALL_HH


#include "../libmath/CVector.hh"
#include "../libmath/CMatrix4.hh"
#include "../libmath/CQuaternion.hh"
#include "CPrimitives.hpp"
#include "opengl.h"

/**
 * class to handle the movement on a 3d ball
 * 
 * this implements a virtual trackball with quaternion
 */
/*
 * we need a start rotationMatrixStart.
 * this is the rotation where the trackball is at its recent position.
 * this rotation is done by using the rotationQuaterionStart.
 * (rotationMatrixStartDrag, rotationQuaterionStartDrag)
 * 
 * during the tracking, a second quanterion is used to rotate the ball further.
 * this is useful to handle the current rotation visualization as well as
 * a correct handling of the visualization
 * (rotationMatrixCurrent, rotationQuaterionCurrent)
 * 
 * for the user program, a matrix array rotationMatrixArray is made available
 * to rotate the current scene
 */
template <typename T>
class CTrackBall
{
	// size of trackball
	T ball_size;

	// set to true to cover the whole viewport with a "virtual" sphere
	bool viewport_sized_trackball;	

	// speed of movement (factor multiplied to the angle)
	T angle_speed;

	/// true, if dragging started
	bool drag_started;

	/// position on sphere where dragging started
	CVector<3,float> drag_position_start;

	// current position
	CVector<3,float> drag_position_end;

	// stretching of frustum to achieve circular trackball
	float frustum_stretch_x, frustum_stretch_y;

	// temporary storage for viewport size
	float viewport_width, viewport_height;
	float ball_dist;	// distance of ball to the viewing position

	// current rotation (dragging)
	CQuaternion<T> rotationQuaterionCurrent;

	CQuaternion<T> rotationQuaterionStartDrag;
	CMatrix4<T> rotationMatrixStartDrag;

public:
	CMatrix4<T> rotationMatrixCurrent;
	T rotationMatrixArray[16];

private:
	void loadViewportSize()
	{
		int viewport[4];
		glGetIntegerv(GL_VIEWPORT, viewport);

		viewport_width = (float)viewport[2];
		viewport_height = (float)viewport[3];
	}

public:
	CTrackBall()
	{
		reset();
	}

	CTrackBall(bool p_viewport_sized_trackball, T p_angle_speed)
	{
		reset();

		viewport_sized_trackball = p_viewport_sized_trackball;
		angle_speed = p_angle_speed;
	}

	void setParameters(bool p_viewport_sized_trackball, T p_angle_speed)
	{
		viewport_sized_trackball = p_viewport_sized_trackball;
		angle_speed = p_angle_speed;
	}

	/**
	 * reset the trackball to the identity rotation
	 */
	void reset()
	{
		ball_size = (T)1;
		drag_started = false;
		ball_dist = (T)2.5;
		if (1)
		{
			viewport_sized_trackball = true;
			angle_speed = 5.0;
		}
		else
		{
			viewport_sized_trackball = false;
			angle_speed = 1.0;
		}

		rotationQuaterionCurrent.loadIdentity();
		rotationMatrixCurrent.loadIdentity();
		
		rotationMatrixStartDrag.loadIdentity();
		rotationMatrixStartDrag.store(rotationMatrixArray);
	}


public:

	/**
	 * call this function when the dragging is started.
	 * 
	 * left upper edge is (0,0)
	 * right lower edge is (viewport_width, viewport_height)
	 */
	void dragStart(float viewport_pos_x, float viewport_pos_y)
	{
		rotationMatrixStartDrag = rotationMatrixCurrent;
		rotationQuaterionStartDrag = rotationQuaterionCurrent;

		CVector<2,float> viewport_pos(viewport_pos_x, viewport_pos_y);
		CVector<3,float> sphere_position;

		if (!computeSpherePosition(viewport_pos, drag_position_start))
			return;

		drag_position_end = drag_position_start;

		drag_started = true;
	}


	/**
	 * call this function when the mouse button is pressed during dragging
	 */
	void dragMove(float viewport_pos_x, float viewport_pos_y)
	{
		if (!drag_started)
			return;

		CVector<2,T> viewport_pos(viewport_pos_x, viewport_pos_y);
		CVector<3,T> sphere_position;

		if (!computeSpherePosition(viewport_pos, drag_position_end))
			return;

		CVector<3,T> v1 = drag_position_start.getNormalized();
		CVector<3,T> v2 = drag_position_end.getNormalized();

		T angle = CMath<T>::acos(v1.getDotProd(v2)) * angle_speed;

		if (CMath<T>::abs(angle) < 0.0001)
		{
			rotationMatrixCurrent = rotationMatrixStartDrag;
			rotationMatrixCurrent.store(rotationMatrixArray);
			return;
		}

		CVector<3,T> rotAxis = drag_position_start.getCrossProd(drag_position_end);
		rotAxis.normalize();

		CQuaternion<T> rotQuatNew;
		rotQuatNew.createRotation(rotAxis, -angle);

		// combine dragStart quaterion with current rotation
		rotationQuaterionCurrent = rotationQuaterionStartDrag * rotQuatNew;

		// normalize the crap
		rotationQuaterionCurrent.normalize();

		// store to current values
		rotationQuaterionCurrent.getRotationMatrix(rotationMatrixCurrent);
		rotationMatrixCurrent.store(rotationMatrixArray);
	}

	void dragEnd()
	{
		if (!drag_started)
			return;

/*
		cout <<	rotationQuaterionCurrent.a << " " <<
				rotationQuaterionCurrent.b << " " <<
				rotationQuaterionCurrent.c << " " <<
				rotationQuaterionCurrent.d << " " << endl;
			
		cout <<	rotationQuaterionCurrent.a*rotationQuaterionCurrent.a +
				rotationQuaterionCurrent.b*rotationQuaterionCurrent.b +
				rotationQuaterionCurrent.c*rotationQuaterionCurrent.c +
				rotationQuaterionCurrent.d*rotationQuaterionCurrent.d << endl;
*/
		drag_started = false;
	}

private:
	/**
	 * compute how the frustum has to be stretched to achieve a circular
	 * sphere while rendering it
	 */
	void computeFrustumStretch()
	{
		loadViewportSize();

		if (viewport_width < viewport_height)
		{
			frustum_stretch_x = 1.0;
			frustum_stretch_y = viewport_height / viewport_width;
		}
		else
		{
			frustum_stretch_x = viewport_width / viewport_height;
			frustum_stretch_y = 1.0;
		}
	}


	/**
	 * compute the position on the sphere
	 * 
	 * \param viewport_pos	position in the viewport from opengl
	 * \param position		reference to return value
	 * \return	true, if point on sphere was found
	 */
	bool computeSpherePosition(CVector<2,T> viewport_pos, CVector<3,T> &position)
	{
		computeFrustumStretch();

		// compute the position of the mouse pointer on the near plane
		// loadViewportSize();	was loaded with computeFrustumStretch

		float near_x = ((2*viewport_pos[0])/viewport_width - 1.0)*frustum_stretch_x;
		float near_y = -((2*viewport_pos[1])/viewport_height - 1.0)*frustum_stretch_y;

		if (viewport_sized_trackball)
		{
			near_x /= ball_size*ball_dist*frustum_stretch_x*CMath<T>::sqrt(2);
			near_y /= ball_size*ball_dist*frustum_stretch_y*CMath<T>::sqrt(2);
		}

		/*
		 * now we have the ray equations:
		 *    x = near_x * t;
		 *    y = near_y * t;
		 * 
		 * center_z: ball center x coordinate
		 * 
		 * sphere equation for (0,0,2) translated sphere:
		 *    x^2 + y^2 + (z-center_z)^2 = ball_size^2
		 * 
		 * combining everything:
		 *    z^2 * (near_x^2 + near_y^2 + 1) + (-2*center_z)*z + (center_z^2 - ball_size^2)
		 * 
		 *    t = near_x^2 + near_y^2 + 1
		 * 
		 *    z^2 * (near_x^2 + near_y^2 + 1) +
		 *    z   * (-2*center_z) +
		 *          (center_z^2 - ball_size^2)
		 *    = 0
		 *    = a(x^2) + bx + c   <<<<<<<<<<<<<<!! Diskriminante
		 */

		float a = near_x*near_x + near_y*near_y + 1.0;
		float b = -2.0*ball_dist;
		float c = ball_dist*ball_dist - ball_size*ball_size;

		// x_1_2 = b^2 - 4*a*c
		float sub_sqrt = b*b - 4.0 * a * c;

		if (sub_sqrt <= 0)
			return false;

		float t = (-b - CMath<T>::sqrt(sub_sqrt)) / (2.0*a);

		position[0] = near_x * t;
		position[1] = near_y * t;
		position[2] = ball_dist - t;

		return true;
	}
	

public:
	/**
	 * render the trackball with the current rotation
	 * 
	 * !!! the depth buffer is cleared and modified !!!
	 * 
	 * light has to set up by the calling function
	 */
	void render()
	{
		glClear(GL_DEPTH_BUFFER_BIT);
		glEnable(GL_DEPTH_TEST);

//		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
//		glCullFace(GL_FRONT);
//		glDisable(GL_CULL_FACE);

		glEnable(GL_LIGHTING);

		GLfloat global_ambient[] = { 0.5f, 0.5f, 0.5f, 1.0f };
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
		
		GLfloat position[] = { 1.5f, 2.0f, 3.0f, 1.0f };
		glLightfv(GL_LIGHT0, GL_POSITION, position);

		glEnable(GL_LIGHT0);


		computeFrustumStretch();

		/**
		 * new PROJECTION and MODELVIEW MATRIX
		 */
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();

		glFrustum(	-frustum_stretch_x, frustum_stretch_x,
					-frustum_stretch_y, frustum_stretch_y,
					1, ball_dist+ball_size+1.0);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

			glTranslatef(0, 0, -ball_dist);
			
			if (drag_started)
			{
				// draw start sphere
				glColor3f(0.5, 0.0, 0.0);
				CPrimitives::sphere(	0.05,
										drag_position_start[0],
										drag_position_start[1],
										drag_position_start[2],
										10);

				CPrimitives::sphere(	0.05,
										drag_position_end[0],
										drag_position_end[1],
										drag_position_end[2],
										10);
			}

			// apply current rotation
			glMultMatrixf(rotationMatrixArray);

			// draw the sphere
			glColor3f(0.5, 0.5, 0.5);

			// BALL
			CPrimitives::sphere(	ball_size,
									0, 0, 0,
									100.0);

			glClear(GL_DEPTH_BUFFER_BIT);

			// X AXIS
			glColor3f(1.0, 0.0, 0.0);
			glBegin(GL_LINES);
				glVertex3f(-ball_size*1.1, 0, 0);
				glVertex3f(ball_size*1.1, 0, 0);
			glEnd();

			CPrimitives::sphere(	0.05,	1,0,0,	10);
	

			// Y AXIS
			glColor3f(0.0, 1.0, 0.0);
			glBegin(GL_LINES);
				glVertex3f(0, -ball_size*1.1,  0);
				glVertex3f(0, ball_size*1.1,  0);
			glEnd();

			CPrimitives::sphere(	0.05,	0,1,0,	10);
	
	
			// Z AXIS
			glColor3f(0.0, 0.0, 1.0);
			glBegin(GL_LINES);
				glVertex3f(0, 0, -ball_size*1.1);
				glVertex3f(0, 0, ball_size*1.1);
			glEnd();

			CPrimitives::sphere(	0.05,	0,0,-1,	10);

			if (drag_started && false)
			{
				// draw the sphere
				// end position
				glColor3f(1.0, 0.0, 0.0);
				CPrimitives::sphere(	0.05,
										drag_position_start[0],
										drag_position_start[1],
										drag_position_start[2],
										10);
/*
				CPrimitives::sphere(	0.05,
										drag_position_end[0],
										drag_position_end[1],
										drag_position_end[2],
										10);
*/
			}

		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
	}
		
	/**
	 * store the rotation matrix in the array matrix
	 * \param matrix	float[4][4] matrix to store the rotation matrix
	 */
	void getRotationMatrix(T *matrix)
	{
			matrix[0]	= rotationMatrixCurrent[0][0];
			matrix[1]	= rotationMatrixCurrent[0][1];
			matrix[2]	= rotationMatrixCurrent[0][2];
			matrix[3]	= rotationMatrixCurrent[0][3];

			matrix[4]	= rotationMatrixCurrent[1][0];
			matrix[5]	= rotationMatrixCurrent[1][1];
			matrix[6]	= rotationMatrixCurrent[1][2];
			matrix[7]	= rotationMatrixCurrent[1][3];

			matrix[8]	= rotationMatrixCurrent[2][0];
			matrix[9]	= rotationMatrixCurrent[2][1];
			matrix[10]	= rotationMatrixCurrent[2][2];
			matrix[11]	= rotationMatrixCurrent[2][3];

			matrix[12]	= rotationMatrixCurrent[3][0];
			matrix[13]	= rotationMatrixCurrent[3][1];
			matrix[14]	= rotationMatrixCurrent[3][2];
			matrix[15]	= rotationMatrixCurrent[3][3];
	}
	
};



#endif
