/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <GL/gl.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
#include "dspObjfile.hpp"
#include "../libmath/CVector.hh"


using namespace std;


static void renderGroup(CObjFile &o, obj_group_s *g)
{
	int *f;

	f = g->faces4;
	glBegin(GL_QUADS);
	for (int c = 0; c < g->faces4_count; c++)
	{
		glNormal3fv(&o.normals[*f][0]);
		glVertex3fv(&o.vertices[*f][0]);
		f++;
		glNormal3fv(&o.normals[*f][0]);
		glVertex3fv(&o.vertices[*f][0]);
		f++;
		glNormal3fv(&o.normals[*f][0]);
		glVertex3fv(&o.vertices[*f][0]);
		f++;
		glNormal3fv(&o.normals[*f][0]);
		glVertex3fv(&o.vertices[*f][0]);
		f++;
	}
	glEnd();

	f = g->faces3;
	glBegin(GL_TRIANGLES);
	for (int c = 0; c < g->faces3_count; c++)
	{
		glNormal3fv(&o.normals[*f][0]);
		glVertex3fv(&o.vertices[*f][0]);
		f++;
		glNormal3fv(&o.normals[*f][0]);
		glVertex3fv(&o.vertices[*f][0]);
		f++;
		glNormal3fv(&o.normals[*f][0]);
		glVertex3fv(&o.vertices[*f][0]);
		f++;
	}
	glEnd();
}


/**
 * render the objdata in CObjFile class
 * \param o	reference to CObjFile class
 * \param drawOnlyMesh	draw only mesh without the border named 'world'
 */
bool renderObjfile(CObjFile &o, bool drawOnlyMesh, float cr, float cg, float cb)
{
	if (!o.ready)	return false;

	for (int i = 0; i < o.group_count; i++)
	{
		obj_group_s *g = &o.groups[i];

		if (strcmp(g->name, "world") != 0)
		{
			if (strncmp(g->name, "input", 5) == 0)
			{
				glColor3f(0.2f, 0.7f, 0.2f);
			}
			else if (strncmp(g->name, "output", 6) == 0)
			{
				glColor3f(0.7f, 0.2f, 0.2f);
			}
			else
			{
				glColor3f(cr, cg, cb);
			}

			if (!drawOnlyMesh)
				renderGroup(o, g);
		}

		if (drawOnlyMesh)
		{
			GLboolean cullface;
			glGetBooleanv(GL_CULL_FACE, &cullface);

//			glDisable(GL_CULL_FACE);
//			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

//			glColor3f(0.5, 0.5, 0.5);
			renderGroup(o, g);

//			if (cullface)	glEnable(GL_CULL_FACE);
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		}
	}
	return 0;
}




/*
 * render a objFile for voxel creation
 * the objfile should be renderes for the area [-1; 1]
 */
static void renderVoxelGroup(CObjFile &o, obj_group_s *g)
{
	int *f;

	f = g->faces3;
	glBegin(GL_TRIANGLES);
	for (int c = 0; c < g->faces3_count; c++)
	{
		glVertex3fv(&o.vertices[f[0]][0]);
		glVertex3fv(&o.vertices[f[1]][0]);
		glVertex3fv(&o.vertices[f[2]][0]);
		f += 3;
	}
	glEnd();

	f = g->faces4;
	glBegin(GL_QUADS);
	for (int c = 0; c < g->faces4_count; c++)
	{
		glVertex3fv(&o.vertices[f[0]][0]);
		glVertex3fv(&o.vertices[f[1]][0]);
		glVertex3fv(&o.vertices[f[2]][0]);
		glVertex3fv(&o.vertices[f[3]][0]);
		f += 4;
	}
	glEnd();
}



/**
 * render the objdata for voxel generation - without any unneccesary information!
 *
 * \param o	reference to CObjFile class
 * \param voxelRes	draw borders
 */
bool renderObjfileVoxel(CObjFile &o, bool drawBorder)
{
	if (!o.ready)	return false;

	for (int i = 0; i < o.group_count; i++)
		if (drawBorder)
		{
			renderVoxelGroup(o, &o.groups[i]);
		}
		else
		{
			if (strcmp(o.groups[i].name, "world") != 0)
				renderVoxelGroup(o, &o.groups[i]);
		}

	return 0;
}

/**
 * WITH NORMALS
 *
 * \param o	reference to CObjFile class
 * \param voxelRes	draw borders
 */
bool renderObjfileVoxelWithNormals(CObjFile &o, bool drawBorder)
{
	if (!o.ready)	return false;

	for (int i = 0; i < o.group_count; i++)
		if (drawBorder)
		{
			renderGroup(o, &o.groups[i]);
		}
		else
		{
			if (strcmp(o.groups[i].name, "world") != 0)
				renderGroup(o, &o.groups[i]);
		}

	return 0;
}

