/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "gltools.hpp"
#include "SDL_image.h"



bool unFrustum(float *matrix, float *left, float *right, float *bottom, float *top, float *near, float *far)
{
	float a = matrix[0];
	float c = matrix[4*2];
	float d = matrix[4*1 + 1];
	float e = matrix[4*2 + 1];
	float g = matrix[4*2 + 2];
	float h = matrix[4*3 + 2];
	
	if (matrix[4*2+3] != -1.0f)	// funny check
		return false;

	*near = (((g+1)/(g-1))*h - h)/2;
	*far = *near*(g-1)/(g+1);
	*left = (*near*2)/(((c+1)/(c-1)-1)*a);
	*right = *left*(c+1)/(c-1);
	*bottom = (*near*2)/(((e+1)/(e-1)-1)*d);
	*top = *bottom*(e+1)/(e-1);

	return true;
}


void printMatrix(float *matrix)
{
	int p;
	for (int i = 0; i < 16; i++)
	{
		if (i % 4 == 0)        
		{
			if (i / 4 == 0)	printf("\n / ");
			else if (i / 4 == 3)	printf("\n \\ ");
			else printf("\n | ");
		}
		p = (i % 4) * 4 + (i / 4);
		printf("%.2f\t", matrix[p]);
		if (i % 4 == 3)        
		{
			if (i / 4 == 0)	printf("\\");
			else if (i / 4 == 3)	printf("/");
			else printf("|");
		}

	}
	printf("\n");
}



void checkGlError_(const char *file, const char *function, int line)
{
	int nr = glGetError();
	const char *error = 0;
	switch(nr)
	{
		case GL_NO_ERROR:
			error = 0;
			break;

		case GL_INVALID_ENUM:
			error = "INVALID_ENUM";
			break;

		case GL_INVALID_VALUE:
			error = "INVALID_VALUE";
			break;
		
		case GL_INVALID_OPERATION:
			error = "INVALID_OPERATION";
			break;

		case GL_STACK_OVERFLOW:
			error = "INVALID_OVERFLOW";
			break;

		case GL_STACK_UNDERFLOW:
			error = "INVALID_UNDERFLOW";
			break;

		case GL_OUT_OF_MEMORY:
			error = "OUT_OF_MEMORY";
			break;
/*
		case GL_TABLE_TOO_LARGE:
			error = "TABLE_TOO_LARGE";
			break;
*/
		default:
			char unknown[100];
			sprintf(unknown, "UNKNOWN %i", nr);
			error = unknown;
	}

	if (error)
	{
		fprintf(stderr, "\033[0;40;31m");
		fprintf(stderr, "GL ERROR  in   %s(%i): %s() -> %s", file, line, function, error);
		fprintf(stderr, "\033[0m");
		fprintf(stderr, "\n");
	}

}




GLuint genMirrorTexture(int iTexSize)
{
	printf("Generating mirror texture...");

	GLuint uiTexture;
	GLubyte *noiseData = new GLubyte[iTexSize*iTexSize*4];
	GLubyte gbVal, aVal;

	for (int x = 0; x<iTexSize; x++) {
		for (int y = 0; y<iTexSize; y++) {
		
			int i = (x*iTexSize+y)*4;
			//float fFadeFac = 1;

			gbVal = rand()%40;

			aVal = rand()%20+80;
			noiseData[i+3]	= aVal;

			noiseData[i+0] = gbVal;	// fFadeFac
			noiseData[i+1] = gbVal;	// fFadeFac
			noiseData[i+2] = gbVal;	// fade to blue
		}
	}

	glGenTextures(1,&uiTexture);
	glBindTexture(GL_TEXTURE_2D,uiTexture);

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8,iTexSize,iTexSize,0, GL_RGBA, GL_UNSIGNED_BYTE, noiseData);

	delete[] noiseData;

	printf(" done.\n");

	return uiTexture;
}



