/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * version 0.2
 *
 * history:
 *
 * 2007-09-02:
 * - removed loadTexture because this is done by CTexture.hpp
 * - renamed WaitFPS to limit_fps
 */

#ifndef GL_TOOLS_H__
#define GL_TOOLS_H__

#include <GL/gl.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>




/*
 * check for gl error and output the line and function where the error occured
 */
#if CHECK_GL_ERROR_OFF
	#define checkGlError()
#else
	#define checkGlError()	checkGlError_(__FILE__, __FUNCTION__, __LINE__)
	void checkGlError_(const char *file, const char *function, int line);
#endif

/*
 * generate a mirror-like texture
 * and return a texture handler
 */
extern GLuint genMirrorTexture(int iTexSize);

/**
 * recalculate the frustum from a matrix and return the values to parameters
 */
bool unFrustum(float *matrix, float *left, float *right, float *bottom, float *top, float *near, float *far);

/**
 * output a 4x4 matrix
 */
void printMatrix(float *matrix);



/*
 * time measure features
 */
/*
// init time structures
#define initTakeTime(timedataname)	static struct timeval timedataname##time_start, timedataname##time_end;

// start taking the time and return the last delta value
#define takeTime(timedataname)	startTakeTime_(&timedataname##time_start, &timedataname##time_end)
inline unsigned int startTakeTime_(struct timeval *time1, struct timeval *time2)
{
	int delta;

	gettimeofday(time2, 0);

	delta = ((time2->tv_sec - time1->tv_sec) * 1000000 + (time2->tv_usec - time1->tv_usec));
	
	time1->tv_sec = time2->tv_sec;
	time1->tv_usec = time2->tv_usec;

	return delta;
}


#define takeTimef(timedataname)	((float)takeTime(timedataname)/1000000)

#define endTakeTimei(timedataname)	endTakeTime_(&timedataname##time_start, &timedataname##time_end)
#define endTakeTimef(timedataname)	((float)endTakeTime(timedataname)/1000000)
inline unsigned int endTakeTime_(struct timeval *time1, struct timeval *time2)
{
	gettimeofday(time2, 0);
	return ((time2->tv_sec - time1->tv_sec) * 1000000 + (time2->tv_usec - time1->tv_usec));
}
*/

/*
 * INVERT MODELVIEW 
 *
 * As discussed in class. Works only with translates & rotates !
 */
inline void GetInvertedModelview(float invM[16]) 
{
	float M[16];
	glGetFloatv(GL_MODELVIEW_MATRIX,M);
	// Let R=upper left 3x3. R is orthonormal (rotation-matrix) <=> R^-1=R^T.
	invM[ 0]=M[ 0];	invM[ 4]=M[ 1];	invM[ 8]=M[ 2];
	invM[ 1]=M[ 4];	invM[ 5]=M[ 5];	invM[ 9]=M[ 6];
	invM[ 2]=M[ 8];	invM[ 6]=M[ 9];	invM[10]=M[10];
	// Let v=upper right 3D-vector. v^-1=-R^T*v. v is the translation vector.
	invM[12]=-invM[ 0]*M[12]-invM[ 4]*M[13]-invM[ 8]*M[14];
	invM[13]=-invM[ 1]*M[12]-invM[ 5]*M[13]-invM[ 9]*M[14];
	invM[14]=-invM[ 2]*M[12]-invM[ 6]*M[13]-invM[10]*M[14];
	// rest is just like ID
	invM[ 3]=0.0f;		invM[ 7]=0.0f;		invM[11]=0.0f;		invM[15]=1.0f;
}


/* 
 * MATRIX VECTOR product, partial 3x3 product
 *
 * v is overwritten with result.
 * No perspective division by w performed
 */
inline void MatrixVectorProd3(float M[16], float v[3]) 
{
	float a=M[0]*v[0]+M[4]*v[1]+M[8]*v[2];
	float b=M[1]*v[0]+M[5]*v[1]+M[9]*v[2];
	v[2]=M[2]*v[0]+M[6]*v[1]+M[10]*v[2];
	v[0]=a;
	v[1]=b;
}


inline void MatrixMatrixProd(float a[16], float b[16]) {
	float c[16];
	for (int x = 0;x<16;x+=4)
		for (int y = 0;y<4;y++) 
			c[x+y] = b[0+y]*a[0+x]+b[4+y]*a[1+x]+b[8+y]*a[2+x]+b[12+y]*a[3+x];

	for (int i=0;i<16;i++) a[i] = c[i];	
}

#define clamp(value, min, max)	(value > max ? max : (value < min ? min : value))


#define DEFAULT_MAX_FPS	30

/**
 * function to limit fps
 * wait until 'fps' fps are reached
 * \param fps	fps to limit simulation
 * \param ticks_actual	actual number of ticks (from sdl)
 */
#define limit_fps(fps, ticks_actual)					\
{									\
	static signed int last_frame_ticks = 0;				\
	signed int delta = ticks_actual - last_frame_ticks;		\
	signed int milli_secs_per_frame = 1000/fps;			\
									\
	if (delta < milli_secs_per_frame)	usleep((milli_secs_per_frame - delta)*1000);	\
	last_frame_ticks += milli_secs_per_frame;			\
									\
	if (last_frame_ticks < ticks_actual)				\
		last_frame_ticks = ticks_actual;			\
}

/*
//	static unsigned int lastframeticks = 0;					\
	unsigned int delta = ticks_actual - lastframeticks;			\
	unsigned int milli_secs_per_frame = 1000/fps;						\
										\
	if (delta < milli_secs_per_frame)	usleep((milli_secs_per_frame - delta)*1000);				\
										\
	if (lastframeticks < ticks_actual)					\
		lastframeticks += (delta%milli_secs_per_frame == 0 ? delta : delta - delta%milli_secs_per_frame + milli_secs_per_frame);	\
	else									\
		lastframeticks = ticks_actual;					\
*/

#endif


#define fbo_write_save_state()							\
	GLboolean ss_lighting, ss_blend, ss_depth_test, ss_cull_face, ss_texture_2d;		\
	GLint ss_texture_env_mode;						\
	glGetBooleanv(GL_LIGHTING, &ss_lighting);				\
	glGetBooleanv(GL_BLEND, &ss_blend);					\
	glGetBooleanv(GL_DEPTH_TEST, &ss_depth_test);				\
	glGetBooleanv(GL_CULL_FACE, &ss_cull_face);				\
	glGetBooleanv(GL_TEXTURE_2D, &ss_texture_2d);				\
	glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &ss_texture_env_mode);	\
										\
	glDisable(GL_LIGHTING);							\
	glDisable(GL_BLEND);							\
	glDisable(GL_DEPTH_TEST);						\
	glDisable(GL_CULL_FACE);						\
	glEnable(GL_TEXTURE_2D);						\
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);


#define fbo_write_restore_state()						\
	if (ss_lighting)	glEnable(GL_LIGHTING);				\
	if (ss_blend)	glEnable(GL_BLEND);					\
	if (ss_depth_test)	glEnable(GL_DEPTH_TEST);			\
	if (ss_cull_face)	glEnable(GL_CULL_FACE);				\
	if (ss_texture_2d)	glEnable(GL_TEXTURE_2D);			\
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, ss_texture_env_mode);

