/*
 * This file is part of the Lattice Boltzmann Open Surface program (lbm_os).
 *
 * lbm_os is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * lbm_os is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with lbm_os.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include <SDL.h>
#include <stdio.h>
#include "CSimulation.hpp"
#include "CGuiSimulation.hpp"
#include "libviz/gltools.hpp"
#include "libviz/CObjFile.hpp"
#include <iostream>
#include <unistd.h>

using namespace std;

bool quit_program = false;
CSimulation *simulation;
CGuiSimulation *gui_simulation;


int main(int argc, char *argv[])
{
#ifdef __WIN32__
	// avoid output on console
	// the program will crash if the second line is created with cout
	freopen(NULL, "w", stdout );
    freopen(NULL, "w", stderr );
#endif
	
	char *objfile_path = 0;
	int res_x = 50, res_y = 50, res_z = 50;
	float gravitation = 9.81;
	float viscosity = 0.012f;
	float length_x = 1.0f;
	bool run_simulation = true;
	bool take_screenshots = false;

	char optchar;
        while ((optchar = getopt(argc, argv, "o:x:y:z:v:g:l:ns")) > 0)
	{
		switch(optchar)
		{
		case 'o':
			objfile_path = optarg;
			break;

		case 's':
			take_screenshots = true;
			break;

		case 'n':
			run_simulation = false;
			break;

		case 'l':
			length_x = atof(optarg);
			if (length_x <= 0)
			{
				cerr << "error: length_x has to be greater than 0" << endl;
				goto parameter_error;
			}
			break;

		case 'x':
			res_x = atoi(optarg);
			if (res_x < 1)
			{
				cerr << "error: resolution has to be greater than 0" << endl;
				goto parameter_error;
			}
			break;

		case 'y':
			res_y = atoi(optarg);
			if (res_y < 1)
			{
				cerr << "error: resolution has to be greater than 0" << endl;
				goto parameter_error;
			}
			break;

		case 'z':
			res_z = atoi(optarg);
			if (res_z < 1)
			{
				cerr << "error: resolution has to be greater than 0" << endl;
				goto parameter_error;
			}
			break;

		case 'g':
			gravitation = atof(optarg);
			break;


		case 'v':
			viscosity = atof(optarg);
			break;

		default:
			cerr << "Unknown option '" << optchar << "'" << endl;
			goto parameter_error;
		}
	}

	if (objfile_path == 0)
	{
		cerr << "No objfile selected" << endl;
		goto parameter_error;
	}

	goto parameters_ok;
parameter_error:
	printf("	-n: dont run simulation immediately\n");
	printf("	-s: take screenshot every frame\n");

	cout << "usage: " << argv[0] << " [-o objfile] [-x resolution_x] [-y resolution_y] [-z resolution_z] [-v viscosity] [-l length along x] [-n] [-s]" << endl;
	cout << "	-n: dont run simulation immediately" << endl;
	cout << "	-s: take screenshot every frame" << endl;
	return -1;

parameters_ok:

	CObjFile *objFile = new CObjFile(objfile_path);
	if (!objFile->ready)
		return -1;

	CGui *gui = new CGui();

	GLenum err = glewInit();
	if (err != GLEW_OK)
	{
		fprintf(stderr, "glewInit failed: %s\n", glewGetErrorString(err));
		return false;
	}

// initialize simulation
	simulation = new CSimulation(objFile, gui, res_x, res_y, res_z, viscosity, length_x, gravitation);
	if (!simulation->ready)
	{
		delete simulation;
		return 1;
	}

// initialize gui for simulation
	gui_simulation = new CGuiSimulation(gui, simulation, run_simulation, take_screenshots);
	if (gui_simulation->error)
	{
		delete gui_simulation;
		delete simulation;
		return 2;
	}


// main loop
//	int main_ticks = gui->get_ticks(), main_ticks_new, main_tick_values = 0;
//	int computation_ticks = 0;
	while (!quit_program)
	{
		gui->init_rendering();

		if (gui_simulation->steps_to_do)
		{
			simulation->computate();

			quit_program = gui_simulation->render();

			// modify this after gui because of particle emission
			gui_simulation->steps_to_do--;
			gui_simulation->simulation_steps++;
		}
		else
		{
			quit_program = gui_simulation->render();
		}

		gui->finish_rendering();
		gui->event_loop();
	}

	delete objFile;
	delete gui_simulation;
	delete simulation;
	delete gui;

	return 0;
}
